/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/


////////////////////////////////////////////////////////////////////
//
// olap_bkend_thread.cpp
//
// This file has the implementation of the backend threads. These threads
// take a sql request, forward them to the backend and fetch the results.
//
// Author: Karthikeyan Ramasamy
// Modified by: Prasad M. Deshpande
//
/////////////////////////////////////////////////////////////////////


#if defined(SOLARIS2)
#define BSD_COMP
#include <sys/ioctl.h>
#undef BSD_COMP
#endif

#include <stdlib.h>

#include "olap_comm_shore.h"
#include "olap_bkend_request.h"
#include "olap_bkend_thread.h" 
#include "olap_tuple_network_stream.h"

const char OLAP_TMP_DIR[]	= "/tmp/" ; 
const char OLAP_USERNAME[] = "sa";
const char OLAP_PASSWD[] = "samitsybase";

static bool willReadBlock( SOCKET m_isSocket )
{
	fd_set		rset ; 
	FD_ZERO(&rset) ; 
	FD_SET(m_isSocket, &rset) ; 
	struct timeval timeout;
	timeout.tv_sec  = 0; timeout.tv_usec = 0;

	int32 n = select( m_isSocket +1 , &rset, NULL, NULL, &timeout ) ;  

	if ( n < 0 ) W_FATAL( fcOS ) ; 

        return n > 0 ? true : false ;
}

SOCKET
C_OLAPBkend_Thread::Synchronize() 
{
	ostrstream	sname ;
	sname << OLAP_TMP_DIR << ".borg." << getlogin() << '.'
		<< m_isSchemaName.c_str() 
		<< "." << m_iMyself << ends ; 

	SOCKET		isSocket = INVALID_SOCKET ; 

	do	
	{
		isSocket = 
		C_OLAPComm_System::CreateMachineDomainCltSocket(sname.str()) ; 

		if ( isSocket == INVALID_SOCKET ) C_OLAPThread_Thread::Yield() ;
	}
	while ( isSocket == INVALID_SOCKET ) ; 

	sname.rdbuf()->freeze(0) ; 

	return isSocket ; ; 
}

vo_ptr C_OLAPBkend_Thread::Run()
{
	strstream	sname ;
	sname << OLAP_TMP_DIR << ".borg." << getlogin() << '.'
		<< m_isSchemaName.c_str() 
		<< "." << m_iMyself << ends ; 

	SOCKET isSocket =
		 C_OLAPComm_System::CreateMachineDomainSvrSocket(sname.str()) ; 

	sname.freeze(0) ; 

	if ( isSocket == SOCKET_ERROR )
	{
		cout << lockp << "Socket creation error" << endl << unlockp ;
		return NULL ; 
	}

	SOCKET isAcceptSocket = 
		 C_OLAPComm_System::AcceptMachineDomainSocket(isSocket) ; 
		
	if ( isAcceptSocket == SOCKET_ERROR )
	{
		cout << lockp << "Socket accept error" << endl << unlockp ;
		return NULL ; 
	}

	C_OLAPComm_System::TearDownSocket(isSocket) ; 

	/*
	if ( setsockopt( isAcceptSocket, SOL_SOCKET,
                         SO_SNDBUF, (char *)&one, sizeof(one)) < 0 )
	{
		cout << lockp << " send buf not able to set" << endl 
		     << unlockp ;  
	}
	*/

	C_OLAPTuple_NativeNetwork_Stream 	tstream(isAcceptSocket) ; 

	fd_set			rset ; 
	FD_ZERO(&rset) ; 
	FD_SET(isAcceptSocket, &rset) ; 

	cout << lockp << "Opening backend db " << endl << unlockp;

	while ( 1 )
	{	
		bool bytes_ready = !willReadBlock( isAcceptSocket ) ; 

		if ( !bytes_ready )
		{ 
			cerr << lockp << "Waiting in select " << unlockp << endl ; 
			int n = select(isAcceptSocket+1, &rset, NULL, NULL, NULL) ;

			if ( n < 0 ) break ;
		}
		else
			cerr << lockp << "Bytes are ready " << unlockp << endl ; 

		if ( bytes_ready || FD_ISSET(isAcceptSocket, &rset) )
		{
			C_OLAPBkend_Request*	request ; 
			if ( recv(isAcceptSocket, (char *)&request, 
					sizeof(request), 0) == SOCKET_ERROR )  
				return NULL ; 

			if ( request->RequestType() == 
					C_OLAPBkend_Request::OLAP_BKEND_SHUTDOWN )
				break ; 

			w_assert3(request->RequestType() == C_OLAPBkend_Request::OLAP_BKEND_SQL )
			C_OLAPBkend_Sql_Request *sql_request;

			sql_request = (C_OLAPBkend_Sql_Request *)request;


			tstream.Reset() ; 
                }

		FD_ZERO(&rset) ; 
		FD_SET(isAcceptSocket, &rset) ; 
	}

	// Reset the stream
	tstream.Reset() ; 

	// Close the connection
	C_OLAPComm_System::TearDownSocket(isAcceptSocket) ; 

	return NULL ; 
}
