/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// apb_queries.cpp
//
// Defines functions for generating query trees corresponding to 
// APB benchmark queries.
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////



// This file contains a test query tree
#include <iostream.h>

#include "querytree.h"
#include "qt_attr.h"
#include "qt_expr.h"
#include "qt_join.h"
#include "qt_measure.h"

#include "borgtypes.h"

extern  ifstream fin;

#define FIELD_LENGTH          12

pQT_MDQuery Q1_ConstructQuery();
pQT_LogicalExpr Q1_ConstructLogicalExpr();
pQT_ProjList Q1_ConstructProjList();


pQT_MDQuery     Q2_ConstructQuery();
pQT_LogicalExpr Q2_ConstructLogicalExpr();
pQT_ProjList    Q2_ConstructProjList();

/*
 pQT_MDQuery     Q3_ConstructQuery();
 pQT_LogicalExpr Q3_ConstructLogicalExpr();
 pQT_ProjList    Q3_ConstructProjList();
*/

pQT_MDQuery     Q4_ConstructQuery();
pQT_LogicalExpr Q4_ConstructLogicalExpr();
pQT_ProjList    Q4_ConstructProjList();


pQT_MDQuery     Q5_ConstructQuery();
pQT_LogicalExpr Q5_ConstructLogicalExpr();
pQT_ProjList    Q5_ConstructProjList();


pQT_MDQuery     Q6_ConstructQuery();
pQT_LogicalExpr Q6_ConstructLogicalExpr();
pQT_ProjList    Q6_ConstructProjList();


pQT_MDQuery     Q7_ConstructQuery();
pQT_LogicalExpr Q7_ConstructLogicalExpr();
pQT_ProjList    Q7_ConstructProjList();


pQT_MDQuery     Q8_ConstructQuery();
pQT_LogicalExpr Q8_ConstructLogicalExpr();
pQT_ProjList    Q8_ConstructProjList();


pQT_MDQuery     Q9_ConstructQuery();
pQT_LogicalExpr Q9_ConstructLogicalExpr();
pQT_ProjList    Q9_ConstructProjList();


pQT_MDQuery     Q10_ConstructQuery();
pQT_LogicalExpr Q10_ConstructLogicalExpr();
pQT_ProjList    Q10_ConstructProjList();


extern pQT_MDQuery ConstructQuery();
extern pQT_MDQuery ConstructQuery2();

//////////////////////////////////////////////////////////
//////////////// Q U E R Y  0 1 //////////////////////////
//////////////////////////////////////////////////////////

pQT_MDQuery Q1_ConstructQuery()
{
    pQT_MDQuery     pmdq;
    pQT_StarJoin    psj;
    pQT_LogicalExpr ple;
    pQT_ProjList    ppl;
    pQT_Measure     pm;

    ple = Q1_ConstructLogicalExpr();
    ppl = Q1_ConstructProjList();

    psj = new QT_StarJoin;
    psj->AddArg(pQT_MDQueryNode(ple));
    psj->AddArg(pQT_MDQueryNode(ppl));

    // Construct the measures HS_DollarSales
    pm = new QT_Measure(Sum);
    pQT_Attr    pa = new QT_Attr;
    pa->L_AttrName() = "HS_DollarSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "HS_UnitSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the query
    pmdq = new QT_MDQuery;
    pmdq->AddArg(pQT_MDQueryNode(psj));

    return pmdq;
}



pQT_LogicalExpr Q1_ConstructLogicalExpr()
{
    pQT_LogicalExpr pleAND;
    pQT_CompExpr    pce;
    pQT_Attr        pa;
    pQT_Value       pv;

    char buf[32];       // To read tuples from stream

    // Expr is:
    //     ScenarioABF = "ACTUAL" 
    // AND PCode = children<?product> 
    // AND CStore = children<?customer>
    // AND ChBase = <?channel>
    // AND TMon = children<?time>

    // Build the tree from the left up
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND = new QT_LogicalExpr(QT_LogicalExpr::And);
    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "ChBase";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ---------------------------------------------- */

    return pleAND;
}



pQT_ProjList Q1_ConstructProjList()
{
    pQT_Attr pa;
    pQT_ProjList ppl = new QT_ProjList;

    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "ChBase";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";
    ppl->AddArg(pQT_MDQueryNode(pa));

    return ppl;
}



//////////////////////////////////////////////////////////
//////////////// Q U E R Y  0 2 //////////////////////////
//////////////////////////////////////////////////////////

pQT_MDQuery Q2_ConstructQuery()
{
    pQT_MDQuery     pmdq;
    pQT_StarJoin    psj;
    pQT_LogicalExpr ple;
    pQT_ProjList    ppl;
    pQT_Measure     pm;

    ple = Q2_ConstructLogicalExpr();
    ppl = Q2_ConstructProjList();

    psj = new QT_StarJoin;
    psj->AddArg(pQT_MDQueryNode(ple));
    psj->AddArg(pQT_MDQueryNode(ppl));

    // Construct the measures HS_DollarSales
    pm = new QT_Measure(Sum);
    pQT_Attr    pa = new QT_Attr;
    pa->L_AttrName() = "HS_UnitSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "PC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "SC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the query
    pmdq = new QT_MDQuery;
    pmdq->AddArg(pQT_MDQueryNode(psj));

    return pmdq;
}



pQT_LogicalExpr Q2_ConstructLogicalExpr()
{
    pQT_LogicalExpr pleAND;
    pQT_CompExpr    pce;
    pQT_Attr        pa;
    pQT_Value       pv;

    char buf[32];       // To read tuples from stream

    // Expr is:
    //     ScenarioABF = "ACTUAL" 
    // AND PCode = children<?product> 
    // AND CStore = children<?customer>
    // AND ChBase = <?channel>
    // AND TMon = children<?time>

    // Build the tree from the left up
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND = new QT_LogicalExpr(QT_LogicalExpr::And);
    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";

    pv = new QT_Value;
    pv->val() = "Actual" ;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ---------------------------------------------- */

    return pleAND;
}



pQT_ProjList Q2_ConstructProjList()
{
    pQT_Attr pa;
    pQT_ProjList ppl = new QT_ProjList;

    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";
    ppl->AddArg(pQT_MDQueryNode(pa));

    return ppl;
}



//////////////////////////////////////////////////////////
//////////////// Q U E R Y  0 4 //////////////////////////
//////////////////////////////////////////////////////////

pQT_MDQuery Q4_ConstructQuery()
{
    pQT_MDQuery     pmdq;
    pQT_StarJoin    psj;
    pQT_LogicalExpr ple;
    pQT_ProjList    ppl;
    pQT_Measure     pm;

    ple = Q4_ConstructLogicalExpr();
    ppl = Q4_ConstructProjList();

    psj = new QT_StarJoin;
    psj->AddArg(pQT_MDQueryNode(ple));
    psj->AddArg(pQT_MDQueryNode(ppl));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pQT_Attr    pa = new QT_Attr;
    pa->L_AttrName() = "HS_UnitSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the query
    pmdq = new QT_MDQuery;
    pmdq->AddArg(pQT_MDQueryNode(psj));

    return pmdq;
}



pQT_LogicalExpr Q4_ConstructLogicalExpr()
{
    pQT_LogicalExpr pleAND;
    pQT_CompExpr    pce;
    pQT_Attr        pa;
    pQT_Value       pv;

    char buf[32];       // To read tuples from stream

    // Expr is:
    //     ScenarioABF = "ACTUAL" 
    // AND PCode = children<?product> 
    // AND CStore = children<?customer>
    // AND ChBase = <?channel>
    // AND TMon = children<?time>

    // Build the tree from the left up
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND = new QT_LogicalExpr(QT_LogicalExpr::And);
    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ---------------------------------------------- */

    return pleAND;
}



pQT_ProjList Q4_ConstructProjList()
{
    pQT_Attr pa;
    pQT_ProjList ppl = new QT_ProjList;

    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "ChBase";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";
    ppl->AddArg(pQT_MDQueryNode(pa));

    return ppl;
}



//////////////////////////////////////////////////////////
//////////////// Q U E R Y  0 5 //////////////////////////
//////////////////////////////////////////////////////////

pQT_MDQuery Q5_ConstructQuery()
{
    pQT_MDQuery     pmdq;
    pQT_StarJoin    psj;
    pQT_LogicalExpr ple;
    pQT_ProjList    ppl;
    pQT_Measure     pm;

    ple = Q5_ConstructLogicalExpr();
    ppl = Q5_ConstructProjList();

    psj = new QT_StarJoin;
    psj->AddArg(pQT_MDQueryNode(ple));
    psj->AddArg(pQT_MDQueryNode(ppl));

    // Construct the measures HS_DollarSales
    pm = new QT_Measure(Sum);
    pQT_Attr    pa = new QT_Attr;
    pa->L_AttrName() = "HS_UnitSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "PC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "SC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the query
    pmdq = new QT_MDQuery;
    pmdq->AddArg(pQT_MDQueryNode(psj));

    return pmdq;
}



pQT_LogicalExpr Q5_ConstructLogicalExpr()
{
    pQT_LogicalExpr pleAND;
    pQT_CompExpr    pce;
    pQT_Attr        pa;
    pQT_Value       pv;

    char buf[32];       // To read tuples from stream

    // Expr is:
    //     ScenarioABF = "ACTUAL" 
    // AND PCode = children<?product> 
    // AND CStore = children<?customer>
    // AND ChBase = <?channel>
    // AND TMon = children<?time>

    // Build the tree from the left up

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";

    pv = new QT_Value;
    pv->val() = "Budget";
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND = new QT_LogicalExpr(QT_LogicalExpr::And);
    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";

    pv = new QT_Value;
    pv->val() = "199601";
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Between);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pv = new QT_Value;
    pv->val() = "199612";
    pv->operation() = QT_Value::None;
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ---------------------------------------------- */

    return pleAND;
}



pQT_ProjList Q5_ConstructProjList()
{
    pQT_Attr pa;
    pQT_ProjList ppl = new QT_ProjList;

    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";
    ppl->AddArg(pQT_MDQueryNode(pa));

    return ppl;
}




//////////////////////////////////////////////////////////
//////////////// Q U E R Y  0 6 //////////////////////////
//////////////////////////////////////////////////////////

pQT_MDQuery Q6_ConstructQuery()
{
    pQT_MDQuery     pmdq;
    pQT_StarJoin    psj;
    pQT_LogicalExpr ple;
    pQT_ProjList    ppl;
    pQT_Measure     pm;

    ple = Q6_ConstructLogicalExpr();
    ppl = Q6_ConstructProjList();

    psj = new QT_StarJoin;
    psj->AddArg(pQT_MDQueryNode(ple));
    psj->AddArg(pQT_MDQueryNode(ppl));

    // Construct the measures HS_DollarSales
    pm = new QT_Measure(Sum);
    pQT_Attr    pa = new QT_Attr;
    pa->L_AttrName() = "HS_UnitSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "PC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "SC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the query
    pmdq = new QT_MDQuery;
    pmdq->AddArg(pQT_MDQueryNode(psj));

    return pmdq;
}



pQT_LogicalExpr Q6_ConstructLogicalExpr()
{
    pQT_LogicalExpr pleAND;
    pQT_CompExpr    pce;
    pQT_Attr        pa;
    pQT_Value       pv;

    char buf[32];       // To read tuples from stream

    // Expr is:
    //     ScenarioABF = "ACTUAL" 
    // AND PCode = children<?product> 
    // AND CStore = children<?customer>
    // AND ChBase = <?channel>
    // AND TMon = children<?time>

    // Build the tree from the left up

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";

    pv = new QT_Value;
    pv->val() = "Budget";
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND = new QT_LogicalExpr(QT_LogicalExpr::And);
    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";

    pv = new QT_Value;
    pv->val() = "1996";
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ---------------------------------------------- */

    return pleAND;
}



pQT_ProjList Q6_ConstructProjList()
{
    pQT_Attr pa;
    pQT_ProjList ppl = new QT_ProjList;

    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";
    ppl->AddArg(pQT_MDQueryNode(pa));

    return ppl;
}




//////////////////////////////////////////////////////////
//////////////// Q U E R Y  0 7 //////////////////////////
//////////////////////////////////////////////////////////

pQT_MDQuery Q7_ConstructQuery()
{
    pQT_MDQuery     pmdq;
    pQT_StarJoin    psj;
    pQT_LogicalExpr ple;
    pQT_ProjList    ppl;
    pQT_Measure     pm;

    ple = Q7_ConstructLogicalExpr();
    ppl = Q7_ConstructProjList();

    psj = new QT_StarJoin;
    psj->AddArg(pQT_MDQueryNode(ple));
    psj->AddArg(pQT_MDQueryNode(ppl));

    // Construct the measures HS_DollarSales
    pm = new QT_Measure(Sum);
    pQT_Attr    pa = new QT_Attr;
    pa->L_AttrName() = "HS_UnitSales";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "PC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the measures HS_UnitSales
    pm = new QT_Measure(Sum);
    pa = new QT_Attr;
    pa->L_AttrName() = "SC_Value";
    pm->AddArg(pQT_MDQueryNode(pa));
    psj->AddArg(pQT_MDQueryNode(pm));

    // Construct the query
    pmdq = new QT_MDQuery;
    pmdq->AddArg(pQT_MDQueryNode(psj));

    return pmdq;
}



pQT_LogicalExpr Q7_ConstructLogicalExpr()
{
    pQT_LogicalExpr pleAND;
    pQT_CompExpr    pce;
    pQT_Attr        pa;
    pQT_Value       pv;

    char buf[32];       // To read tuples from stream

    // Expr is:
    //     ScenarioABF = "ACTUAL" 
    // AND PCode = children<?product> 
    // AND CStore = children<?customer>
    // AND ChBase = <?channel>
    // AND TMon = children<?time>

    // Build the tree from the left up

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";

    pv = new QT_Value;
    pv->val() = "Forecast";
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND = new QT_LogicalExpr(QT_LogicalExpr::And);
    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));


    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";


    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::Children;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ------------------ */
    // Next operator
    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";

    pv = new QT_Value;
    fin >> buf;
    pv->val() = buf;
    pv->operation() = QT_Value::None;

    // Insert a children operator here...
    pce = new QT_CompExpr(QT_CompExpr::Eq);
    pce->AddArg(pQT_MDQueryNode(pa));
    pce->AddArg(pQT_MDQueryNode(pv));

    pleAND->AddArg(pQT_MDQueryNode(pce));

    /* ---------------------------------------------- */

    return pleAND;
}



pQT_ProjList Q7_ConstructProjList()
{
    pQT_Attr pa;
    pQT_ProjList ppl = new QT_ProjList;

    pa = new QT_Attr;
    pa->L_AttrName() = "ScenarioABF";
    ppl->AddArg(pQT_MDQueryNode(pa));
    
    pa = new QT_Attr;
    pa->L_AttrName() = "PCode";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "CStore";
    ppl->AddArg(pQT_MDQueryNode(pa));

    pa = new QT_Attr;
    pa->L_AttrName() = "TMon";
    ppl->AddArg(pQT_MDQueryNode(pa));

    return ppl;
}


