/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_client.cpp
//
// Implements BORG client. It reads input and generates the appropriate
// query to be issued to the BORG server.
//
// Authors: Karthikeyan Ramasamy, Amit Shukla, Prasad Deshpande
//
/////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include <math.h>

#include "olap_comm_generic.h"
#include "olap_tuple_client_native.h"
#include "olap_thread.h"

extern "C" long random() ; 

char* hostName = NULL ; 
u_int32         OlapMsgInquiryPort = 0;
u_int32         OlapMsgServicePort = 0;
u_int32         OlapClientServicePort = 0;
u_int32         OlapServerShutdownPort = 0;
u_int32         OlapDbServerPort = 0;

class C_OLAP_Client_Request_Thread : public C_OLAPThread_Thread
{
public :

	C_OLAP_Client_Request_Thread() 
	{ } ;  
	
	virtual ~C_OLAP_Client_Request_Thread()
	{ } ;

	vo_ptr Run() ; 
} ; 

void SetPortNumbers()
{
        OlapMsgInquiryPort = OLAP_MSG_INQUIRY_PORT;
        OlapMsgServicePort = OLAP_MSG_SERVICE_PORT;
        OlapClientServicePort = OLAP_CLIENT_SERVICE_PORT;
        OlapServerShutdownPort = OLAP_SERVER_SHUTDOWN_PORT;
        OlapDbServerPort = OLAP_DB_SERVER_PORT;


        ifstream inp(".ports");
        if (!inp) return;

        inp >> OlapMsgInquiryPort ;
        while (inp.get() != '\n');
        inp >> OlapMsgServicePort ;
        while (inp.get() != '\n');
        inp >> OlapClientServicePort ;
        while (inp.get() != '\n');
        inp >> OlapServerShutdownPort ;
        while (inp.get() != '\n');
        inp >> OlapDbServerPort ;
}


main( int argc, char* argv[] )
{
	SetPortNumbers();
	if ( argc < 3 || argc > 4 )
	{
		cerr << argv[0] << " <num-clients> <host-name> [<host-port>]" << endl ; 
		exit(1) ; 
	}

	if ( argc == 4 ) OlapClientServicePort = atoi(argv[3]) ; 
	hostName = argv[2] ; 


	// Initialize the comm system
	W_COERCE( C_OLAPComm_System::Initialize() ) ; 

	int32 nthreads = atoi( argv[1] ) ; 

	C_OLAP_Client_Request_Thread** th = new
				C_OLAP_Client_Request_Thread*[nthreads] ; 
	for ( int32 i = 0 ; i < nthreads ; i++ )
	{
		th[i] = new C_OLAP_Client_Request_Thread ; 
		if ( !th[i] ) W_FATAL( BORG_OUTOFMEMORY ) ;  
		th[i]->Start() ; 
	}

	for (int32 i=0; i<nthreads; i++)
	{
		th[i]->Wait();
		delete th[i];
	}

	delete [] th;

//	while (1) ;
}

ifstream 					fin(0);

vo_ptr C_OLAP_Client_Request_Thread::Run()
{ 
	// Connect to the server
	SOCKET s = C_OLAPComm_System::CreateNetDomainCltSocket( hostName, 
                                                                OlapClientServicePort ) ; 
	if ( s == INVALID_SOCKET )
	{
		cerr << "Unable to connect to server" << endl ; 
		exit(1) ; 
	}

	C_OLAPComm_System_Generic_Client		isClient(s,true) ; 
	C_OLAPComm_Reconciler				isReconciler ; 
	int32						nreq = 0 ; 

	while (! fin.eof())
	{
	    {
		// C_OLAPComm_Empty_Packet			isEmpty ; 	
                C_OLAP_TupleInfo 			tupInfo;
        	C_OLAP_Tuple     			tuple;

        	C_OLAPComm_DataLinearizer               isQuery ;
		int					querynum;

		// Send the queries from here... 

		fin >> querynum;
		if (fin.eof()) break;

        	cerr << "Sending Query Number " << querynum << endl;
        	switch(querynum)
        	{
		}
	
	        C_OLAPComm_Generic_Packet isPacket(isQuery) ;

		// Send a request 
		W_COERCE( isClient.SendData( isReconciler, isPacket ) ) ; 
		nreq++ ; 

        	// Receive the results
	        C_OLAPComm_Packet*              pPacket = NULL ;

 	       	if ( isClient.ReceiveData( isReconciler, pPacket ) <= 0 )
        	{
                	cerr << "Unable to recv results..." << endl ;
                	::exit(1) ;
        	}

        	C_OLAPComm_DataDelinearizer*            isNetData ;
        	isNetData = ((C_OLAPComm_Generic_Packet*)pPacket)->GetDataDelinearizer() ;

		isNetData->Unpack(tupInfo);
        	tuple.Init(tupInfo);
	
		// Get a stream of results back	
		C_OLAPTuple_NativeClientStream		clientStream(s) ;  

		int32					count = 0 ; 
		bool					eos ; 
		bool					exception ; 

		W_COERCE( clientStream.read(tuple, eos, exception ) ) ;
	

		while ( 1 )
		{
			if ( eos || exception )	break ; 

			tuple.print(cout);
			count++ ; 

			W_COERCE( clientStream.read(tuple, eos, exception ) ) ;
		}	

		cout << lockp 
                     << "Number of tuples received " << count 
		     << endl << unlockp ; 
		//if (nreq == 5) break;
	    }
	}

	C_OLAPComm_System::TearDownSocket( s ) ; 

	return NULL ; 
}
