/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_system_generic.cpp
//
// This file contains the implementations of the transport layer for
// sockets. The abstracted read/write interface is mapped onto the 
// socket read and write. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if defined(SOLARIS2)
#define BSD_COMP
#include <sys/ioctl.h>
#undef BSD_COMP
#endif

#include "olap_types.h"
#include "olap_comm_types.h"
#include "olap_comm_system_generic.h"

w_rc_t
C_OLAPComm_System_Generic::WillBlockRead( bool& block ) 
{
	fd_set		rset ; 
	FD_ZERO(&rset) ; 
	FD_SET(m_isSocket, &rset) ; 
	struct timeval timeout;
	timeout.tv_sec  = 0; timeout.tv_usec = 0;

	int32 n = select( m_isSocket +1 , &rset, NULL, NULL, &timeout ) ;  

	if ( n < 0 ) W_FATAL( fcOS ) ; 

        block = n > 0 ? true : false ;

	return RCOK ; 
}

w_rc_t 
C_OLAPComm_System_Generic::Write( void* ptr, int32 len, int32& pkt_write )
{
	char* szNetworkData = (char *)ptr ; 
	int32 temp = 0 ;

	pkt_write = 0 ; 
	while ( pkt_write < len ) 
	{
		if ( (temp = send( m_isSocket, szNetworkData + pkt_write, 
	       		len - pkt_write, 0)) == SOCKET_ERROR )
			return RC(BORGCOMM_UNABLE_TO_SEND) ; 
	
		pkt_write += temp ; 
	}

	return RCOK ; 
}

w_rc_t 
C_OLAPComm_System_Generic::Read( void* ptr, int32 len, int32& pkt_read )
{
	char* szNetworkData = (char *)ptr ; 
	int temp = 0 ;

	pkt_read = 0 ; 
	while (pkt_read < len)
	{
		if ( (temp = recv(m_isSocket, szNetworkData + pkt_read, 
			len - pkt_read, 0)) == SOCKET_ERROR )
			return RC(BORGCOMM_UNABLE_TO_RECEIVE) ;

		pkt_read += temp;
	}

	return RCOK ; 
}

w_rc_t
C_OLAPComm_System_Generic_Server::Bind( const u_int16 iPort ) 
{
	if ( m_bBind )
		return RC(BORGCOMM_ALREADY_BOUND) ; 

	m_isSocket = CreateNetDomainSvrSocket( iPort ) ; 
	return m_isSocket == INVALID_SOCKET ? 
			RC(BORGCOMM_UNABLE_TO_BIND) : RCOK ; 
}

w_rc_t 
C_OLAPComm_System_Generic_Server::Accept( C_OLAPComm_System_Generic_Server& isWorker )
{
	struct sockaddr_in	isClientAddr ;
	int32			iAddrLen = sizeof(isClientAddr) ; 

	memset( &isClientAddr, 0, sizeof(isClientAddr) ) ;

	SOCKET			isClientSocket ;

	if ( (isClientSocket = accept( m_isSocket, 
				(struct sockaddr *)&isClientAddr,
				&iAddrLen )) == INVALID_SOCKET )
		return RC(BORGCOMM_UNABLE_TO_ACCEPT) ;

#ifdef _WINDOWS
	closesocket(isWorker.m_isSocket) ; 
#else
        close(isWorker.m_isSocket) ;
#endif

	isWorker.m_isSocket = isClientSocket ;

	return RCOK ;
}

w_rc_t
C_OLAPComm_System_Generic_Client::Connect( const char* szHostName, 
					   const u_int16 iPort ) 
{
	if ( m_bConnected ) return RC(BORGCOMM_ALREADY_CONNECTED) ;

	m_isSocket = CreateNetDomainCltSocket( szHostName, iPort ) ; 

	return m_isSocket == INVALID_SOCKET ? 
		RC(BORGCOMM_UNABLE_TO_CONNECT) : RCOK ; 
}

