/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_system_generic.h
//
// This file defines concrete implementation interface of the transport
// layer for sockets. Specifically abstracts the data read and write 
// interface to hide the various differences. Makes the component 
// portable across various platforms and also provides a framework for use 
// with various forms of transport other than sockets
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////


#if !defined(__OLAP_COMM_SYSTEM_GENERIC_H)
#define __OLAP_COMM_SYSTEM_GENERIC_H

#if !defined(_WINDOWS)
#include <unistd.h>
#include <string.h>
#endif

#include "olap_comm_msgheader.h"
#include "olap_comm_pkt.h"
#include "olap_comm_reconciler.h"
#include "olap_comm_system.h"

class C_OLAPComm_System_Generic : public C_OLAPComm_System {

public :

	C_OLAPComm_System_Generic()
	{ } 
 
	C_OLAPComm_System_Generic( SOCKET s ) : m_isSocket(s)
	{ }

	virtual ~C_OLAPComm_System_Generic()
	{
		TearDownSocket( m_isSocket ) ; 
	}

	virtual w_rc_t WillBlockRead( bool& block ) ; 

protected :

	virtual w_rc_t Read(void* ptr, int32 len, int32& pkt_read) ; 
	virtual w_rc_t Write(void* ptr, int32 len, int32& pkt_write) ; 

	SOCKET			m_isSocket ;

} ;

class C_OLAPComm_System_Generic_Server : public C_OLAPComm_System_Generic {

public :

	C_OLAPComm_System_Generic_Server()
	{
		m_bBind = false ;
	}

	C_OLAPComm_System_Generic_Server( SOCKET s, bool bound = false ) :
		C_OLAPComm_System_Generic(s), m_bBind(bound)
	{ }

	~C_OLAPComm_System_Generic_Server() 
	{ }

	w_rc_t Bind( const u_int16 iPort ) ;
	w_rc_t Accept( C_OLAPComm_System_Generic_Server& isWorker ) ;

private :

	bool			m_bBind ;

} ;
	
class C_OLAPComm_System_Generic_Client : public C_OLAPComm_System_Generic {

public :

	C_OLAPComm_System_Generic_Client()
	{
		m_bConnected = false ;
	}

	C_OLAPComm_System_Generic_Client( SOCKET s, bool connected = false ) :
		C_OLAPComm_System_Generic(s), m_bConnected(connected)
	{ }

	~C_OLAPComm_System_Generic_Client()
	{ }

	w_rc_t Connect( const char* szHostName, const u_int16 iPort ) ;

private :

	bool			m_bConnected ;
} ;

#endif
