/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_msgheader.h
//
// This file contains the class defintion for protocol message header 
// for communications. The header defines the packet type of data being
// sent, endian type of sending machine and the length of the packet
// being sent
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////


#if !defined(__OLAP_COMM_MSGHEADER_H)
#define __OLAP_COMM_MSGHEADER_H

#if !defined(_WINDOWS)
#include <sys/types.h>
#include <netinet/in.h>
#endif

#include <iostream.h>

class C_OLAPComm_Header {

public :

	C_OLAPComm_Header()
	{
		m_iPktType = OLAPCOMM_NULL_PACKET ;
		m_iCtlInfo = u_int8(0);
		m_iLength = 0 ;
	}

	C_OLAPComm_Header( OLAPCOMM_PACKET_TYPE isPktType, 
		               OLAPCOMM_ENDIAN_TYPE isEndian, 
					   bool bMoreData,
					   int32 iLength )
	{
		Init( isPktType, isEndian, bMoreData, iLength ) ;
	}

	~C_OLAPComm_Header()
	{

	}

	OLAPCOMM_PACKET_TYPE GetPacketType()
	{
		return OLAPCOMM_PACKET_TYPE(m_iPktType) ;
	}

	OLAPCOMM_ENDIAN_TYPE GetEndianType()
	{
		return OLAPCOMM_ENDIAN_TYPE(m_iCtlInfo & ms_iEndianExtractor);
	}

	bool GetMoreData()
	{
		return m_iCtlInfo & ms_iMoreDataExtractor ? true : false ;
	}

	u_int32 GetPacketLength()
	{
		return m_iLength ;
	}

	void
	ConvertToHostOrder()
	{
		m_iPktType = ntohl(m_iPktType) ; 
		m_iCtlInfo = ntohl(m_iCtlInfo) ; 
		m_iLength = ntohl(m_iLength) ;
	}

	void
	ConvertToNetworkOrder()
	{
		m_iPktType = htonl(m_iPktType) ;
		m_iCtlInfo = htonl(m_iCtlInfo) ; 
		m_iLength = htonl(m_iLength) ;
	}

	friend ostream& operator << (ostream& os, 
				const C_OLAPComm_Header& header) ;
 
protected :

	friend class C_OLAPComm_Reconciler ;

	void Init( OLAPCOMM_PACKET_TYPE isPktType, 
		       OLAPCOMM_ENDIAN_TYPE isEndian, bool bMoreData,
		       int32 iLength )
	{
		m_iPktType = int32(isPktType) ; m_iCtlInfo = int32(isEndian) ;
		m_iLength = iLength ;

		if ( bMoreData )
			m_iCtlInfo |= ms_iMoreDataExtractor ;
	}

private :

	int32			m_iPktType ;	// Packet type
	int32			m_iCtlInfo ;	// endian, etc
	int32			m_iLength ;	// Length of the data packet

	static const int32	ms_iEndianExtractor ;
	static const int32      ms_iMoreDataExtractor ;
} ; 

#endif
