/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_pkt.h
//
// This file contains the abstract class definition of the data packets
// that exchange various data between client and server. Also consists
// of various packets definitions inherited from the abstract class.
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////


#if !defined(__OLAP_COMM_PKT_H)
#define __OLAP_COMM_PKT_H

#include "olap_comm_data.h"

class C_OLAPComm_Packet {

public :

	C_OLAPComm_Packet()
	{
		m_szNetworkData = NULL ;
		m_iNetworkDataLength = 0 ; 
		m_bMemoryOut = false ;
	}

	virtual ~C_OLAPComm_Packet()
	{

	}

	virtual OLAPCOMM_PACKET_TYPE GetPktType() = 0 ; 

	virtual bool PeekAhead() = 0 ; 

	virtual bool Linearize() = 0 ;

	virtual bool LinearizeNext() = 0 ; 

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE isEndian ) = 0 ; 

	virtual bool DelinearizeNext() = 0 ;

	char* GetNetworkData()
	{
		return m_szNetworkData ;
	}

	int32 GetNetworkDataLength()
	{
		return m_iNetworkDataLength ;
	}


protected:

	void Init( char* szData, int32 iDataLength )
	{
		m_szNetworkData = szData ;
		m_iNetworkDataLength = iDataLength ;
	}

	char*		m_szNetworkData ;
	int32		m_iNetworkDataLength ;
	bool		m_bMemoryOut ;

	friend class C_OLAPComm_Reconciler ;
} ;

class C_OLAPComm_Generic_Packet : public C_OLAPComm_Packet {

public :

	C_OLAPComm_Generic_Packet( C_OLAPComm_DataLinearizer& isData ) :
		m_isDataLinearizer(&isData) 
	{
		m_isDataDelinearizer = NULL ; 
	}

	virtual ~C_OLAPComm_Generic_Packet()
	{
		if ( !m_bMemoryOut && m_isDataDelinearizer ) 
			delete m_isDataDelinearizer ;
	} 

	virtual OLAPCOMM_PACKET_TYPE GetPktType()
	{
		return OLAPCOMM_GENERIC_PACKET ;
	}

	virtual bool PeekAhead()
	{
		return false ;
	}


	virtual bool Linearize()
	{
		m_szNetworkData = m_isDataLinearizer->GetLinearizedData() ;
		m_iNetworkDataLength = m_isDataLinearizer->GetLinearizedDataLength() ;
		return true ;
	}

	virtual bool LinearizeNext()
	{
		return false ;
	}

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE isEndian ) ;

	virtual bool DelinearizeNext()
	{
		return false ;
	}

	C_OLAPComm_DataDelinearizer* GetFreezedDataDelinearizer()
	{
		m_bMemoryOut = true ;
		return m_isDataDelinearizer ;
	}

	C_OLAPComm_DataDelinearizer* GetDataDelinearizer()
	{
		return m_isDataDelinearizer ;
	}

protected:

	C_OLAPComm_DataLinearizer*	m_isDataLinearizer ;
	C_OLAPComm_DataDelinearizer*	m_isDataDelinearizer ;

	C_OLAPComm_Generic_Packet() : m_isDataLinearizer(NULL),
		                          m_isDataDelinearizer(NULL)
	{

	}

	friend class C_OLAPComm_Reconciler ;
} ;

class C_OLAP_Page ; 

class C_OLAPComm_Stream_Packet : public C_OLAPComm_Packet {

public:

	C_OLAPComm_Stream_Packet( C_OLAP_Page& page ) : m_isPage(&page) 
	{  } 
	
	virtual ~C_OLAPComm_Stream_Packet() ;

	virtual OLAPCOMM_PACKET_TYPE GetPktType()
	{
		return OLAPCOMM_STREAM_PACKET ;
	}

	virtual bool PeekAhead()
	{
		return false ;
	}

	virtual bool Linearize() ;

	virtual bool LinearizeNext()
	{
		return false ;
	}

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE /*isEndian*/ ) ;

	virtual bool DelinearizeNext()
	{
		return false ;
	}

	C_OLAP_Page* GetFreezedPage()
	{
		m_bMemoryOut = true ;
		return m_isPage ;
	}

	C_OLAP_Page* GetPage()
	{
		return m_isPage ;
	}

protected:

	C_OLAPComm_Stream_Packet() : m_isPage(NULL)
	{  }  

	C_OLAP_Page*		m_isPage ;

	friend class C_OLAPComm_Reconciler ;
} ; 

class C_OLAPComm_Error_Packet : public C_OLAPComm_Packet {

public:

	C_OLAPComm_Error_Packet( char* szError )
	{
		if ( !szError ) return ;

		m_iNetworkDataLength = ::strlen(szError) + 1 ;
		m_szNetworkData = szError ;
	}

	virtual ~C_OLAPComm_Error_Packet()
	{
		if ( !m_bMemoryOut ) delete [] m_szNetworkData ;
	}

	virtual OLAPCOMM_PACKET_TYPE GetPktType()
	{
		return OLAPCOMM_ERROR_PACKET ;
	}

	virtual bool PeekAhead()
	{
		return false ;
	}

	virtual bool Linearize() 
	{
		return true ;
	}

	virtual bool LinearizeNext()
	{
		return false ;
	}

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE /*isEndian*/ )
	{
		return true ;
	}

	virtual bool DelinearizeNext()
	{
		return false ;
	}

	char* GetFreezedErrorMessage()
	{
		m_bMemoryOut = true ;
		return m_szNetworkData ;
	}

	char* GetErrorMessage()
	{
		return m_szNetworkData ;
	}

protected:

	C_OLAPComm_Error_Packet()
	{

	}

	friend class C_OLAPComm_Reconciler ;
} ;

class C_OLAPComm_Exit_Packet : public C_OLAPComm_Packet {

public:

	C_OLAPComm_Exit_Packet()
	{ }

	virtual ~C_OLAPComm_Exit_Packet()
	{ }

	virtual OLAPCOMM_PACKET_TYPE GetPktType()
	{
		return OLAPCOMM_EXIT_PACKET ;
	}

	virtual bool PeekAhead()
	{
		return false ;
	}

	virtual bool Linearize() 
	{
		return true ;
	}

	virtual bool LinearizeNext()
	{
		return false ;
	}

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE /*isEndian*/ )
	{
		return true ;
	}

	virtual bool DelinearizeNext()
	{
		return false ;
	}

protected:

	friend class C_OLAPComm_Reconciler ;
} ;

class C_OLAPComm_Empty_Packet : public C_OLAPComm_Packet {

public:

	C_OLAPComm_Empty_Packet()
	{ }

	virtual ~C_OLAPComm_Empty_Packet()
	{ }

	virtual OLAPCOMM_PACKET_TYPE GetPktType()
	{
		return OLAPCOMM_EMPTY_PACKET ;
	}

	virtual bool PeekAhead()
	{
		return false ;
	}

	virtual bool Linearize() 
	{
		return true ;
	}

	virtual bool LinearizeNext()
	{
		return false ;
	}

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE /*isEndian*/ )
	{
		return true ;
	}

	virtual bool DelinearizeNext()
	{
		return false ;
	}

protected:

	friend class C_OLAPComm_Reconciler ;
} ; 

class C_OLAPComm_Shutdown_Packet : public C_OLAPComm_Packet {

public:

	C_OLAPComm_Shutdown_Packet()
	{ }

	virtual ~C_OLAPComm_Shutdown_Packet()
	{ }

	virtual OLAPCOMM_PACKET_TYPE GetPktType()
	{
		return OLAPCOMM_SHUTDOWN_PACKET ;
	}

	virtual bool PeekAhead()
	{
		return false ;
	}

	virtual bool Linearize() 
	{
		return true ;
	}

	virtual bool LinearizeNext()
	{
		return false ;
	}

	virtual bool Delinearize( OLAPCOMM_ENDIAN_TYPE /*isEndian*/ )
	{
		return true ;
	}

	virtual bool DelinearizeNext()
	{
		return false ;
	}

protected:

	friend class C_OLAPComm_Reconciler ;
} ; 

#endif

