/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_reconciler.cpp
//
// This file contains the method definitions related with the
// C_OLAPComm_Reconciler. For evey new data packet added, one must
// add an instance create statement in the CreatePacket. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////


#if !defined(_WINDOWS)
#include <sys/types.h>
#include <netinet/in.h>
#endif
#include "olap_types.h"
#include "olap_comm_types.h"
#include "olap_comm_reconciler.h"

void
C_OLAPComm_Reconciler::GenerateMsgHeader( C_OLAPComm_Packet& isPacket, 
										  C_OLAPComm_Header& isHeader )
{
	isHeader.Init( isPacket.GetPktType(), gc_isMyEndian,
				   isPacket.PeekAhead() ? true : false, 
				   isPacket.GetNetworkDataLength() ) ; 
}

void 
C_OLAPComm_Reconciler::GenerateNetworkData( C_OLAPComm_Packet& isPacket,
		                         char*& szNetworkData,
							     int32& iNetworkDataLength )
{
	szNetworkData = isPacket.GetNetworkData() ;

	iNetworkDataLength = isPacket.GetNetworkDataLength() ;
}

void 
C_OLAPComm_Reconciler::ExtractNetworkData( C_OLAPComm_Packet& isPacket,
							               char*   szNetworkData,
							               C_OLAPComm_Header& isHeader )
{
	isPacket.Init( szNetworkData, isHeader.GetPacketLength() ) ;
}

C_OLAPComm_Packet*
C_OLAPComm_Reconciler::CreatePacket( C_OLAPComm_Header& isHeader )
{
	C_OLAPComm_Packet*	isPacket = NULL ;

	switch( isHeader.GetPacketType() )
	{
		case OLAPCOMM_GENERIC_PACKET :
			isPacket = new C_OLAPComm_Generic_Packet ;
			break ;

		case OLAPCOMM_STREAM_PACKET :
			isPacket = new C_OLAPComm_Stream_Packet ;
			break ;

		case OLAPCOMM_ERROR_PACKET :
			isPacket = new C_OLAPComm_Error_Packet ;
			break ;

		case OLAPCOMM_EXIT_PACKET :
			isPacket = new C_OLAPComm_Exit_Packet ;
			break ;

		case OLAPCOMM_EMPTY_PACKET :
			isPacket = new C_OLAPComm_Empty_Packet ;
			break ;

		case OLAPCOMM_SHUTDOWN_PACKET :
			isPacket = new C_OLAPComm_Shutdown_Packet ;
			break ;

		default :
			break ;
	}

	return isPacket ;
}
