/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_server_test.cpp
//
// This file contains code for testing the communication component.
// Specifically, a server instances is created and listens on an arbitrary
// port. Server echoes the data back received. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////


#include <iostream.h>
#include <strstream.h>

#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "olap_types.h"
#include "olap_comm_types.h"
#include "olap_comm_system_generic.h"

int main(int argc, char* argv[])
{
	int i;

	if ( argc != 2 )
	{
		cerr << argv[0] << " : port-number " << endl ;
		exit(1) ;
	}

	ostrstream out;
	

	C_OLAPComm_System::Initialize() ;

	C_OLAPComm_System_Generic_Server	isServer ; 
	C_OLAPComm_System_Generic_Server	isWorker ;
	C_OLAPComm_Reconciler			isReconciler ;

	if ( isServer.Bind(atoi(argv[1])) < 0 )
	{
		cerr << "Error in connecting..." << endl ;
		exit(1) ;
	}

	C_OLAPComm_Packet* pPacket = NULL ;

	while(1)
        {
		if ( isServer.Accept(isWorker) < 0 )
		{
	    		cerr << "Error in accepting..." << endl ;
	    		exit(1) ;
		}

                char *sqlstmt;

        	if ( isWorker.ReceiveData( isReconciler, pPacket ) < 0 )
        	{
        		cerr << "Unable to recv data..." << endl ;
        		exit(1) ;
        	}
        
        	C_OLAPComm_DataDelinearizer  *qData;
                C_OLAPComm_Generic_Packet    *recvPkt ; 
        	int32 len;
        	
                recvPkt = (C_OLAPComm_Generic_Packet *)pPacket ; 

        	qData = recvPkt->GetDataDelinearizer();
        	qData->Unpack(len);
                cout << len << endl;
        	sqlstmt = new char[len];
        	qData->Unpack(sqlstmt, len);
        
        	cout << "Received query : " << endl;
        	cout << sqlstmt << endl;
        	
        	out.seekp(0);
                for (i=0; i<1000; i++)
			out << "junk" << i << " ";
        
        	C_OLAPComm_DataLinearizer		isData ;
        
        	isData.Pack( int32(out.pcount()) );
        	isData.Pack( out.str() );
        
        	C_OLAPComm_Generic_Packet isPacket(isData) ;
        
        	cout << "Sending Results " << endl;
		cout << out.str() << endl;
        	
        	if ( isWorker.SendData( isReconciler, isPacket ) < 0 )
        	{
        		cerr << "Unable to send data..." << endl ;
        		exit(1) ;
        	}
        	
        	out.freeze(false);
                delete recvPkt;
        }
	C_OLAPComm_System::Shutdown() ;

	exit(0) ;
}
