/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_system.cpp
//
// This file contains the method definitions of the transport layer
// of the communication component. It contains routines for shipping
// data, receiving data, sending protocol header, receiving protocol
// header and other miscellaneous utilies for startup, shutdown,
// socket creation and destruction.
// 
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_types.h"
#include "olap_comm_types.h"
#include "olap_comm_system.h"

int32 C_OLAPComm_System::DNS_RETRIES = 4 ;

/*
struct hostent* C_OLAPComm_System::GetHostByName( const char* szHostName ) ;

w_rc_t C_OLAPComm_System::Initialize() ;
w_rc_t C_OLAPComm_System::Shutdown() ;

SOCKET C_OLAPComm_System::CreateNetDomainSvrSocket( const u_int16 port ) ;
SOCKET C_OLAPComm_System::CreateNetDomainCltSocket( const char* hostname, 
						    const u_int16 port ) ;
SOCKET C_OLAPComm_System::AcceptNetDomainSocket( SOCKET sock ) ;

SOCKET C_OLAPComm_System::CreateMachineDomainSvrSocket( const char* name ) ;
SOCKET C_OLAPComm_System::CreateMachineDomainCltSocket( const char* name ) ;
SOCKET C_OLAPComm_System::AcceptMachineDomainSocket( SOCKET sock ) ;

void C_OLAPComm_System::TearDownSocket( SOCKET sock ) ; 
*/

struct hostent*
C_OLAPComm_System::GetHostByName( const char* szHostName )
{
	if ( !szHostName ) return NULL ;

	for ( int i = 0 ; i < DNS_RETRIES ; i++ )
	{
		struct hostent* isHostEnt = gethostbyname( szHostName ) ;
		if ( isHostEnt ) return isHostEnt ;
	}

	return NULL ;
}

w_rc_t
C_OLAPComm_System::Initialize()
{
#ifdef _WINDOWS
	WORD	wVersionRequested;
	WSADATA wsaData;
	int		iError;
 
	wVersionRequested = MAKEWORD( 2, 0 );
 
	iError = WSAStartup( wVersionRequested, &wsaData );
	if ( iError != 0 ) {
		return RC(BORGCOMM_UNABLE_TO_INITIALIZE) ;
	}
#endif

	return RCOK ;
}

w_rc_t
C_OLAPComm_System::Shutdown()
{
#ifdef _WINDOWS
	if ( WSACleanup () == SOCKET_ERROR )
		return RC(BORGCOMM_UNABLE_TO_SHUTDOWN) ;
#endif
	return RCOK ;
}

SOCKET
C_OLAPComm_System::CreateNetDomainSvrSocket( const u_int16 iPort )
{
	SOCKET isSocket = socket( AF_INET, SOCK_STREAM, 0 ) ;
	if ( isSocket == INVALID_SOCKET ) return -1 ;

	struct sockaddr_in		isServAddr ;
	memset( &isServAddr, 0, sizeof(isServAddr) ) ;

	isServAddr.sin_family = AF_INET ;
	isServAddr.sin_port = htons( iPort ) ;
	isServAddr.sin_addr.s_addr = htonl(INADDR_ANY) ;

	if ( bind( isSocket, (struct sockaddr *)&isServAddr,
		                 sizeof(isServAddr) ) == SOCKET_ERROR )
		return INVALID_SOCKET ;

	if ( listen( isSocket, SOMAXCONN ) == SOCKET_ERROR )
		return INVALID_SOCKET ;

	return isSocket ; 
}

SOCKET
C_OLAPComm_System::CreateNetDomainCltSocket( 
				const char* szHostName, 
				const u_int16 iPort )
{
	SOCKET isSocket = socket( AF_INET, SOCK_STREAM, 0 ) ;
	if ( isSocket == INVALID_SOCKET ) return -1 ;

	struct hostent*		isServHost ;

	if ( (isServHost = GetHostByName(szHostName)) != NULL )
	{
		struct sockaddr_in		isServAddr ;

		memset( &isServAddr, 0, sizeof(isServAddr) ) ;

		isServAddr.sin_family = AF_INET ;
		isServAddr.sin_port = htons( iPort ) ;

		struct in_addr*  isInAddr = (struct in_addr *) *(isServHost->h_addr_list) ;

		isServAddr.sin_addr.s_addr = inet_addr(inet_ntoa(*isInAddr)) ;

		if ( connect( isSocket, (struct sockaddr *)&isServAddr, 
				                       sizeof(isServAddr) ) == SOCKET_ERROR )
			return INVALID_SOCKET ;
	}

	return isSocket ; 
}

SOCKET 
C_OLAPComm_System::AcceptNetDomainSocket( SOCKET sock ) 
{
	struct sockaddr_in      isCltAddr ;

	int32 iCltAddrLen = sizeof(isCltAddr) ;
	memset( &isCltAddr, 0, sizeof(isCltAddr) ) ;

        SOCKET isAcceptSocket = 
		accept( sock, (struct sockaddr *)&isCltAddr, &iCltAddrLen ) ;

	if ( isAcceptSocket == SOCKET_ERROR )
		return INVALID_SOCKET ;

	return isAcceptSocket ;
}

SOCKET
C_OLAPComm_System::CreateMachineDomainSvrSocket( const char* name )
{
	SOCKET isSocket = socket( AF_UNIX, SOCK_STREAM, 0 ) ;
	if ( isSocket == INVALID_SOCKET ) return -1 ;

	// Remove the file or socket if one exists
        unlink(name) ;

	struct  sockaddr_un     isServAddr ;

	// Bind the socket
	memset( &isServAddr, 0, sizeof(isServAddr)) ;

	isServAddr.sun_family = AF_UNIX ;
	strcpy(isServAddr.sun_path, name) ;

	if ( bind( isSocket, (struct sockaddr *)&isServAddr,
		                 sizeof(isServAddr) ) == SOCKET_ERROR )
		return INVALID_SOCKET ;

	if ( listen( isSocket, SOMAXCONN ) == SOCKET_ERROR )
		return INVALID_SOCKET ;

	return isSocket ; 
}

SOCKET
C_OLAPComm_System::CreateMachineDomainCltSocket( const char* name )
{
	SOCKET isSocket = socket( AF_UNIX, SOCK_STREAM, 0 ) ;
	if ( isSocket == INVALID_SOCKET ) return -1 ;

	struct sockaddr_un		isServAddr ;

	memset( &isServAddr, 0, sizeof(isServAddr) ) ;
	isServAddr.sun_family = AF_UNIX ;
	strcpy(isServAddr.sun_path, name) ;

	if ( connect( isSocket, (struct sockaddr *)&isServAddr,
                                  sizeof(isServAddr) ) == SOCKET_ERROR )
	{
		TearDownSocket( isSocket ) ; 
		return INVALID_SOCKET ;
	}

	return isSocket ; 
} 

SOCKET 
C_OLAPComm_System::AcceptMachineDomainSocket( SOCKET sock ) 
{
	struct sockaddr_un      isCltAddr ;

	int32 iCltAddrLen = sizeof(isCltAddr) ;
	memset( &isCltAddr, 0, sizeof(isCltAddr) ) ;

        SOCKET isAcceptSocket = 
		accept( sock, (struct sockaddr *)&isCltAddr, &iCltAddrLen ) ;

	if ( isAcceptSocket == SOCKET_ERROR )
		return INVALID_SOCKET ;

	return isAcceptSocket ;
}

void
C_OLAPComm_System::TearDownSocket( SOCKET sock ) 
{
#ifdef _WINDOWS
	closesocket( sock ) ; 
#else
	close( sock ) ; 
#endif
}

w_rc_t 
C_OLAPComm_System::SendProtocolHeader( C_OLAPComm_Header& isHeader )
{
	isHeader.ConvertToNetworkOrder() ;
	
	int nwritten ; 

	W_DO( Write(&isHeader, sizeof(isHeader), nwritten) ) ;

	return RCOK ; 
}

w_rc_t 
C_OLAPComm_System::ReceiveProtocolHeader( C_OLAPComm_Header& isHeader )
{
	int nread ;

	W_DO( Read(&isHeader, sizeof(isHeader), nread) ) ;

	isHeader.ConvertToHostOrder() ;

	return RCOK ; 
}

w_rc_t
C_OLAPComm_System::SendData( C_OLAPComm_Reconciler& isReconciler, 
				 C_OLAPComm_Packet& isPacket )
{
	C_OLAPComm_Header	isHeader ;
	char*			szNetworkData ;
	int32			iNetworkDataLength ;

	while (1)
	{
		isPacket.Linearize() ;

		isReconciler.GenerateMsgHeader( isPacket, isHeader ) ;

		isReconciler.GenerateNetworkData( isPacket, szNetworkData,
				              iNetworkDataLength ) ;

		W_DO( SendProtocolHeader( isHeader ) ) ;  

		if ( iNetworkDataLength > 0 )
		{
			int32 nwritten ; 

			W_DO( Write(szNetworkData, iNetworkDataLength, 
						nwritten) );
		}

		if ( !isPacket.PeekAhead() ) break ;
			
		isPacket.LinearizeNext() ; 
	}

	return RCOK ; 
}

w_rc_t
C_OLAPComm_System::ReceiveData( C_OLAPComm_Reconciler& isReconciler, 
				    C_OLAPComm_Packet*& isPacket )
{
	C_OLAPComm_Header	isHeader ;
	char*			szNetworkData ;

	isPacket = NULL ;

	while (1)
	{
		W_DO( ReceiveProtocolHeader( isHeader ) ) ;

		isPacket = isPacket ? isPacket : 
			  isReconciler.CreatePacket( isHeader ) ; 

		if ( !isPacket ) W_FATAL(BORGCOMM_UNABLE_TO_CREATE_PACKET) ; 

		if ( isHeader.GetPacketLength() )
		{
			szNetworkData = new char[isHeader.GetPacketLength()] ;

			if ( !szNetworkData ) 
				W_FATAL(BORG_OUTOFMEMORY) ; ; 
			
			int nread ; 
			W_DO( Read(szNetworkData,isHeader.GetPacketLength(),
					 nread) ) ; 

			isReconciler.ExtractNetworkData( *isPacket, 
						szNetworkData,
						isHeader ) ;
			isPacket->Delinearize( isHeader.GetEndianType() ) ; 

		}

		isPacket->DelinearizeNext() ; 

		if ( !isHeader.GetMoreData() ) 
			break ;
	}

	return RCOK ; 
}
