/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_system.h
//
// This file defines the interface for the transport layer of the 
// communication component. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////


#if !defined(__OLAP_COMM_SYSTEM_H)
#define	__OLAP_COMM_SYSTEM_H

#include "w.h"
#include "olap_comm_reconciler.h"
#include "olap_comm_pkt.h"

class C_OLAPComm_System {

public :

	w_rc_t SendData( C_OLAPComm_Reconciler& isReconciler, 
		            C_OLAPComm_Packet& isPacket ) ; 

	w_rc_t ReceiveData( C_OLAPComm_Reconciler& isReconciler, 
		               C_OLAPComm_Packet*& isPacket ) ; 
	
	virtual w_rc_t WillBlockRead( bool& block ) = 0 ; 

	static w_rc_t Initialize() ;
	static w_rc_t Shutdown() ;

	static int32 DNS_RETRIES ;

	static struct hostent* GetHostByName( const char* szHostName ) ;

	static SOCKET CreateNetDomainSvrSocket( const u_int16 port ) ; 
	static SOCKET CreateNetDomainCltSocket( const char* hostname, 
					const u_int16 port ) ; 
	static SOCKET AcceptNetDomainSocket( SOCKET sock ) ; 

	static SOCKET CreateMachineDomainSvrSocket( const char* name ) ; 
	static SOCKET CreateMachineDomainCltSocket( const char* name ) ; 
	static SOCKET AcceptMachineDomainSocket( SOCKET sock ) ; 

	static void TearDownSocket( SOCKET sock ) ; 

protected :

	w_rc_t SendProtocolHeader( C_OLAPComm_Header& isHeader ) ;
	w_rc_t ReceiveProtocolHeader( C_OLAPComm_Header& isHeader ) ;

	virtual w_rc_t Read( void* ptr, int32 len, int32& pkt_read ) = 0 ; 
	virtual w_rc_t Write( void* ptr, int32 len, int32& pkt_write ) = 0 ; 
} ;

#endif
