/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_system_shore.h
//
// This file contains interface definitions of the transport layer for
// shore sfiles. It is inherited from the comm system so that it supports
// shipping and receiving operations. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_COMM_SYSTEM_SHORE_H)
#define __OLAP_COMM_SYSTEM_SHORE_H

#if !defined(_WINDOWS)
#include <unistd.h>
#include <string.h>
#endif

#include "olap_server_globals.h"
#include "sfile.h"

#include "olap_comm_msgheader.h"
#include "olap_comm_pkt.h"
#include "olap_comm_reconciler.h"
#include "olap_comm_system.h"

class C_OLAPComm_System_Shore : public C_OLAPComm_System {

public :

	C_OLAPComm_System_Shore() ;

	C_OLAPComm_System_Shore(const SOCKET fd) ;

	C_OLAPComm_System_Shore(sfile_t* sd) 
	{
		m_isSocket = sd ; m_bAllocated = false ; 
	}

	virtual ~C_OLAPComm_System_Shore()
	{ 
		if ( m_bAllocated ) delete m_isSocket ; 
	}

	virtual w_rc_t WillBlockRead( bool& block ) ; 

protected :

	virtual w_rc_t Read(void* ptr, int32 len, int32& pkt_read) ;
	virtual w_rc_t Write(void* ptr, int32 len, int32& pkt_write) ;

	sfile_t*			m_isSocket ;
	bool				m_bAllocated ; 

} ;

class C_OLAPComm_System_Shore_Server : public C_OLAPComm_System_Shore {

public :

	C_OLAPComm_System_Shore_Server()
	{
		m_bBind = false ;
	}

	C_OLAPComm_System_Shore_Server(sfile_t* fd, bool bind) 
				: C_OLAPComm_System_Shore(fd) 
	{
		m_bBind = bind ;
	}

	C_OLAPComm_System_Shore_Server(const SOCKET fd, bool bind) 
				: C_OLAPComm_System_Shore(fd) 
	{
		m_bBind = bind ;
	}

	virtual ~C_OLAPComm_System_Shore_Server() 
	{ }

	w_rc_t Bind( const u_int16 iPort ) ;
	w_rc_t Accept( C_OLAPComm_System_Shore_Server& isWorker ) ;

private :

	bool			m_bBind ;

} ;
	
class C_OLAPComm_System_Shore_Client : public C_OLAPComm_System_Shore {

public :

	C_OLAPComm_System_Shore_Client()
	{
		m_bConnected = false ;
	}

	C_OLAPComm_System_Shore_Client(const SOCKET fd, bool conn) 
				: C_OLAPComm_System_Shore(fd) 
	{
		m_bConnected = conn ;
	}

	C_OLAPComm_System_Shore_Client(sfile_t* fd, bool conn) 
				: C_OLAPComm_System_Shore(fd) 
	{
		m_bConnected = conn ;
	}

	virtual ~C_OLAPComm_System_Shore_Client()
	{ }

	w_rc_t Connect( const char* szHostName, const u_int16 iPort ) ;

private :

	bool			m_bConnected ;
} ;

#endif
