/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_msg.h
//
// This file contains the class definition of the data message types
// send out from the olap server for debugging purposes.
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_MSG_H)
#define	__OLAP_MSG_H

#include <string>

#include "olap_types.h"
#include "olap_client_id.h"
#include "w.h"

#include "olap_comm_pkt.h"

class C_OLAPMsg : public C_OLAPComm_Data
{
public:

	static int32 OLAP_MSG_ALL_CLIENTS ;

	enum OLAP_Msg_Class
	{
		OLAP_MSG_ERROR = 1,
		OLAP_MSG_VERBOSE,
		OLAP_MSG_EXIT,
		OLAP_MSG_SERVER_EXIT
	} ;

	enum OLAP_Msg_SubClass 
	{
		OLAP_MSG_SERVER_FATAL	= 1,
		OLAP_MSG_CLIENT_FATAL,
		OLAP_MSG_SERVER_WARNING,
		OLAP_MSG_CLIENT_WARNING,
		OLAP_MSG_SERVER_VERBOSE
	} ;

	C_OLAPMsg()
	{ } 

	C_OLAPMsg(
		int32			id, 
		OLAP_Msg_Class		msg_class,
		OLAP_Msg_SubClass	msg_subclass,	
		int32 			msg_code,
		const string&		err_string
		) : m_isClientId(id), 
			m_isMsgClass(msg_class), 
			m_isMsgSubClass(msg_subclass),
			m_iMsgCode(msg_code), 
			m_isErrString(err_string)
	{ } 

	~C_OLAPMsg()
	{ } 

	C_OLAPMsg& operator = (const C_OLAPMsg& msg_item) ;

	C_OLAPClient_Id& GetClientId()
	{
		return	m_isClientId ;
	}

	OLAP_Msg_Class GetMsgClass()
	{
		return m_isMsgClass ; 
	}

	OLAP_Msg_SubClass GetMsgSubClass()
	{
		return m_isMsgSubClass ; 
	}

	string& GetErrorString()
	{
		return m_isErrString ; 
	}

	virtual void Linearize( C_OLAPComm_DataLinearizer &dLin ) ; 
        virtual void DeLinearize( C_OLAPComm_DataDelinearizer &deLin ) ;

	friend ostream& operator << (ostream& os, const C_OLAPMsg& msg_item) ;

protected:

	C_OLAPClient_Id		m_isClientId ;
	OLAP_Msg_Class		m_isMsgClass ; 
	OLAP_Msg_SubClass	m_isMsgSubClass ; 
	int32			m_iMsgCode ;
	string			m_isErrString ; 	
	w_link_t		m_isLink ; 

	friend class C_OLAPMsg_Queue ; 
} ;

#endif
