/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_msg_queue.cpp
//
// This file consists of the implementation of various types of message
// queues. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_macros.h"
#include "olap_msg_queue.h"

#include "olap_thread.h"

C_OLAPMsg_Queue::C_OLAPMsg_Queue() : 
		m_isFreeList(offsetof(C_OLAPMsg, m_isLink)),
		m_isMsgQueue(offsetof(C_OLAPMsg, m_isLink)),
		m_isQueueCond("error queue condition")  
{
	for ( int i = 0 ; i < max_msg_items ; i++ )
	{
		C_OLAPMsg* isMsgItem = new C_OLAPMsg ; 
		if ( !isMsgItem ) W_FATAL( BORG_OUTOFMEMORY ) ; 
 
		m_isFreeList.append( isMsgItem ) ; 
	}
}

C_OLAPMsg_Queue::~C_OLAPMsg_Queue()
{
        C_OLAPMsg*                      isMsgItem = NULL ;

        while ( (isMsgItem = m_isFreeList.pop()) )
		delete isMsgItem ; 

        while ( (isMsgItem = m_isMsgQueue.pop()) )
		delete isMsgItem ; 
}

w_rc_t C_OLAPMsg_Queue::enqueue( const C_OLAPMsg& msg_item )
{
	W_DO( m_isQueueMutex.acquire() ) ; 

	C_OLAPMsg*	msg = NULL ;

	while ( (msg = m_isFreeList.pop()) == NULL )
		W_DO( m_isQueueCond.wait( m_isQueueMutex ) ) ;

	*msg = msg_item ; 
	m_isMsgQueue.append( msg ) ; 

	m_isQueueMutex.release() ; 
	m_isQueueCond.signal() ; 

	return RCOK ; 
}

w_rc_t C_OLAPMsg_Queue::dequeue( C_OLAPMsg& msg_item )
{
	W_DO( m_isQueueMutex.acquire() ) ; 

	C_OLAPMsg*	msg = NULL ;

	while ( (msg = m_isMsgQueue.pop()) == NULL )
		W_DO( m_isQueueCond.wait( m_isQueueMutex ) ) ;

	msg_item = *msg ; 
	m_isFreeList.append( msg ) ; 

	m_isQueueMutex.release() ; 
	m_isQueueCond.signal() ; 

	return RCOK ; 
}

w_rc_t C_OLAPMsg_Client_Queues::add_queue( 
			const queue_id_t id , 
			C_OLAPMsg_Queue* new_queue )
{
	new_queue->m_isQueueId = id ;  
	W_DO( m_isMutex.acquire() ) ; 
	m_isQueueList.push( new_queue ) ; 
	m_isMutex.release() ; 

	return RCOK ; 
}

w_rc_t C_OLAPMsg_Client_Queues::distribute( C_OLAPMsg& msg_item )
{
	W_DO( m_isMutex.acquire() ); 
	w_hash_i<C_OLAPMsg_Queue, queue_id_t> 	iter(m_isQueueList) ; 
	C_OLAPMsg_Queue*			queue = NULL ; 

	while ( (queue = iter.next()) ) 
	{
		if ( queue->GetInterestedClient() == msg_item.GetClientId() ||
                     msg_item.GetClientId().id() == C_OLAPMsg::OLAP_MSG_ALL_CLIENTS )
		{
			w_rc_t rc = queue->enqueue( msg_item ) ; 
	 
			if ( rc )
			{
				m_isMutex.release() ;
				return rc ; 
			}
		}
	}

	m_isMutex.release() ; 

	return RCOK ; 
}

w_rc_t C_OLAPMsg_Client_Queues::delete_queue( 
			const queue_id_t id, 
			C_OLAPMsg_Queue* queue )
{
	C_OLAPMsg	msg_item ;

	queue->m_isQueueId = id ; 
	while ( m_isMutex.is_locked() )
		W_DO( queue->dequeue( msg_item ) ) ; 	

	W_DO( m_isMutex.acquire() ) ; 
	m_isQueueList.remove(id) ; 
	m_isMutex.release() ; 

	return RCOK ;
}


#ifdef __GNUG__
template class w_hash_t<C_OLAPMsg_Queue, queue_id_t>;
template class w_hash_i<C_OLAPMsg_Queue, queue_id_t>;
template class w_list_t<C_OLAPMsg_Queue>;
template class w_list_i<C_OLAPMsg_Queue>;
#endif
