/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_msg_queue.h
//
// This file consists of definitions of various message queues defined
// used by the messaging sub system. The messages are queued in a global
// queue and disseminated by msg router thread to the interested clients 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_MSG_QUEUE_H)
#define	__OLAP_MSG_QUEUE_H

#include "olap_types.h"
#include "olap_server_globals.h"
#include "olap_client_id.h"
#include "olap_msg.h"

#include "olap_comm_pkt.h"

typedef	smthread_t::id_t	queue_id_t ; 

class C_OLAPMsg_Queue
{
public:

	enum {
		max_msg_items = 128
	} ; 

	C_OLAPMsg_Queue() ; 

	~C_OLAPMsg_Queue() ; 

	w_rc_t enqueue( const C_OLAPMsg& msg_item ) ;
	w_rc_t dequeue( C_OLAPMsg& msg_item ) ;  

	C_OLAPClient_Id& GetInterestedClient()
	{
		return m_InterestedClnt ;
	}

protected:

	w_list_t<C_OLAPMsg> 	m_isFreeList ; 
	w_list_t<C_OLAPMsg>	m_isMsgQueue ;	

	C_OLAPClient_Id		m_InterestedClnt ;
	queue_id_t		m_isQueueId ; 

	scond_t			m_isQueueCond ; 
	smutex_t		m_isQueueMutex ; 	
	w_link_t		m_isLink ; 

	friend class C_OLAPMsg_Client_Queues ; 

private:
	
} ;

class C_OLAPMsg_Client_Queues
{
public:
		
	enum {
		max_queues =  32 
	} ; 

	C_OLAPMsg_Client_Queues() : 
		m_isQueueList( max_queues, 
			offsetof(C_OLAPMsg_Queue, m_isQueueId), 
			offsetof(C_OLAPMsg_Queue, m_isLink) )
	{ }

	~C_OLAPMsg_Client_Queues()
	{ } 

	w_rc_t add_queue(
			const queue_id_t id, 
			C_OLAPMsg_Queue* new_queue
		); 

	w_rc_t distribute(C_OLAPMsg& msg_item) ; 
	w_rc_t delete_queue(
			const queue_id_t id,
			C_OLAPMsg_Queue* old_queue
		) ; 

protected:

	smutex_t					m_isMutex ; 	
	w_hash_t<C_OLAPMsg_Queue, queue_id_t>		m_isQueueList ; 
} ; 

extern C_OLAPMsg_Queue			g_MsgQueue ; 
extern C_OLAPMsg_Client_Queues		g_MsgClientQueues ;

#endif
