/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_msg_thread.cpp
//
// This file consists of the implementation of various types of olap
// messaging threads.  
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_globals.h"
#include "olap_msg.h"
#include "olap_msg_thread.h"
#include "olap_client_thread.h"

#include "olap_thread.h"
#include "w_dynarray.h"

extern	C_OLAP_Clients		g_Clients ;
C_OLAPMsg_Queue			g_MsgQueue ; 
C_OLAPMsg_Client_Queues		g_MsgClientQueues ; 

C_OLAPMsg_Router_Thread*	g_RouterThread = NULL ; 
C_OLAPMsg_Inquiry_Thread*	g_MsgInquiryThread = NULL ; 

extern smutex_t			g_startMutex ;
extern scond_t			g_startCond ;

void C_OLAPMsg_Router_Thread::run()
{
	while (1)
	{
		C_OLAPMsg	msg_item ; 

		// Get the next item from the global queue
		W_COERCE( g_MsgQueue.dequeue( msg_item ) ) ; 

		// Distribute the items to interested clients
		W_COERCE( g_MsgClientQueues.distribute( msg_item ) ) ; 

		if ( msg_item.GetMsgClass() == C_OLAPMsg::OLAP_MSG_SERVER_EXIT )
			break ; 

		yield() ; 
	}
}

void C_OLAPMsg_Service_Thread::run()
{
	w_rc_t	rc ; 

	cout << "Starting msg routing service..." ; 
	// Now start the msg router thread
	g_RouterThread = new C_OLAPMsg_Router_Thread ;
	if ( !g_RouterThread ) W_FATAL(BORG_OUTOFMEMORY) ; 
	
	W_COERCE( g_RouterThread->fork() ) ; 		
	cout << "Done" << endl ; 

	cout << "Starting msg inquiry service..." ; 
	// Now start the msg inquiry thread
	g_MsgInquiryThread = 
			new C_OLAPMsg_Inquiry_Thread(BORG_Msg_Inquiry_Sock) ;
	if ( !g_MsgInquiryThread ) W_FATAL(BORG_OUTOFMEMORY) ; 
	
	W_COERCE( g_MsgInquiryThread->fork() ) ; 		
	cout << "Done" << endl ; 

	g_startCond.signal() ; 
	while (1)
	{
		sfile_t*	msg_client_fhdl = new sfile_t ; 
		if ( !msg_client_fhdl ) W_FATAL(BORG_OUTOFMEMORY) ; 
	
		struct sockaddr_in	msg_clt_addr ;
        	int msg_clt_addr_len = sizeof(msg_clt_addr) ;

		w_rc_t rc = m_isMsgListener->accept( 
				(struct sockaddr *)&msg_clt_addr,
				&msg_clt_addr_len, *msg_client_fhdl )  ; 

		if ( rc )
                {
                        delete msg_client_fhdl ; break ;
		}
			
		// now fork of msg thread to serve the requests
		C_OLAPMsg_Thread*	msg_thread = 
			new C_OLAPMsg_Thread( msg_client_fhdl ) ;
		if ( !msg_thread ) W_FATAL(BORG_OUTOFMEMORY) ; 

		W_COERCE( msg_thread->fork() ) ; 		
	}
}

C_OLAPMsg_Thread::~C_OLAPMsg_Thread() 
{
	W_COERCE( m_isSock->close() ) ; 
	delete m_isSock ; 
}

void C_OLAPMsg_Thread::run()
{
	C_OLAPComm_Reconciler 		isReconciler ;

	// Get the client interested from the msg client
	C_OLAPComm_DataDelinearizer*	qData = NULL ;
	C_OLAPComm_Packet*		pPacket = NULL ;
	C_OLAPComm_Generic_Packet*	recvPkt = NULL ;

	w_rc_t rc = m_isMsgClientServer.ReceiveData(isReconciler, pPacket)  ; 
	if ( rc ) return ;

	recvPkt = (C_OLAPComm_Generic_Packet *)pPacket ; 
	qData = recvPkt->GetDataDelinearizer();
	m_isMyQueue.GetInterestedClient().DeLinearize( *qData ) ; 

	// Add the queue in global set of queues
	W_COERCE(g_MsgClientQueues.add_queue(ThreadId(), &m_isMyQueue)) ; 

	while (1)
	{
		C_OLAPMsg	msg_item ; 

		W_COERCE( m_isMyQueue.dequeue( msg_item ) ) ; 

		if ( msg_item.GetMsgClass() == C_OLAPMsg::OLAP_MSG_EXIT ||
		     msg_item.GetMsgClass() == C_OLAPMsg::OLAP_MSG_SERVER_EXIT )
			break ; 
		
		// Send the msg item to the client
		C_OLAPComm_DataLinearizer	isData ;
		msg_item.Linearize( isData ) ; 

		C_OLAPComm_Generic_Packet	isPacket(isData) ; 
		if ( m_isMsgClientServer.SendData( isReconciler, 
					isPacket ) != RCOK )
			break ; 
	}

	W_COERCE( g_MsgClientQueues.delete_queue( ThreadId(), &m_isMyQueue ) ); 

	C_OLAPComm_Exit_Packet	exitPkt ; 
	rc = m_isMsgClientServer.SendData( isReconciler, exitPkt ) ;

	cerr << lockp << "Msg client thread done" << unlockp << endl ; 
}

void C_OLAPMsg_Inquiry_Thread::run()
{
	C_OLAPComm_Reconciler	isReconciler ; 

	while (1)
	{
		sfile_t		msg_client_fhdl ; 

		struct sockaddr_in	msg_clt_addr ;
        	int msg_clt_addr_len = sizeof(msg_clt_addr) ;

		w_rc_t rc = m_isMsgInquiry->accept( 
				(struct sockaddr *)&msg_clt_addr,
				&msg_clt_addr_len, msg_client_fhdl )  ; 

		if ( rc ) break ;

		SOCKET fd = int32(msg_client_fhdl.fd()) ; 

		// Now send all the clients currently active
		C_OLAPComm_System_Shore_Server	m_isInquiry(fd, false) ;

		/*
		C_OLAPComm_Packet*		pPacket = NULL ;
		C_OLAPComm_Empty_Packet*	recvPkt = NULL ;

		if ( m_isInquiry.ReceiveData(isReconciler, pPacket) < 0 )
			W_FATAL( m_isInquiry.rc.err_num() ) ; 

		recvPkt = (C_OLAPComm_Empty_Packet *)pPacket ; 
		delete recvPkt ; 
		*/
		
		C_OLAPComm_DataLinearizer	isData ;
		g_Clients.GetActiveClients( isData ) ; 

		C_OLAPComm_Generic_Packet	isPacket(isData) ; 
		rc = m_isInquiry.SendData( isReconciler, isPacket ) ;
	}
}
