/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_msg_thread.h
//
// This file contains the various thread definitions for the messaging
// subsystem. Service thread deals waits for msg clients to join. 
// Router thread routes the messages to the interested clients. 
// Inquiry thread gives a list of clients that are active in the server. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_MSG_THREAD_H)
#define	__OLAP_MSG_THREAD_H

#include "olap_server_globals.h"
#include "sfile.h"

#include "olap_comm_shore.h"
#include "olap_msg_queue.h"
#include "olap_borg_thread.h"

class C_OLAPMsg_Service_Thread : public C_OLAPBorg_Thread
{
public:
	C_OLAPMsg_Service_Thread(sfile_t* sd) :
		m_isMsgListener(sd)
	{ }  

	virtual ~C_OLAPMsg_Service_Thread()
	{ }

	virtual void run() ; 

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_MSG_SERVICE_THREAD ;
	}

protected: 

	sfile_t*			m_isMsgListener ;
} ;

class C_OLAPMsg_Thread : public C_OLAPBorg_Thread
{
public:

        C_OLAPMsg_Thread(sfile_t* sd) : 
			m_isMsgClientServer(sd, true) 
	{ 
		m_isSock = sd ;
	} 

        virtual ~C_OLAPMsg_Thread() ;

        virtual void run() ;

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_MSG_THREAD ;
	}

protected:

	sfile_t*				m_isSock ; 
	C_OLAPMsg_Queue				m_isMyQueue ; 
	C_OLAPComm_System_Shore_Server		m_isMsgClientServer ; 
} ;

class C_OLAPMsg_Router_Thread : public C_OLAPBorg_Thread
{
public:
	
	C_OLAPMsg_Router_Thread()
	{ }

	~C_OLAPMsg_Router_Thread()
	{ }

	virtual void run() ; 

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_MSG_ROUTER_THREAD ; 
	}

protected:
} ; 

class C_OLAPMsg_Inquiry_Thread : public C_OLAPBorg_Thread
{
public:
	
	C_OLAPMsg_Inquiry_Thread()
	{ }

	C_OLAPMsg_Inquiry_Thread(sfile_t* sd) :
		m_isMsgInquiry(sd)
	{ }  

	~C_OLAPMsg_Inquiry_Thread()
	{ }

	virtual void run() ; 

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_MSG_ROUTER_THREAD ; 
	}

protected:

	sfile_t*			m_isMsgInquiry ;
} ; 

#endif 
