/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_cache_operator.cpp
//
// This file consists of the implementation of the olap cache operator
//
// Author: Prasad Deshpande 
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_globals.h"
#include "olap_bkend_request.h"
#include "olap_cache_operator.h"
#include "evaluate.h"
#include "evaluator_cache.h"

extern cache g_QueryCache ;

class AutoReleaseEval
{
public:
    AutoReleaseEval(SchemaEvaluatorCache *secache, query_eval_info *qevalinfo)
    {
	m_secache = secache;
	m_qevalinfo = qevalinfo;
    }

    ~AutoReleaseEval()
    {
	W_COERCE(m_secache->ReleaseQueryEvaluator(m_qevalinfo));
    }

private:
    SchemaEvaluatorCache 	*m_secache;
    query_eval_info 		*m_qevalinfo;
};

w_rc_t C_OLAPOperator_Cache::start()
{
    m_isInputStreams[0] = NULL;
    return RCOK;
}

w_rc_t C_OLAPOperator_Cache::run()
{
    query_eval_info 	*qevalinfo;
    bool 		readexception;
    bool 		writeexception;
    AutoReleaseEval	*areval;

    // ToDo: Get query evaluator corresponding to this schema
    W_DO (m_isCacheNode->m_secache->GetQueryEvaluator(qevalinfo) );

    areval = new AutoReleaseEval(m_isCacheNode->m_secache, qevalinfo);
    w_auto_delete_t<AutoReleaseEval> autorel(areval);

    W_DO (qevalinfo->m_qeval->evaluate(m_isExecSql, m_isCacheNode->m_selPred,
				       m_isCacheNode->m_gbList,
				       m_isCacheNode->m_projList,
				       g_QueryCache,
				       m_isOutputStreams[0],
				       m_isCacheNode->op_tuple_type(0),
				       readexception,
				       writeexception,
				       true) );
    m_isInputStreams[0] = m_isExecSql.getStream();

    W_EXCEPTION( m_isOutputStreams[0], writeexception );
    W_EXCEPTION( m_isInputStreams[0], readexception );

    W_DO(m_isOutputStreams[0]->SendEndOfStream());
    return RCOK;
}

w_rc_t C_OLAPOperator_Cache::shutdown()
{
    return RCOK;
}

#ifdef __GNUG__
template class w_auto_delete_t<AutoReleaseEval>;
#endif
