/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_merge_operator.cpp
//
// This file consists of the implementation of the merge operator.
// It uses a simple round robin for pulling out the tuples from the
// set of input streams. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_merge_operator.h"

C_OLAPOperator_Merge::C_OLAPOperator_Merge( C_OLAPPlan_Merge_Node* node ) :
               	C_OLAPOperator(node, node->num_child(), 1),
                m_isEndOfStrReached( node->num_child() )
{
	m_isMergeNode = node ;
	m_isEndOfStrReached.zero() ;
}

w_rc_t C_OLAPOperator_Merge::start()
{
	m_isEndOfStrReached.zero() ; 
	return RCOK ; 
}

w_rc_t C_OLAPOperator_Merge::shutdown()
{
	return RCOK ; 
}

w_rc_t C_OLAPOperator_Merge::run()
{
	bool		eos ; 
	bool 		exception ;

	int32		count = 0 ; 

	while (1)
	{
		for ( u_int32 i = 0 ; i < m_isEndOfStrReached.size() ; i++ )
		{
			if ( m_isEndOfStrReached.is_clr(i) )
			{
			   W_DO( ReadAndWrite( i, eos, exception ) ) ;  

			   if ( eos )
				m_isEndOfStrReached.set(i) ; 	
			   else
			   	count++ ; 
			}
		}

		if ( m_isEndOfStrReached.num_set() == 
		                       m_isEndOfStrReached.size() )
			break ; 
	}

	// Send end of stream
	W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 

	return RCOK ; 
}

w_rc_t C_OLAPOperator_Merge::ReadAndWrite( int32 ipStream, 
                                           bool& eos, bool& exception )
{
	C_OLAP_Tuple tuple; 

	// Read the tuple 
	W_DO( m_isInputStreams[ipStream]->read(tuple, eos, exception) ) ; 

	if ( exception )
		cerr << "Merge received exception 1" << endl ; 

	W_EXCEPTION( m_isInputStreams[ipStream], exception ) ; 

	if ( !eos && !exception )
	{
		// Write the tuple 
		W_DO( m_isOutputStreams[0]->write( tuple, exception ) ) ; 

		if ( exception )
			cerr << "Merge received exception 2" << endl ; 

		W_EXCEPTION( m_isOutputStreams[ipStream], exception ) ; 
	}
 
	return RCOK ; 
}
