/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_sql_operator.cpp
//
// This file contains the implementation of the sql operator 
// used to contact the backend, execute the sql and get the results. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_globals.h"
#include "olap_bkend_request.h"

#include "olap_sql_operator.h"

extern C_OLAPBkend_WrkrThreads	g_bkend_thread_alloc ;

w_rc_t C_OLAPOperator_Sql::start()
{
	// Create a back end request
	m_isRequest = new C_OLAPBkend_Sql_Request( 
				m_isSqlNode->GetDatabase(), 
				m_isSqlNode->GetSql(), NULL,
				m_isPlanNode->ip_tuple_type(0) ) ;
	if ( !m_isRequest) W_FATAL( BORG_OUTOFMEMORY ) ; 

	// Get a free back end thread
	W_DO( g_bkend_thread_alloc.GetFreeWorker( m_isSqlNode->GetDatabase(), m_isFreeWorker) ) ;  

        // Now allocate a stream at to receive the tuples
        m_isStream = new C_OLAPTuple_ShoreNetwork_Stream(
                                        m_isFreeWorker->m_isSocket);
        if ( !m_isStream ) W_FATAL( BORG_OUTOFMEMORY ) ;

	m_isInputStreams[0] = m_isStream;

	return RCOK ; 
}

w_rc_t C_OLAPOperator_Sql::run()
{
	W_DO( SubmitBkendRequest( m_isFreeWorker ) ) ; 

	C_OLAP_Tuple		tuple ; 
	bool 			eos ; 
	bool			exception ; 

	// Now get tuple from the backend thread and pass it up
	W_DO( m_isInputStreams[0]->read(tuple, eos, exception) ) ; 

	while ( !eos && !exception )
	{
		W_DO( m_isOutputStreams[0]->write(tuple, exception) ) ; 
		W_EXCEPTION( m_isOutputStreams[0], exception ) ; 

		W_DO( m_isInputStreams[0]->read( tuple, eos, exception) ) ;
		W_EXCEPTION( m_isInputStreams[0], exception ) ; 
	}
 
	// Send end of stream
	if ( eos ) 
		W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 
	
	return RCOK ; 
}

w_rc_t C_OLAPOperator_Sql::shutdown()
{
	if ( m_isRequest ) delete m_isRequest ; 
	m_isRequest = NULL ; 

        if ( m_isStream ) delete m_isStream ;
        m_isStream = NULL ;

	// Release the back end worker thread
	W_DO( g_bkend_thread_alloc.ReleaseWorker( m_isSqlNode->GetDatabase(), m_isFreeWorker) ) ;  

	return RCOK ; 
}

w_rc_t C_OLAPOperator_Sql::SubmitBkendRequest(C_OLAPBkend_WrkrInfo* info)
{
	// Send the pointer to the backend request.
	int32 nbytes ; 
	W_DO( info->m_isSocket->send( &m_isRequest, sizeof(m_isRequest),
				      0, nbytes ) ) ;    

	return RCOK ; 
}
