/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_test_operator.cpp
//
// This file consists of the implementation of the test operators
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include "olap_bkend_request.h"

#include "olap_test_operator.h"

extern "C" {
       long random();
}

w_rc_t 
C_OLAPOperator_Producer::run()
{
	int i = 0 ; 
	bool exception ; 

        C_OLAP_TupleInfo tupInfo;
        C_OLAP_Tuple     tuple;

        char buf[100];
        tupInfo.SetNumAttrs(3);
        tupInfo.SetAttrDesc(0, BORG_STR);
        tupInfo.SetAttrDesc(1, BORG_INT);
        tupInfo.SetAttrDesc(2, BORG_STR);
        tupInfo.EndAttrDesc();
        tuple.Init(tupInfo);

	for ( i = 0 ; i < m_isProducerNode->GetNumTuples() ; i++ )
	{
                memset(buf, 0, 100);
                sprintf(buf, "%d", i);
		tuple.Reset();
                tuple.FillAttr(tupInfo[0], buf, 100);
                tuple.FillAttr(tupInfo[1], (void *)&i);
                tuple.FillAttr(tupInfo[2], buf, 50);

		W_DO( m_isOutputStreams[0]->write( tuple, exception ) ) ; 
	
		if ( exception ) cerr << "Producer received exception" << endl ; 

		W_EXCEPTION( m_isOutputStreams[0], exception ) ; 
	}

	cerr << "Producer produced " << i << " tuples" << endl ; 

	if ( i >= m_isProducerNode->GetNumTuples() ) 
		W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 

	return RCOK ; 
}

extern C_OLAPBkend_WrkrThreads	g_bkend_thread_alloc ;

w_rc_t C_OLAPOperator_BkendProducer::start()
{
	// Create a back end request
	m_isRequest = new C_OLAPBkend_Sql_Request("sales", "xx", NULL) ;
	if ( !m_isRequest) W_FATAL( BORG_OUTOFMEMORY ) ; 

	long xx = random() % 1000000 ; 

	cerr << "Getting a free worker----------------------- " << xx << endl ; 

	// Get a free back end thread
	W_DO( g_bkend_thread_alloc.GetFreeWorker("sales", m_isFreeWorker) ) ;  

	cerr << "Got a free worker ////////////////////////// " << xx << endl ; 

	// Now allocate a stream to receive the tuples 
	m_isStream = new C_OLAPTuple_ShoreNetwork_Stream(
					m_isFreeWorker->m_isSocket); 
	if ( !m_isStream ) W_FATAL( BORG_OUTOFMEMORY ) ; 
	
	return RCOK ; 
}

w_rc_t C_OLAPOperator_BkendProducer::run()
{
	w_rc_t			rc ;
	C_OLAP_Tuple		tuple ; 
	int32			count = 0 ; 

	bool 			eos ; 
	bool 			exception ; 

	// hook up the stream
	m_isInputStreams[0] = m_isStream ; 

	// Submit the backend request 
	W_DO( m_isFreeWorker->SubmitBkendRequest( m_isRequest ) ) ; 

	cerr << "Submitted the backend request%%%%%%%%%%%%%%%" << endl ; 

	// Now get tuple from the backend thread and pass it up
	W_DO( m_isInputStreams[0]->read( tuple, eos, exception ) ) ; 
 
	cerr << "Got the first tuple************************" << endl ; 

	while ( !eos && !exception ) 
	{
		W_DO( m_isOutputStreams[0]->write(tuple, exception) ) ; 
		W_EXCEPTION( m_isOutputStreams[0], exception ) ;  

		count++ ; 
		W_DO( m_isInputStreams[0]->read(tuple, eos, exception) );
		W_EXCEPTION( m_isInputStreams[0], exception ) ;  
	}
 
	// Send end of stream
	if ( eos )
		W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 
	
	// Unhook the back end stream
	m_isInputStreams[0] = NULL ; 
	return RCOK ; 
}

w_rc_t C_OLAPOperator_BkendProducer::shutdown()
{
	if ( m_isRequest ) delete m_isRequest ; 
	m_isRequest = NULL ; 

	if ( m_isStream ) delete m_isStream ; 
	m_isStream = NULL ; 

	// Release the back end worker thread
	W_DO( g_bkend_thread_alloc.ReleaseWorker("sales", m_isFreeWorker) ) ;  

	cerr << "Released the backend worker thread" << endl ; 

	return RCOK ; 
}

/* ------------------------------------------------------------------ */
// Code added by Amit

w_rc_t 
C_OLAPOperator_Producer_1::run()
{
    bool exception ; 
//    char buf[1024];
    C_OLAP_Tuple     tuple1;
    C_OLAP_TupleInfo tupInfo1;
    unsigned int len;

//    tuple1.SetTuple(buf, 1024);

    tupInfo1.SetNumAttrs(3);
    tupInfo1.SetAttrDesc(0, BORG_INT);
    tupInfo1.SetAttrDesc(1, BORG_FLOAT);
    tupInfo1.SetAttrDesc(2, BORG_STR);
    tupInfo1.EndAttrDesc();
    tuple1.Init(tupInfo1);

    int i, j, jc;
    int ival;
    float fval;
    char sval[32];

    for (i = 0; i < m_isProducerNode_1->GetNumTuples() ; i++)
    {
        jc = i % 6;

        ival = jc;
        fval = i;
        for (j=0; j < i%10+1; j++)
            sval[j] = i + '0';
        sval[j] = '\0';

        /* this step is essential for reusing tuples with
           variable length attributes */
        tuple1.Reset();

        tuple1.FillAttr(tupInfo1[0], (char *)&jc);
        tuple1.FillAttr(tupInfo1[1], (char *)&fval);
        tuple1.FillAttr(tupInfo1[2], sval, strlen(sval) + 1);

	cerr << " Pruducer 1 = ("
             << *(int *)tuple1.GetAttr(0, len) << ", "
             << *(float *)tuple1.GetAttr(1, len) << ", "
             << (char *)tuple1.GetAttr(2, len) << ")\n";

	W_DO( m_isOutputStreams[0]->write( tuple1, exception ) ) ; 
	
	if ( exception ) cerr << "Producer received exception" << endl ; 
	
	W_EXCEPTION( m_isOutputStreams[0], exception ) ; 
    }

    cerr << "Producer produced " << i << " tuples" << endl ; 
    
    if ( i >= m_isProducerNode_1->GetNumTuples() ) 
	W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 

    return RCOK ; 
}


w_rc_t 
C_OLAPOperator_Producer_2::run()
{
    bool exception ; 
//    char buf[1024];
    C_OLAP_Tuple     tuple1;
    C_OLAP_TupleInfo tupInfo1;
    unsigned int len;

//    tuple1.SetTuple(buf, 1024);

    tupInfo1.SetNumAttrs(3);
    tupInfo1.SetAttrDesc(0, BORG_INT);
    tupInfo1.SetAttrDesc(1, BORG_FLOAT);
    tupInfo1.SetAttrDesc(2, BORG_STR);
    tupInfo1.EndAttrDesc();
    tuple1.Init(tupInfo1);

    int i, j, jc;
    int ival;
    float fval;
    char sval[32];


    for (i = 0; i < m_isProducerNode_2->GetNumTuples() ; i++)
    {
        jc = i % 9;

        ival = jc;
        fval = i;
        for (j=0; j<i%10 + 1; j++)
            sval[j] = i + '0';
        sval[j] = '\0';

        /* this step is essential for reusing tuples with
           variable length attributes */
        tuple1.Reset();

        tuple1.FillAttr(tupInfo1[0], (char *)&jc);
        tuple1.FillAttr(tupInfo1[1], (char *)&fval);
        tuple1.FillAttr(tupInfo1[2], sval, strlen(sval) + 1);

	cerr << " Pruducer 2 = ("
             << *(int *)tuple1.GetAttr(0, len) << ", "
             << *(float *)tuple1.GetAttr(1, len) << ", "
             << (char *)tuple1.GetAttr(2, len) << ")\n";

	W_DO( m_isOutputStreams[0]->write( tuple1, exception ) ) ; 
	
	if ( exception ) cerr << "Producer received exception" << endl ; 
	
	W_EXCEPTION( m_isOutputStreams[0], exception ) ; 
    }

    cerr << "Producer produced " << i << " tuples" << endl ; 
    
    if ( i >= m_isProducerNode_2->GetNumTuples() ) 
	W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 

    return RCOK ; 
}



w_rc_t 
C_OLAPOperator_ExprTest::run()
{
    bool eos, exception ; 
    unsigned int len;

    // Setup the output tuple buffer...
/*
    m_isExprInfo.tup[ColumnIdentifier::output]->Init(
	m_isExprInfo.tInfo[ColumnIdentifier::output]);
*/

    C_OLAP_Tuple* outTup = m_isExprInfo.tup[ColumnIdentifier::output];
    C_OLAP_Tuple* leftTup = m_isExprInfo.tup[ColumnIdentifier::left];
    C_OLAP_Tuple* rightTup = m_isExprInfo.tup[ColumnIdentifier::right];

    while (1)
    {
	W_DO( m_isInputStreams[0]->read(*leftTup, eos, exception)); 
	if ( eos ) break ; 

	cerr << " Consumer: Input1 = ("
             << *(int *)m_isInputTuples[0].GetAttr(0, len) << ", "
             << *(float *)m_isInputTuples[0].GetAttr(1, len) << ", "
             << (char *)m_isInputTuples[0].GetAttr(2, len) << ")\n";

	W_DO( m_isInputStreams[1]->read(*rightTup, eos, exception));
	if ( eos ) break ; 

	cerr << " Consumer: Input2 = ("
             << *(int *)m_isInputTuples[1].GetAttr(0, len) << ", "
             << *(float *)m_isInputTuples[1].GetAttr(1, len) << ", "
             << (char *)m_isInputTuples[1].GetAttr(2, len) << ")\n";

	if ( m_isPlanNode->predicate()->evaluate(&m_isExprInfo))
	{
	    cerr << "---------------- Condition satisfied" << endl;
	    ExprValue ev;

	    outTup->Reset();
	    m_isPlanNode->project_list()->Eval( m_isExprInfo, ev );

	    W_DO( m_isOutputStreams[0]->write(*outTup, exception)); 
	    
	    if ( exception ) cerr << "ExprTest received exception" << endl ; 
	    
	    W_EXCEPTION( m_isOutputStreams[0], exception ) ; 
	}
    }

    W_DO( m_isOutputStreams[0]->SendEndOfStream() ) ; 

    cout << "ExprTest Completed" << endl ; 
    
    return RCOK ; 
}

