/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_plan.cpp
//
// This file contains the implementation for physical plan, starting
// up the physical plan and shutting down the operators of a plan
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_globals.h"
#include "olap_stream.h"
#include "olap_operator.h"
#include "olap_plan_node.h"
#include "olap_plan.h"

#include "olap_qproc.h"
#include "olap_tuple.h"

extern C_OLAPOperator_Queue     g_OperatorExecQueue ;

/*
w_rc_t C_OLAP_Physical_Plan::startOperators( C_OLAPPlan_Node* root_node ) ; 
w_rc_t C_OLAP_Physical_Plan::waitOpTreeCompletion( C_OLAPPlan_Node* root_node ) ; 

w_rc_t C_OLAP_Physical_Plan::instantiatePlan( C_OLAPPlan_Node* root_node ) ; 
void C_OLAP_Physical_Plan::destroyPlan( C_OLAPPlan_Node* root_node ) ; 

w_rc_t C_OLAP_Physical_Plan::getOperatorError( C_OLAPPlan_Node* root_node ) ; 
*/

C_OLAP_Physical_Plan::~C_OLAP_Physical_Plan()
{
	destroyPlan( m_isRoot ) ; 
}

void C_OLAP_Physical_Plan::testPlan1()
{
	C_OLAPPlan_BkendProducer_Node* producer = NULL ; 

	producer =  new C_OLAPPlan_BkendProducer_Node ;		
	if ( !producer ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	m_isRoot = producer ; 
}

void C_OLAP_Physical_Plan::testPlan5()
{
        const char TEST_SQL[] = "Select * from cChannel" ;

        C_OLAP_TupleInfo tupInfo;
        tupInfo.SetNumAttrs(3);

        tupInfo.SetAttrDesc(0, BORG_STR);
        tupInfo.SetAttrDesc(1, BORG_INT);
        tupInfo.SetAttrDesc(2, BORG_INT);
        tupInfo.EndAttrDesc();

        GenSqlPlan("apb", TEST_SQL, &tupInfo) ;
}

void C_OLAP_Physical_Plan::GenSqlPlan(const w_string &database,
                                          const w_string &sql,
                                          C_OLAP_TupleInfo *tupInfo)
{
        C_OLAPPlan_Sql_Node* producer = NULL ;

        producer =  new C_OLAPPlan_Sql_Node(database, sql) ;
        if ( !producer ) W_FATAL( BORG_OUTOFMEMORY ) ;

        // Create the tuple types array...
        C_OLAP_TupleInfo* tupInfo1 = new C_OLAP_TupleInfo,
                        * tupInfo2 = new C_OLAP_TupleInfo;

        (*tupInfo2) = (*tupInfo1) = (*tupInfo);

        producer->add_ip_tuple_type( tupInfo1 ) ;
        producer->add_op_tuple_type( tupInfo2 ) ;

        m_isRoot = producer ;
}



void C_OLAP_Physical_Plan::testPlan4()
{
	C_OLAPPlan_BkendProducer_Node* producer1 = NULL ; 

	producer1 =  new C_OLAPPlan_BkendProducer_Node ;		
	if ( !producer1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_BkendProducer_Node* producer2 = NULL ; 

	producer2 =  new C_OLAPPlan_BkendProducer_Node ;		
	if ( !producer2 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_Merge_Node* merge1 = NULL ; 

	merge1 = new C_OLAPPlan_Merge_Node ; 
	if ( !merge1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	merge1->add_child( producer1 ) ; 
	merge1->add_child( producer2 ) ; 

	m_isRoot = merge1 ; 
}

void C_OLAP_Physical_Plan::testPlan3()
{
    // Create the tuple types array...
    C_OLAP_TupleInfo* tupInfo1 = new C_OLAP_TupleInfo, 
	*tupInfo2 = new C_OLAP_TupleInfo, 
	*otupInfo = new C_OLAP_TupleInfo;

    tupInfo1->SetNumAttrs(3);
    tupInfo1->SetAttrDesc(0, BORG_INT);
    tupInfo1->SetAttrDesc(1, BORG_FLOAT);
    tupInfo1->SetAttrDesc(2, BORG_STR);
    tupInfo1->EndAttrDesc();

    tupInfo2->SetNumAttrs(3);
    tupInfo2->SetAttrDesc(0, BORG_INT);
    tupInfo2->SetAttrDesc(1, BORG_FLOAT);
    tupInfo2->SetAttrDesc(2, BORG_STR);
    tupInfo2->EndAttrDesc();

    otupInfo->SetNumAttrs(4);
    otupInfo->SetAttrDesc(0, BORG_INT);
    otupInfo->SetAttrDesc(1, BORG_FLOAT);
    otupInfo->SetAttrDesc(2, BORG_INT);
    otupInfo->SetAttrDesc(3, BORG_STR);
    otupInfo->EndAttrDesc();

    // Create the predicate and the join expression

    // Create an predicate expression tree...
    C_OLAPAttr * a1 = new C_OLAPAttr;
    a1->ColID.strm = ColumnIdentifier::left;
    a1->ColID.colNum = 0;
    a1->AttrType = BORG_INT;

    C_OLAPValue* v = new C_OLAPValue;
    v->int_val = 0;
    v->type    = BORG_INT;

    C_OLAPAttr * a11 = new C_OLAPAttr;
    a11->ColID.strm = ColumnIdentifier::left;
    a11->ColID.colNum = 0;
    a11->AttrType = BORG_INT;

    C_OLAPValue* v11 = new C_OLAPValue;
    v11->int_val = 0;
    v11->type    = BORG_INT;

    C_OLAPAttr * a2 = new C_OLAPAttr;
    a2->ColID.strm = ColumnIdentifier::right;
    a2->ColID.colNum = 0;
    a2->AttrType = BORG_INT;

    C_OLAPArithExpr* ae = new C_OLAPArithExpr;
    ae->aetype() = C_OLAPArithExpr::Plus;
    ae->AddArg(a2);
    ae->AddArg(v);

    C_OLAPCompExpr * ce = new C_OLAPCompExpr;
    ce->cetype() = C_OLAPCompExpr::Lt;
    ce->AddArg(a1);
    ce->AddArg(ae);

    C_OLAPCompExpr* ce2 = new C_OLAPCompExpr;
    ce2->cetype() = C_OLAPCompExpr::Eq;
    ce2->AddArg(a11);
    ce2->AddArg(v11);

    C_OLAPLogicalExpr * le = new C_OLAPLogicalExpr;
    le->letype() = C_OLAPLogicalExpr::And;
    le->AddArg(ce);
    le->AddArg(ce2);

    C_OLAP_GeneralExpr * GenExpr = new C_OLAP_GeneralExpr(le);

    // Setup the projection list
    C_OLAPAttr *ja1, *ja2, *ja3, *ja4;
    ja1 = new C_OLAPAttr;
    ja1->ColID.strm = ColumnIdentifier::left;
    ja1->ColID.colNum = 0;
    ja1->AttrType = BORG_INT;

    ja2 = new C_OLAPAttr;
    ja2->ColID.strm = ColumnIdentifier::right;
    ja2->ColID.colNum = 0;
    ja2->AttrType = BORG_FLOAT;

    ja3 = new C_OLAPAttr;
    ja3->ColID.strm = ColumnIdentifier::left;
    ja3->ColID.colNum = 1;
    ja3->AttrType = BORG_INT;

    ja4 = new C_OLAPAttr;
    ja4->ColID.strm = ColumnIdentifier::right;
    ja4->ColID.colNum = 2;
    ja4->AttrType = BORG_STR;

    C_OLAPProjectList* jl = new C_OLAPProjectList;
    jl->AddArg(ja1);
    jl->AddArg(ja2);
    jl->AddArg(ja3);
    jl->AddArg(ja4);


    /* ------------------------------------------------------ */
    // Create the plan nodes

    C_OLAPPlan_Producer_Node_1* producer1 = NULL ; 
    C_OLAPPlan_Producer_Node_2* producer2 = NULL ;     
    C_OLAPPlan_ExprTest_Node  * consumer  = NULL ;

    producer1 =  new C_OLAPPlan_Producer_Node_1(20);
    if ( !producer1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

    producer2 =  new C_OLAPPlan_Producer_Node_2(20);
    if ( !producer2 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

    consumer  =  new C_OLAPPlan_ExprTest_Node;
    if ( !consumer ) W_FATAL( BORG_OUTOFMEMORY ) ; 

    consumer->add_child( producer1 ) ; 
    consumer->add_child( producer2 ) ; 

    consumer->add_ip_tuple_type( tupInfo1 ) ; 
    consumer->add_ip_tuple_type( tupInfo2 ) ; 
    consumer->add_op_tuple_type( otupInfo ) ;

    consumer->set_predicate(GenExpr) ; 
    consumer->set_project_list( jl) ; 
    m_isRoot = consumer ; 
}

void C_OLAP_Physical_Plan::testPlan2()
{
	C_OLAPPlan_Producer_Node* producer1 = NULL ; 

	producer1 =  new C_OLAPPlan_Producer_Node(100);		
	if ( !producer1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_Producer_Node* producer2 = NULL ; 

	producer2 =  new C_OLAPPlan_Producer_Node(50);		
	if ( !producer2 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_Merge_Node* merge1 = NULL ; 

	merge1 = new C_OLAPPlan_Merge_Node ; 
	if ( !merge1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	merge1->add_child( producer1 ) ; 
	merge1->add_child( producer2 ) ; 

	m_isRoot = merge1 ; 
}

void C_OLAP_Physical_Plan::testPlan()
{
	C_OLAPPlan_Producer_Node* producer1 = NULL ; 
	C_OLAPPlan_Producer_Node* producer2 = NULL ; 
	C_OLAPPlan_Producer_Node* producer3 = NULL ; 
	C_OLAPPlan_Producer_Node* producer4 = NULL ; 
	C_OLAPPlan_Producer_Node* producer5 = NULL ; 
	C_OLAPPlan_Producer_Node* producer6 = NULL ; 

	producer1 =  new C_OLAPPlan_Producer_Node(20);		
	if ( !producer1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 
	
	producer2 =  new C_OLAPPlan_Producer_Node(40);		
	if ( !producer2 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	producer3 =  new C_OLAPPlan_Producer_Node(30);		
	if ( !producer3 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	producer4 =  new C_OLAPPlan_Producer_Node(25);		
	if ( !producer4 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	producer5 =  new C_OLAPPlan_Producer_Node(35);		
	if ( !producer5 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	producer6 =  new C_OLAPPlan_Producer_Node(45);		
	if ( !producer6 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_Merge_Node* merge1 = NULL ; 

	merge1 = new C_OLAPPlan_Merge_Node ; 
	if ( !merge1 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_Merge_Node* merge2 = NULL ; 

	merge2 = new C_OLAPPlan_Merge_Node ; 
	if ( !merge2 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	C_OLAPPlan_Merge_Node* merge3 = NULL ; 

	merge3 = new C_OLAPPlan_Merge_Node ; 
	if ( !merge3 ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	merge1->add_child(producer1) ; 	
	merge1->add_child(producer2) ; 
	merge1->add_child(producer3) ; 

	merge2->add_child(merge1) ; 
	merge2->add_child(producer4) ; 

	merge3->add_child(merge2) ; 
	merge3->add_child(producer5) ; 
	merge3->add_child(producer6) ; 

	m_isRoot = merge3 ; 
}

w_rc_t C_OLAP_Physical_Plan::executePlan( C_OLAPStream* result )
{
	// Instantiate the root operator
	m_isRoot->MakeOperator() ;

	// Inform that it is the last operator 
	m_isRoot->Operator()->LastOperator() ;

	// Make a dynamic instance of the plan
	W_DO( instantiatePlan( m_isRoot ) ) ; 	

	// Set up the client output stream
	m_isRoot->Operator()->SetOutputStream(0,result) ; 

	// Start the operators 
	W_DO( startOperators( m_isRoot ) ) ; 

	// Wait until the all the opertors complete
	W_DO( waitOpTreeCompletion( m_isRoot ) ) ;

	// Check if any exception occurred during processing 
	if ( m_isRoot->Operator()->DidErrorOccurred() )
	{
		// Now traverse down the tree to get an error 
		return getOperatorError( m_isRoot ) ; 
	}

	return RCOK ; 
}

w_rc_t C_OLAP_Physical_Plan::startOperators( C_OLAPPlan_Node* root_node )
{
	w_list_t<C_OLAPPlan_Node> queue(offsetof(C_OLAPPlan_Node,m_isLink)) ; 
	C_OLAPPlan_Node*	  pnode = root_node ; 

	queue.push( pnode ) ; 
	while ( (pnode = queue.pop()) )
	{
		W_DO( g_OperatorExecQueue.enqueue( pnode->Operator() ) ) ; 

		for ( int32 i = 0 ; i < pnode->num_child() ; i++ )
			queue.append( pnode->child(i) ) ; 
	}

	return RCOK ; 
}

w_rc_t C_OLAP_Physical_Plan::instantiatePlan( C_OLAPPlan_Node* root_node )
{
	for ( int32 i = 0 ; i < root_node->num_child() ; i++ )
	{
		cerr << "Allocating a memory stream " << endl ; 	
		C_OLAPTuple_MemoryStream* ip_stream = new C_OLAPTuple_MemoryStream ;
		if ( !ip_stream ) W_FATAL( BORG_OUTOFMEMORY ) ; 		 
		root_node->Operator()->SetInputStream( i, ip_stream ) ; 

		root_node->child(i)->MakeOperator() ; 
		root_node->child(i)->Operator()->SetOutputStream( 0, ip_stream ) ;
		W_DO( instantiatePlan( root_node->child(i) ) ) ; 
	}

	return RCOK ; 
}

void C_OLAP_Physical_Plan::destroyPlan( C_OLAPPlan_Node* root_node )
{
	for ( int32 i = 0 ; i < root_node->num_child() ; i++ )
	{
		if ( root_node->Operator()->InputStream(i) )
		{
			cerr << "deallocating a memory stream " << endl ; 	
			delete root_node->Operator()->InputStream(i) ; 
		}

		destroyPlan( root_node->child(i) ) ; 
	}

	delete root_node ; 
}

w_rc_t C_OLAP_Physical_Plan::waitOpTreeCompletion( C_OLAPPlan_Node* root_node )
{
	W_DO( root_node->Operator()->WaitForCompletion() ) ; 

	for ( int32 i = 0 ; i < root_node->num_child() ; i++ )
	{
		W_DO( waitOpTreeCompletion( root_node->child(i) ) ) ;
	}

	return RCOK ; 

}

w_rc_t C_OLAP_Physical_Plan::getOperatorError( C_OLAPPlan_Node* root_node )
{
	for ( int32 i = 0 ; i < root_node->num_child() ; i++ )
	{
		if ( root_node->Operator()->GetRC() && 
		     root_node->Operator()->GetRC().err_num() != BORGOP_ABNORMAL_TERMINATION )
			return root_node->Operator()->GetRC() ; 

		W_DO( getOperatorError( root_node->child(i) ) ) ; 
	}

	return RCOK ; 
}

#ifdef __GNUG__
template class w_list_t<C_OLAPPlan_Node>;
template class w_list_i<C_OLAPPlan_Node>;
#endif
