/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_plan_node.cpp
//
// This file defines the implementation of a plan node in the physical
// plan. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include <w.h>
#include <sm_vas.h>

#include "olap_plan_node.h"
#include "olap_operator.h"

C_OLAPPlan_Node::C_OLAPPlan_Node() : m_isChild(4), 
	m_isInputTupleTypes(4), m_isOutputTupleTypes(2)
{
	m_isOp = NULL ; m_isPred = NULL ; m_isProj = NULL ;
}

C_OLAPPlan_Node::~C_OLAPPlan_Node() 
{
	if ( m_isOp ) delete m_isOp ; 
	if ( m_isPred ) delete m_isPred ;
	if ( m_isProj ) delete m_isProj ;

	for ( u_int32 i = 0 ; i < m_isInputTupleTypes.getsize() ; i++ )
	{
		delete m_isInputTupleTypes[i] ; 
		m_isInputTupleTypes[i] = NULL ; 
	}

	for ( u_int32 i = 0 ; i < m_isOutputTupleTypes.getsize() ; i++ )
	{
		delete m_isOutputTupleTypes[i] ; 
		m_isOutputTupleTypes[i] = NULL ; 
	}
}

void C_OLAPPlan_Merge_Node::MakeOperator()
{
	m_isOp = new C_OLAPOperator_Merge( this ) ; 
	if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ; 
}

void C_OLAPPlan_Producer_Node::MakeOperator()
{
	m_isOp = new C_OLAPOperator_Producer( this ) ; 
	if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ; 
}

void C_OLAPPlan_BkendProducer_Node::MakeOperator()
{
	m_isOp = new C_OLAPOperator_BkendProducer( this ) ; 
	if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ; 
}

void C_OLAPPlan_Sql_Node::MakeOperator()
{
	m_isOp = new C_OLAPOperator_Sql( this ) ; 
	if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ; 
}

// -----------------------------------------------------------
// Added by Amit to test the expression processing

void C_OLAPPlan_Producer_Node_1::MakeOperator()
{
        m_isOp = new C_OLAPOperator_Producer_1( this ) ;  
        if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ;
}

void C_OLAPPlan_Producer_Node_2::MakeOperator()
{
        m_isOp = new C_OLAPOperator_Producer_2( this ) ;
        if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ;
}

void C_OLAPPlan_ExprTest_Node::MakeOperator()
{
        m_isOp = new C_OLAPOperator_ExprTest( this ) ;
        if ( !m_isOp ) W_FATAL(BORG_OUTOFMEMORY) ;
}

#ifdef __GNUG__
template class dynarray<C_OLAPPlan_Node*>;
template class dynarray<C_OLAP_TupleInfo*>;
#endif
