/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_plan_node.h
//
// This file defines the interface for the plan node for physical plan 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_PHYSICAL_PLAN_H)
#define	__OLAP_PHYSICAL_PLAN_H

#include <w.h>
#include "olap_types.h"
#include "olap_basics.h"

#include "olap_join_common.h"
#include "w_string.h"

class C_OLAPOperator ; 
class C_OLAPPlan_Node ; 
class C_OLAP_TupleInfo ; 

typedef dynarray<C_OLAPPlan_Node*>	ChildArgs; 

class C_OLAPProjectList;
class C_OLAP_Predicate ; 

class C_OLAPPlan_Node  
{
public:

	enum OLAP_Plan_Node_Type {
		merge_p = 0,
		exec_sql_p, 
		producer_p, 
		bkend_producer_p,
		meta_all_p,
		meta_schema_p
	} ;
		
	C_OLAPPlan_Node() ;  

	virtual ~C_OLAPPlan_Node() ;
	
	C_OLAPOperator* Operator()
	{
		return m_isOp ; 
	}

	C_OLAPPlan_Node* child(int32 n)
	{ 
		return m_isChild[n] ;
	}
	
	void add_child(int32 n, C_OLAPPlan_Node* aNode)
	{
		m_isChild[n] = aNode ; 
	}

	void add_child(C_OLAPPlan_Node* aNode)
	{
		m_isChild.push_back(aNode) ; 
	}

	C_OLAP_TupleInfo* ip_tuple_type(int32 n)
	{
		return m_isInputTupleTypes[n] ; 
	}

	void add_ip_tuple_type(int32 n, C_OLAP_TupleInfo* aTupInfo) 
	{
		m_isInputTupleTypes[n] = aTupInfo ; 
	}

	void add_ip_tuple_type(C_OLAP_TupleInfo* aTupInfo) 
	{
		m_isInputTupleTypes.push_back( aTupInfo ) ; 
	}

	C_OLAP_TupleInfo* op_tuple_type(int32 n)
	{
		return m_isOutputTupleTypes[n] ; 
	}

	void add_op_tuple_type(int32 n, C_OLAP_TupleInfo* aTupInfo) 
	{
		m_isOutputTupleTypes[n] = aTupInfo ; 
	}

	void add_op_tuple_type(C_OLAP_TupleInfo* aTupInfo) 
	{
		m_isOutputTupleTypes.push_back(aTupInfo) ; 
	}

	int32 num_child()
	{
		return m_isChild.getsize() ; 
	}

	C_OLAPProjectList* project_list()
	{
		return m_isProj ; 
	}

	C_OLAP_Predicate* predicate()
	{
		return m_isPred ;
	}

	virtual OLAP_Plan_Node_Type NodeType() = 0 ; 
	virtual void MakeOperator() = 0 ; 

protected:

	void set_project_list( C_OLAPProjectList* projList )
	{
		m_isProj = projList ; 
	}
	
	void set_predicate( C_OLAP_Predicate* pred )
	{
		m_isPred = pred ; 
	}

	C_OLAP_Predicate*		m_isPred;		// optional predicate list
	C_OLAPProjectList*		m_isProj;		// optional project list

	ChildArgs 			m_isChild ; 		// number of children
	TupleTypes			m_isInputTupleTypes ; 	// types of input tuples
	TupleTypes			m_isOutputTupleTypes ;	// types of output tuples

	C_OLAPOperator*			m_isOp ; 
	w_link_t			m_isLink ;

	friend class C_OLAP_Physical_Plan ;
	friend class C_OLAPOperator;
} ; 

class C_OLAPPlan_Merge_Node : public C_OLAPPlan_Node
{
public:

	C_OLAPPlan_Merge_Node() 
	{ } 

	virtual ~C_OLAPPlan_Merge_Node() 
	{ } 

	virtual OLAP_Plan_Node_Type NodeType() 
	{
		return merge_p ;
	}

	virtual void MakeOperator() ; 

protected:
} ; 

class C_OLAPPlan_BkendProducer_Node : public C_OLAPPlan_Node
{
public:

	C_OLAPPlan_BkendProducer_Node() 
	{ 
	} 
	
	virtual ~C_OLAPPlan_BkendProducer_Node() 
	{ } 

	virtual OLAP_Plan_Node_Type NodeType() 
	{
		return bkend_producer_p ;
	}

	virtual void MakeOperator() ;

protected:
} ;

class C_OLAPPlan_Producer_Node : public C_OLAPPlan_Node
{
public:

	C_OLAPPlan_Producer_Node( int32 num_tuples ) 
	{ 
		m_iNumTuples = num_tuples ; 
	} 
	
	virtual ~C_OLAPPlan_Producer_Node() 
	{ } 

	virtual OLAP_Plan_Node_Type NodeType() 
	{
		return producer_p ;
	}

	virtual void MakeOperator() ;

	int32 GetNumTuples()
	{
		return m_iNumTuples ; 
	}

protected:

	int32 		m_iNumTuples ; 
} ;

class C_OLAPPlan_Sql_Node : public C_OLAPPlan_Node
{
public:

        C_OLAPPlan_Sql_Node(const w_string &database, const w_string &sql) :
            m_isDatabase(database), m_isSql(sql)
        { }


	virtual ~C_OLAPPlan_Sql_Node() 
	{ } 

	virtual OLAP_Plan_Node_Type NodeType()
	{
		return exec_sql_p ; 
	}

	virtual void MakeOperator() ;

	w_string&	GetDatabase()
	{
		return m_isDatabase ; 
	}

	w_string& GetSql()
	{
		return m_isSql ; 
	}

protected:

	w_string			m_isDatabase ;
	w_string			m_isSql ; 
} ; 


// -----------------------------------------------------------
// Added by Amit

class C_OLAPPlan_Producer_Node_1 : public C_OLAPPlan_Node
{
public:

    C_OLAPPlan_Producer_Node_1( int32 num_tuples )
    { 
        m_iNumTuples   = num_tuples ; 
    } 
	
    virtual ~C_OLAPPlan_Producer_Node_1() 
    { } 

    virtual OLAP_Plan_Node_Type NodeType() 
    {
        return producer_p ;
    }

    int32 GetNumTuples()
    {
        return m_iNumTuples ; 
    }

    virtual void MakeOperator() ;

protected:

    int32 	       	m_iNumTuples ; 

} ;

class C_OLAPPlan_Producer_Node_2 : public C_OLAPPlan_Node
{
public:

    C_OLAPPlan_Producer_Node_2( int32 num_tuples )
    { 
        m_iNumTuples   = num_tuples ; 
    } 
	
    virtual ~C_OLAPPlan_Producer_Node_2() 
    { 
    } 

    virtual OLAP_Plan_Node_Type NodeType() 
    {
        return producer_p ;
    }

    int32 GetNumTuples()
    {
        return m_iNumTuples ; 
    }

    virtual void MakeOperator() ;

protected:

    int32 	       m_iNumTuples ; 
} ;


class C_OLAPPlan_ExprTest_Node : public C_OLAPPlan_Node
{
public:

    C_OLAPPlan_ExprTest_Node() 
    { 
    } 
	
    virtual ~C_OLAPPlan_ExprTest_Node() 
    { 
    } 

    virtual OLAP_Plan_Node_Type NodeType() 
    {
        return producer_p ;
    }

    virtual void MakeOperator() ;

protected:
} ;

#endif
