/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_attr.cpp
//
// This file contains the implementation for attributes, values and
// expressions.  
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////

#include <stdio.h>

#include "olap_attr.h"

/* ------------------- Eval: C_OLAPAttr ---------------------- */

bool C_OLAPAttr::Eval(ExprInfo& einfo, ExprValue& expval)
{
    expval.type = AttrType;

    unsigned int len;
    int int_val;
    float float_val;
    char *ptr;
    char buf[1024];    // Has to be set to MAXFLDLENGTH...

    // The stored type might be different from the required type
    // This means that the actual type will have to be converted to the
    // required type...
    BorgTypes atype = einfo.tInfo[ColID.strm]->GetAttrDesc(ColID.colNum).m_type->m_typeid;

    switch (atype)
    {
    case BORG_INT:
        int_val = * (int*) einfo.tup[ColID.strm]->GetAttr(ColID.colNum, len);
        switch (AttrType)
        {
        case BORG_INT:
            expval.i = int_val;
            return true;

        case BORG_FLOAT:
            expval.f = float(int_val);
            return true;

        case BORG_STR:
            sprintf(expval.buf, "%d", int_val);
            return true;
        default:
            assert(0);       // Error
            return false;
        }
        break;

    case BORG_FLOAT:
        float_val = * (float*) einfo.tup[ColID.strm]->GetAttr(ColID.colNum, len);
        switch (AttrType)
        {
        case BORG_INT:
            expval.i = int(float_val);
            return true;

        case BORG_FLOAT:
            expval.f = float_val;
            return true;

        case BORG_STR:
            sprintf(expval.buf, "%f", float_val);
            return true;
        default:
            assert(0);       // Error
            return false;
        }
        break;

    case BORG_STR:
        ptr = (char*) einfo.tup[ColID.strm]->GetAttr(ColID.colNum, len);
        strncpy(buf, ptr, len);
        buf[len] = '\0';
        expval.i = len;

        switch (AttrType)
        {
        case BORG_INT:
            expval.i = atoi(buf);   // Assume NULL termination
            return true;

        case BORG_FLOAT:
            expval.f = atof(buf);
            return true;

        case BORG_STR:
            strncpy(expval.buf, ptr, len);
            return true;
        default:
            assert(0);       // Error
            return false;
        }
        break;
    default:
        assert(0);
    }

    return true;
}


void C_OLAPAttr::WriteTree(ostream& fs)
{
  C_OLAPExprNode::WriteTree(fs);
  fs << "(C_OLAPAttr " << " (ColID.strm = " << int(ColID.strm)
     << ",ColID.colNum = " << ColID.colNum << ") ";
  fs << "AttrType = " << int(AttrType) << ")\n";
}


void C_OLAPAttr::ReadTree(istream& fs)
{
  C_OLAPExprNode::ReadTree(fs);
  char buf[32];
  int32 tint;
  fs >> buf >> buf >> buf;
  fs >> tint >> buf >> buf;
  ColID.strm = ColumnIdentifier::StreamID(tint);
  fs >> ColID.colNum >> buf >> buf >> buf;
  fs >> tint >> buf;
  AttrType = BorgTypes(tint);
}




/* ------------------- Eval: C_OLAPProjectList --------------- */

bool C_OLAPProjectList::Eval(ExprInfo& einfo, ExprValue& expval)
{
    bool bRet = false;

    for (int i = 0; i < nargs(); i++)
    {
        bRet = arg(i)->Eval(einfo, expval);
        if (!bRet) return bRet;

        BorgTypes outtype = einfo.tInfo[ColumnIdentifier::output]->GetAttrDesc(i).m_type->m_typeid;
        MakeType(outtype, expval);

        switch(outtype)
        {
        case BORG_INT:
            einfo.tup[ColumnIdentifier::output]->FillAttr(
                (*einfo.tInfo[ColumnIdentifier::output])[i], //->GetAttrDesc(i),
                (char*) &expval.i);
            break;

        case BORG_FLOAT:
            einfo.tup[ColumnIdentifier::output]->FillAttr(
                (*einfo.tInfo[ColumnIdentifier::output])[i], //->GetAttrDesc(i),
                (char*) &expval.f);
            break;

        case BORG_STR:
            einfo.tup[ColumnIdentifier::output]->FillAttr(
                (*einfo.tInfo[ColumnIdentifier::output])[i], //->GetAttrDesc(i),
                expval.buf, expval.i+1);
            break;
        default:
            assert(0);
        }
    }

    return true;
}


void C_OLAPProjectList::MakeType(BorgTypes outtype, ExprValue& expval)
{
    switch (outtype)
    {
    case BORG_INT:
        switch (expval.type)
        {
        case BORG_INT:   break;
        case BORG_FLOAT: expval.i = int(expval.f); break;
        case BORG_STR:   expval.i = atoi(expval.buf); break;
        default:         assert(0);
        }
        break;

    case BORG_FLOAT:
        switch (expval.type)
        {
        case BORG_INT:   expval.f = float(expval.i); break;
        case BORG_FLOAT: break;
        case BORG_STR:   expval.f = atof(expval.buf); break;
        default:         assert(0);
        }
        break;

    case BORG_STR:
        switch (expval.type)
        {
        case BORG_INT:   sprintf(expval.buf, "%d", expval.i); 
                         expval.i = strlen(expval.buf);
                         break;
        case BORG_FLOAT: sprintf(expval.buf, "%f", expval.f); 
                         expval.i = strlen(expval.buf);
                         break;
        case BORG_STR:   break;
        default:         assert(0);
        }
        break;

    default:
        assert(0);
    }

    expval.type = outtype;

    return;
}


void C_OLAPProjectList::WriteTree(ostream& fs)
{
  C_OLAPExprNode::WriteTree(fs);
  fs << "(C_OLAPProjectList ";
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->WriteTree(fs);
  
  fs << ")";
}

void C_OLAPProjectList::ReadTree(istream& fs)
{
  C_OLAPExprNode::ReadTree(fs);
  char buf[32];
  fs >> buf;
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->ReadTree(fs);
  
  fs >> buf;
}


/* ------------------- Eval: C_OLAPValue --------------------- */

bool C_OLAPValue::Eval(ExprInfo& /*einfo*/, ExprValue& expval)
{
    expval.type = type;
    switch(type)
    {
    case BORG_INT:    expval.i = int_val;   break;
    case BORG_FLOAT:  expval.f = float_val; break;
    case BORG_STR:    strcpy(expval.buf, str_val.c_str());
                      expval.i = strlen(expval.buf);
                      break;
    default:          return false;
    }

    return true;
}

void C_OLAPValue::ReadTree(istream& fs)
{
  char str[32];
  int32 tint;
  fs >> str >> str >> str >> str >> tint;
  type = BorgTypes(tint);
  if (type == BORG_INT)   fs >> int_val;
  if (type == BORG_FLOAT) fs >> float_val;
  if (type == BORG_STR)   fs >> str_val;
  fs >> str;
}

void C_OLAPValue::WriteTree(ostream& fs)
{ 
  fs << "( C_OLAPValue, Type = " << int(type) << " ";
  if (type == BORG_INT)   fs << int_val;
  if (type == BORG_FLOAT) fs << float_val;
  if (type == BORG_STR)   fs << str_val;
  fs << ")\n";
}
