/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_attr.h
//
// This file contains the interface defintion for attributes, values
// and project list for query evaluation.
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////


#if !defined(__OLAP_ATTR_H)
#define __OLAP_ATTR_H

#include "olap_expr_abstract.h"

/* -------------------------------------------------------------- */ 
/* -- Node types:                                              -- */
/* --  o C_OLAPAttr     - an attribute node (dimension name)   -- */
/* --  o C_OLAPProjList - the projection list for this query   -- */
/* --  o C_OLAPValue    - the value that the dimension is equal to*/
/* -------------------------------------------------------------- */ 

// C_OLAPAttr store properties of an attribute such as
// logical attribute name, physical attribute name,
// dimension table name, attribute type

/* ---------------------- C_OLAPAttr ---------------------------- */

class C_OLAPAttr : public C_OLAPExprNode {

public:
    // Information specific to this node
    ColumnIdentifier ColID;
    BorgTypes AttrType;	// Type

    /* --------------------------------------------------------- */
    // Constructor and destructor
    C_OLAPAttr(ColumnIdentifier& colid, BorgTypes t) : C_OLAPExprNode(Attribute)
	{ ColID = colid; AttrType = t; }

    C_OLAPAttr() : C_OLAPExprNode(Attribute)
	{ AttrType = BORG_UNDEF; }

    ~C_OLAPAttr()
	{}

    C_OLAPAttr&   operator=(C_OLAPAttr &rhs)
	{ ColID = rhs.ColID; AttrType = rhs.AttrType; return *this; }

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    // It returns false on error...
    bool Eval(ExprInfo& einfo, ExprValue& expval);

    /* ------------------------------------------------------ */
    // Function which dumps the tree to/from a file 'fs'
    void WriteTree(ostream& fs);
    void ReadTree(istream& fs);
};



/* ------------------- C_OLAPProjectList ------------------- */
// A Project list which consists of a set of attributes

class C_OLAPProjectList : public C_OLAPExprNode {
private:
    // Information specific to this node

public:
    // Functions to access this node
    C_OLAPProjectList() : C_OLAPExprNode(ProjList) { }
    void AddAttr(pC_OLAPExprNode pqn) { AddArg(pqn); }

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    bool Eval(ExprInfo& einfo, ExprValue& expval);
    void MakeType(BorgTypes outtype, ExprValue& expval);

    /* ------------------------------------------------------ */
    // Function which dumps the tree to/from a file 'fs'
    void WriteTree(ostream& fs);    
    void ReadTree(istream& fs);    
};





/* -------------------- C_OLAPValue ----------------------- */

// A value which is a string storing the actual value. The
// type of the value is also stored

class C_OLAPValue : public C_OLAPExprNode {
public:
    // the type enums are defined in borgtypes.h
    // enum BorgTypes BORG_UNDEF, BORG_INT, BORG_FLOAT, BORG_STR };

public:
    // Information specific to this node
    string    str_val;              // The value is stored as a string
    int32     int_val;
    float     float_val;

    BorgTypes type;                 // used to determine which of the above variables should be used

public:
    C_OLAPValue(string val, BorgTypes t)  : C_OLAPExprNode(Value)
	{ str_val = val; type = t; }
    C_OLAPValue(int    i_val, BorgTypes t)  : C_OLAPExprNode(Value)
	{ int_val = i_val; type = t; }
    C_OLAPValue(float  f_val, BorgTypes t)  : C_OLAPExprNode(Value)
	{ float_val = f_val; type = t; }
    C_OLAPValue()  : C_OLAPExprNode(Value)
	{ type = BORG_UNDEF; }

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    // It returns false on error...
    bool Eval(ExprInfo& einfo, ExprValue& expval);

    /* --------------------------------------------------- */
    // Function which dumps the tree to/from a file 'fs'
    void WriteTree(ostream& fs);
    void ReadTree(istream& fs);
    
};

#endif
