/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_expr_abstract.cpp
//
// This file contains some of the interface implementation of the 
// abstract expression class.
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////

#include "olap_expr_abstract.h"
#include <stdio.h>    // sprintf
#include <stdlib.h>   // atoi etc...

#include "olap_expr.h"
#include "olap_attr.h"


C_OLAP_GeneralExpr::~C_OLAP_GeneralExpr()
{
	delete pen;
}

bool C_OLAP_GeneralExpr::evaluate(ExprInfo* pei)
{
    ExprValue ev;
    pen->Eval(*pei, ev);
    return ev.b;
}


void ExprInfo::Init( u_int32 numInputStreams, u_int32 numOutputStreams,
			    TupleTypes &ipTypes, TupleTypes &opTypes,
			    C_OLAP_Tuple *ipTuples, C_OLAP_Tuple *opTuples,
			    bool isProject)
{
    tInfo[ColumnIdentifier::left]   = NULL;
    tInfo[ColumnIdentifier::right]  = NULL;
    tInfo[ColumnIdentifier::output] = NULL;

    tup[ColumnIdentifier::left]     = NULL;
    tup[ColumnIdentifier::right]    = NULL;
    tup[ColumnIdentifier::output]   = NULL;

    // Assume that the tuple info structures has been setup.
    w_assert3(numInputStreams <= 2); w_assert3(numOutputStreams <= 1);

    for ( u_int32 i = 0 ; i < numInputStreams; i++)
    {
        tInfo[i] = ipTypes[i];
        tup[i]   = &ipTuples[i];
	tup[i]->Init(*tInfo[i]);
    }
  
    if ( isProject )
    {
 	for ( u_int32 j = 0 ; j < numOutputStreams ; j++ )
	{
           tInfo[ColumnIdentifier::output] = opTypes[j];
           tup[ColumnIdentifier::output]   = &opTuples[j];
	   tup[ColumnIdentifier::output]->Init(*tInfo[ColumnIdentifier::output]);
	}
    }
    else    // Left tuple is the output by default
    {
        tInfo[ColumnIdentifier::output] = tInfo[ColumnIdentifier::left];
        tup[ColumnIdentifier::output]   = tup[ColumnIdentifier::left];
	tup[ColumnIdentifier::output]->Init(*tInfo[ColumnIdentifier::left]);
    }
}


/* ------------------------------------------------------------------ */
// Initialization for the base class...
/* ------------------------------------------------------------------ */

// darray::enlarge returns the new size of the array
int C_OLAPExprNode::darray::enlarge(void)
{
    pC_OLAPExprNode* temp = new pC_OLAPExprNode [_size+increment];
    int i;
    for (i = 0; i < _size; i++)
        temp[i] = _arr[i];
    for (i = _size; i < _size+increment; i++)
	temp[i] = NULL;
    delete [] _arr;
    _arr = temp;
    _size += increment;
    return _size;
}


void C_OLAPExprNode::darray::set_size(int new_size)
{
    // Set size
    if (_size < new_size)
    {
        pC_OLAPExprNode* temp = new pC_OLAPExprNode [new_size];
	int i;
        for (i = 0; i < _size; i++)
            temp[i] = _arr[i];
	for (i = _size; i < new_size - _size; i++)
	    temp[i] = NULL;

        delete [] _arr;
        _arr = temp;
        _size  = new_size;
    }

    return;
}

/* ------------------------------------------------------------------ */

/* ------------------------------------------------------------------ */

// Add an argument to this Query tree node
bool C_OLAPExprNode::AddArg(pC_OLAPExprNode pqn)
{
    if (_nargs == _args.size())    // Array full?
        _args.enlarge();

    _args[_nargs] = pqn;
    _nargs++;

    return true;
}


// Add an argument to this Query tree node
bool C_OLAPExprNode::RemoveArg(int argnum)
{
    if (_nargs >= _args.size())    // Invalid argument number
        return false;

    for (int i = argnum; i < _nargs-1; i++)
       _args[i] = _args[i+1];

    _nargs--;
    return true;
}



void C_OLAPExprNode::ReadTree(istream& fs)
{
    char buf[32];
    int32 temp, val;
    fs >> buf >> buf >> buf >> temp >> buf >> buf >> buf >> _nargs >> buf;
    _opcode = opcode_t(temp);
    _args.set_size(_nargs);
    for (int i = 0; i < _nargs; i++)
    {
        fs >> val;              // Unpack the opcode
        // Create node of the appropriate type
        switch (val)
        {
        case LogicalExpr:  arg(i) = new C_OLAPLogicalExpr; break;
        case ArithExpr:    arg(i) = new C_OLAPArithExpr;   break;
        case CompExpr:     arg(i) = new C_OLAPCompExpr;    break;
        case Attribute:    arg(i) = new C_OLAPAttr;        break;
        case ProjList:     arg(i) = new C_OLAPProjectList; break;
        case Value:        arg(i) = new C_OLAPValue;       break;
        default:           assert(0);        // Should not happen
        }
    }
}


void C_OLAPExprNode::WriteTree(ostream& fs)
{
  fs << "(C_OLAPExprNode (opcode = " << int(_opcode) << "), (nargs =  " << nargs() << ") ";
  
  // Pack the types of the arguments
  for (int i = 0; i < nargs(); i++)
    fs << "(arg(" << i << ")->opcode " << int(arg(i)->opcode()) << ") ";
  
  fs << ")\n";
}
