/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// test_pred.cpp
//
// This file contains test code for testing predicates.  
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <borgtypes.h>
#include <olap_tuple.h>
#include <strings.h>
#include <olap_operator_common.h>

main()
{
    /* ---- Tuple creation begin... ----- */

    C_OLAP_Tuple     tuple1, tuple2, otuple;
    C_OLAP_TupleInfo tupInfo1, tupInfo2, otupInfo;
    C_OLAP_TupleIter tupIter1(tupInfo1, tuple1), 
	             tupIter2(tupInfo2, tuple2),
	             otupIter(otupInfo, otuple);

    tupInfo1.SetNumAttrs(3);
    tupInfo1.SetAttrDesc(0, BORG_INT);
    tupInfo1.SetAttrDesc(1, BORG_FLOAT);
    tupInfo1.SetAttrDesc(2, BORG_STR);
    tupInfo1.EndAttrDesc();

    tupInfo2.SetNumAttrs(3);
    tupInfo2.SetAttrDesc(0, BORG_INT);
    tupInfo2.SetAttrDesc(1, BORG_FLOAT);
    tupInfo2.SetAttrDesc(2, BORG_STR);
    tupInfo2.EndAttrDesc();

    otupInfo.SetNumAttrs(4);
    otupInfo.SetAttrDesc(0, BORG_INT);
    otupInfo.SetAttrDesc(1, BORG_FLOAT);
    otupInfo.SetAttrDesc(2, BORG_INT);
    otupInfo.SetAttrDesc(3, BORG_STR);
    otupInfo.EndAttrDesc();

    cout << "Fixed Length of tuple = " << tupInfo1.GetFixedLen() << endl;

    int i, j, ival;
    u_int32 len;
    float fval;
    char sval[50];
    char buf1[30], buf2[30];
    char obuf[40];

    /* this step is optional. the buffer can be provided externally or the
     * tuple will internally allocate memory
     */
    tuple1.SetTuple(buf1, 30);
    tuple2.SetTuple(buf2, 30);
    otuple.SetTuple(obuf, 40);

    /* this step is essential for tuples with variable length attributes */
    tuple1.SetFixedLen(tupInfo1.GetFixedLen());
    tuple2.SetFixedLen(tupInfo2.GetFixedLen());
    otuple.SetFixedLen(otupInfo.GetFixedLen());

    /* ---- Tuple creation complete ----- */
    /* ---------------------------------- */

    // Create an predicate expression tree...
    C_OLAPAttr * a1 = new C_OLAPAttr;
    a1->ColID.strm = ColumnIdentifier::left;
    a1->ColID.colNum = 0;
    a1->AttrType = BORG_INT; 

    C_OLAPValue* v = new C_OLAPValue;
    v->int_val = 0;
    v->type    = BORG_INT;

    C_OLAPAttr * a11 = new C_OLAPAttr;
    a11->ColID.strm = ColumnIdentifier::left;
    a11->ColID.colNum = 0;
    a11->AttrType = BORG_INT; 

    C_OLAPValue* v11 = new C_OLAPValue;
    v11->int_val = 0;
    v11->type    = BORG_INT;

    C_OLAPAttr * a2 = new C_OLAPAttr;
    a2->ColID.strm = ColumnIdentifier::right;
    a2->ColID.colNum = 0;
    a2->AttrType = BORG_INT; 

    C_OLAPArithExpr* ae = new C_OLAPArithExpr;
    ae->aetype() = C_OLAPArithExpr::Plus;
    ae->AddArg(a2);
    ae->AddArg(v);

    C_OLAPCompExpr * ce = new C_OLAPCompExpr;
    ce->cetype() = C_OLAPCompExpr::Lt;
    ce->AddArg(a1);
    ce->AddArg(ae);

    C_OLAPCompExpr* ce2 = new C_OLAPCompExpr;
    ce2->cetype() = C_OLAPCompExpr::Eq;
    ce2->AddArg(a11);
    ce2->AddArg(v11);

    C_OLAPLogicalExpr * le = new C_OLAPLogicalExpr;
    le->letype() = C_OLAPLogicalExpr::And;
    le->AddArg(ce);
    le->AddArg(ce2);

    // Setup the expression info structure
    ExprInfo einfo;
    einfo.tInfo[0] = &tupInfo1;
    einfo.tInfo[1] = &tupInfo2;
    einfo.tup[0] = &tuple1;
    einfo.tup[1] = &tuple2;
    einfo.tIter[0] = &tupIter1;
    einfo.tIter[1] = &tupIter2;

    einfo.tInfo[2] = &otupInfo;
    einfo.tup[2] = &otuple;
    einfo.tIter[2] = &otupIter;

    // Setup the projection list
    C_OLAPAttr *ja1, *ja2, *ja3, *ja4;
    ja1 = new C_OLAPAttr;
    ja1->ColID.strm = ColumnIdentifier::left;
    ja1->ColID.colNum = 0;
    ja1->AttrType = BORG_INT; 

    ja2 = new C_OLAPAttr;
    ja2->ColID.strm = ColumnIdentifier::right;
    ja2->ColID.colNum = 0;
    ja2->AttrType = BORG_FLOAT; 

    ja3 = new C_OLAPAttr;
    ja3->ColID.strm = ColumnIdentifier::left;
    ja3->ColID.colNum = 1;
    ja3->AttrType = BORG_INT; 

    ja4 = new C_OLAPAttr;
    ja4->ColID.strm = ColumnIdentifier::right;
    ja4->ColID.colNum = 2;
    ja4->AttrType = BORG_STR; 

    C_OLAPProjectList* jl = new C_OLAPProjectList;
    jl->AddArg(ja1);
    jl->AddArg(ja2);
    jl->AddArg(ja3);
    jl->AddArg(ja4);

    ExprValue expval;
    int jc1, jc2;

    for (i = 0; i < 30; i++)
    {
	jc1 = i % 2;
	jc2 = i % 3;

	ival = jc1;
	fval = i;
	for (j=0; j<i; j++)
	    sval[j] = i + '0';
	sval[j] = '\0';

	/* this step is essential for reusing tuples with 
	   variable length attributes */
	tuple1.Reset();

	tuple1.FillAttr(tupInfo1[0], (char *)&jc1);
	tuple1.FillAttr(tupInfo1[1], (char *)&fval);
	tuple1.FillAttr(tupInfo1[2], sval, strlen(sval) + 1);

	tuple2.Reset();

	tuple2.FillAttr(tupInfo2[0], (char *)&jc2);
	tuple2.FillAttr(tupInfo2[1], (char *)&fval);
	tuple2.FillAttr(tupInfo2[2], sval, strlen(sval) + 1);

	cout << " Tuple (1) = (" 
	     << *(int *)tupIter1.GetAttr(0, len) << ", " 
	     << *(float *)tupIter1.GetAttr(1, len) << ", " 
	     << (char *)tupIter1.GetAttr(2, len) << ")\t";
//	cout << " Variable size = " << len << "\t";

	cout << " Tuple (2) = ("
	     << *(int *)tupIter2.GetAttr(0, len) << ", " 
	     << *(float *)tupIter2.GetAttr(1, len) << ", " 
	     << (char *)tupIter2.GetAttr(2, len) << ")";
//	cout << " Variable size = " << len << '\t';


	// Check if these tuples satisfy the predicates...
	le->Eval(einfo, expval);
	if (expval.b)
        {
            cout << " ---- ";
            jl->Eval(einfo, expval);

            // Print the projected tuple...

            cout << "Output Tuple = (" 
                 << *(int *) otupIter.GetAttr(0, len) << ", " 
                 << *(float *) otupIter.GetAttr(1, len) << ", " 
                 << *(int *) otupIter.GetAttr(2, len) << ", " 
                 << (char *) otupIter.GetAttr(3, len) << ")\n";
//            cout << " Variable size = " << len << "\t";

        }
        cout << endl;
    }

    delete le;
    delete jl;
}
