/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_bkend_request.cpp
//
// This file defines the implementation for bkend request queues,
// allocation tables.  
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_bkend_request.h"
#include "olap_client_id.h"

#include "sfile.h"

C_OLAPBkend_Sql_Request::C_OLAPBkend_Sql_Request( 
			const w_string&		database, 
			const w_string& 		sql, 
			C_OLAPClient_Id*	client_id,
			C_OLAP_TupleInfo	*tupInfo)
{
	m_isDatabaseName = database ; m_isSql = sql ;
	m_isClientId = client_id ; 
	m_tupInfo = tupInfo ; 
}

ostream& operator << (ostream& os, const C_OLAPBkend_Sql_Request& request)
{
	os << *request.m_isClientId << " " 
	   << request.m_isDatabaseName << " " 
	   << request.m_isSql ;

	return os ; 
}

w_rc_t C_OLAPBkend_WrkrInfo::SubmitBkendRequest( C_OLAPBkend_Request* request ) 
{
        int32 nbytes = 0;
        W_DO( m_isSocket->write((char *)&request, sizeof(request), nbytes) ) ;

	return RCOK ; 
}

C_OLAPBkend_DbWrkrThreads::~C_OLAPBkend_DbWrkrThreads()
{
	for (u_int32 i = 0; i<m_isInfo.getsize(); i++)
		delete m_isInfo[i];
}

w_rc_t C_OLAPBkend_DbWrkrThreads::GetFreeWorker( C_OLAPBkend_WrkrInfo*& info )
{
	W_DO( m_isMutex.acquire() ) ;
	int32 freeWrkr ; 

	while ( (freeWrkr = m_isAllocBitmap.first_clr(0)) < 0 )	
		W_DO( m_isCond.wait( m_isMutex ) ) ; 
	
	m_isAllocBitmap.set(freeWrkr) ; 
	m_isInfo[freeWrkr]->m_iPos = freeWrkr ; 
	
        info = m_isInfo[freeWrkr] ; 
	m_isMutex.release() ; 

	return RCOK ; 
}

w_rc_t C_OLAPBkend_DbWrkrThreads::ReleaseWorker( C_OLAPBkend_WrkrInfo*& info )
{
	W_DO( m_isMutex.acquire() ) ;
	m_isAllocBitmap.clr(info->m_iPos) ;
	m_isMutex.release() ;

	m_isCond.signal() ; 

	return RCOK ; 
}

C_OLAPBkend_WrkrThreads::~C_OLAPBkend_WrkrThreads() 
{
	while( m_isMap.num_members() > 0 )
	{
		w_hash_i<C_OLAPBkend_DbWrkrThreads, w_string> iter(m_isMap);

		C_OLAPBkend_DbWrkrThreads* db_threads = iter.next() ; 
		
		m_isMap.remove(db_threads);

		delete db_threads ; 
	}
}


w_rc_t C_OLAPBkend_WrkrThreads::GetWorkerList(
			const w_string&			database,
			C_OLAPBkend_DbWrkrThreads*&     dbthreads )
{
	W_DO( m_isMutex.acquire() ) ; 
	dbthreads = m_isMap.lookup(database);

	if ( !dbthreads )
		return RCOK ; 	// TO DO : return the appropriate error

	m_isMutex.release() ; 

	return RCOK ; 
}

w_rc_t C_OLAPBkend_WrkrThreads::GetFreeWorker( 
			const w_string& 		database, 
			C_OLAPBkend_WrkrInfo*&	info ) 
{
	C_OLAPBkend_DbWrkrThreads* dbthreads = NULL ; 

	W_DO( m_isMutex.acquire() ) ; 
	dbthreads = m_isMap.lookup(database) ;  

	if ( dbthreads == NULL )
	{
		m_isMutex.release() ; 
		return RCOK ; // TO DO: return the appropriate error
	}

	m_isMutex.release() ; 

	W_DO( dbthreads->GetFreeWorker(info) ) ; 
	return RCOK ; 
}

w_rc_t C_OLAPBkend_WrkrThreads::ReleaseWorker( 
			const w_string&		database, 
			C_OLAPBkend_WrkrInfo*& 	info ) 
{
	C_OLAPBkend_DbWrkrThreads* dbthreads = NULL ; 
		
	W_DO( m_isMutex.acquire() ) ; 
	dbthreads = m_isMap.lookup(database) ;  
	m_isMutex.release() ; 

	if ( dbthreads )
		W_DO( dbthreads->ReleaseWorker( info ) ) ; 

	return RCOK ; 
}

#ifdef __GNUG__
template class dynarray<C_OLAPBkend_WrkrInfo*>;
#endif

#ifdef __GNUG__
template class w_hash_t<C_OLAPBkend_DbWrkrThreads, w_string>;
template class w_hash_i<C_OLAPBkend_DbWrkrThreads, w_string>;
template class w_list_t<C_OLAPBkend_DbWrkrThreads>;
template class w_list_i <C_OLAPBkend_DbWrkrThreads>;
#endif
