/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_bkend_request.h
//
// This file defines the interface for sql and shutdown request. 
// It also defines the allocation structure for the backend worker 
// threads. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_BKEND_REQUEST_H)
#define	__OLAP_BKEND_REQUEST_H

#include "w.h"
#include "w_dynarray.h"
#include "w_bitmap.h"
#include "w_string.h"

#include "olap_server_globals.h"

#include "olap_types.h"

class C_OLAPBkend_Queue ; 
class C_OLAPClient_Id ;
class C_OLAPBkend_Thread ;
class C_OLAP_TupleInfo ;
class sfile_t ; 

class C_OLAPBkend_Request
{
public:

	enum Olap_Bkend_Request_Type
	{
		OLAP_BKEND_SQL,
		OLAP_BKEND_SHUTDOWN
	} ; 

	C_OLAPBkend_Request()
	{ } ;

	virtual ~C_OLAPBkend_Request() 
	{ } 

	virtual Olap_Bkend_Request_Type RequestType() = 0 ;

} ; 
	

class C_OLAPBkend_Sql_Request : public C_OLAPBkend_Request
{
public :

	C_OLAPBkend_Sql_Request() 
	{ } 
		
	C_OLAPBkend_Sql_Request( 
			const w_string&		database, 
			const w_string&		sql, 
			C_OLAPClient_Id*	client_id,
			C_OLAP_TupleInfo*	tupInfo=NULL ) ; 

	virtual ~C_OLAPBkend_Sql_Request()
	{ } 

	virtual Olap_Bkend_Request_Type RequestType() 
	{
		return OLAP_BKEND_SQL ; 
	}

	w_string& GetDbName()
	{
		return m_isDatabaseName ;
	}
 
	w_string& GetSql()
	{
		return m_isSql ; 
	}

	C_OLAP_TupleInfo& GetTupInfo()
	{
		return *m_tupInfo ; 
	}

	C_OLAPClient_Id* GetClientId()
	{
		return m_isClientId ; 
	}

	friend ostream&
	operator << ( ostream& os, const C_OLAPBkend_Sql_Request& request ) ; 

protected :

	w_string			m_isDatabaseName ;
	w_string			m_isSql ; 		
	C_OLAP_TupleInfo	*m_tupInfo ; 
	C_OLAPClient_Id*	m_isClientId ; 

	friend class C_OLAPBkend_DbWrkrThreads ; 
	friend class C_OLAPBkend_WrkrThreads ; 
} ;

class C_OLAPBkend_Shutdown_Request : public C_OLAPBkend_Request
{
public:

	C_OLAPBkend_Shutdown_Request() 
	{ } 

	virtual ~C_OLAPBkend_Shutdown_Request()
	{ } 

	virtual Olap_Bkend_Request_Type RequestType() 
	{
		return OLAP_BKEND_SHUTDOWN ; 
	}
} ; 

class C_OLAPBkend_WrkrInfo
{
public:
	
	C_OLAPBkend_WrkrInfo()
	{
		m_isSocket = NULL ; m_iPos = -1 ; 
	}

	~C_OLAPBkend_WrkrInfo()
	{ } 

	w_rc_t SubmitBkendRequest( C_OLAPBkend_Request* request) ; 

	sfile_t*		m_isSocket ; 

protected:

	int32			m_iPos ;
	friend class C_OLAPBkend_DbWrkrThreads ; 
} ; 

typedef	dynarray<C_OLAPBkend_WrkrInfo*>	WorkerArray; 

class C_OLAPBkend_DbWrkrThreads
{
public:

	enum {
		max_db_threads = 1 
	} ; 

	C_OLAPBkend_DbWrkrThreads() : m_isAllocBitmap(max_db_threads), 
				m_isInfo(max_db_threads) 
        { }

	~C_OLAPBkend_DbWrkrThreads() ; 

	w_rc_t GetFreeWorker( C_OLAPBkend_WrkrInfo*& info ) ; 
	w_rc_t ReleaseWorker( C_OLAPBkend_WrkrInfo*& info ) ;
		
protected :

	smutex_t				m_isMutex ;
	scond_t					m_isCond ;
	w_bitmap_t				m_isAllocBitmap ;
	WorkerArray				m_isInfo ; 
	w_link_t				m_isLink ;

private :

	w_string					m_isDatabase ;

	friend class C_OLAPBkend_WrkrThreads ;
	friend class C_OLAPMain_Thread ; 
} ; 

class C_OLAPBkend_WrkrThreads
{
	enum {
		max_databases = 32
	} ;

public:
	C_OLAPBkend_WrkrThreads() : 
		m_isMap(max_databases, 
			offsetof(C_OLAPBkend_DbWrkrThreads, m_isDatabase),
			offsetof(C_OLAPBkend_DbWrkrThreads, m_isLink))
	{ } 

	~C_OLAPBkend_WrkrThreads() ;
		

	w_rc_t GetWorkerList( 
		const w_string&			database,
		C_OLAPBkend_DbWrkrThreads*& 	dbthreads ) ; 

	w_rc_t GetFreeWorker( 
		const w_string&		database, 
		C_OLAPBkend_WrkrInfo*&	info ) ;

	w_rc_t ReleaseWorker( 
		const w_string&		database,
		C_OLAPBkend_WrkrInfo*&	info ) ;
			
protected:

	smutex_t					m_isMutex ; 	
	w_hash_t<C_OLAPBkend_DbWrkrThreads, w_string>	m_isMap ; 

	friend class C_OLAPMain_Thread ; 
} ; 

#endif
