/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_client_thread.h
//
// This file defines the service and worker threads for clients on  
// BORG server 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_CLIENT_THREAD_H)
#define	__OLAP_CLIENT_THREAD_H

#include "olap_borg_thread.h"
#include "olap_client_id.h"

#include "olap_server_globals.h"
#include "sfile.h"

#include "olap_comm_shore.h"

class C_OLAPClient_Thread ; 

class C_OLAP_Clients
{
private:

 	class CListElem
	{
	public :

		C_OLAPClient_Id		m_isClientId ;
		bool			m_bActive ; 
		w_link_t		m_isLink ; 
		C_OLAPClient_Thread*	m_isThread ;

		CListElem()
		{ 
			m_bActive = false ; m_isThread = NULL ; 
		} 
	} ;	 

	smutex_t			m_isMutex ;	
	w_list_t<CListElem>		m_isList ;				

public:

	C_OLAP_Clients() : m_isList(offsetof(CListElem,m_isLink))
	{
	}

	~C_OLAP_Clients() ;

	rc_t insert( const C_OLAPClient_Id& client_id, 
		     C_OLAPClient_Thread* client_thread ) ;

	rc_t remove( const C_OLAPClient_Id& client_id ) ; 
	
	rc_t GetActiveClients( C_OLAPComm_DataLinearizer& lin ) ; 

	rc_t RemoveInactiveClients() ; 

	rc_t WaitForActiveClientsFinish() ; 
} ;

class C_OLAPClient_Service_Thread : public C_OLAPBorg_Thread
{
public :

	C_OLAPClient_Service_Thread(sfile_t* fd) : 
		 m_isListener(fd)
	{ }

	~C_OLAPClient_Service_Thread()
	{ } 

	virtual void run() ;

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_CLIENT_SERVICE_THREAD ;
	}


protected :

	sfile_t*			m_isListener ; 
} ; 

class C_OLAPClient_Thread : public C_OLAPBorg_Thread
{
public :

	C_OLAPClient_Thread() 
	{ }

	C_OLAPClient_Thread( sfile_t* sock, C_OLAPClient_Id client_id ) : 
			m_isClientId(client_id), m_isClientSock(sock) 
	{ } 
 
	~C_OLAPClient_Thread() ;

	virtual void run() ; 

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_CLIENT_THREAD ;
	}

	C_OLAPClient_Id& GetClientId()
	{
		return m_isClientId ; 
	}

protected :

	C_OLAPClient_Id		m_isClientId ; 
	sfile_t*		m_isClientSock ; 
} ;

#endif
