/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_main_thread.cpp
//
// This file contains the implementation of the main thread for
// initializing the server and starting the appropriate services. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_macros.h"
#include "w_dynarray.h"

#include "olap_bkend.h"
#include "olap_bkend_request.h"

#include "olap_main_thread.h"
#include "olap_client_thread.h"
#include "olap_msg_thread.h"
#include "olap_operator.h"

extern	option_t*				dev_opt ;
extern	option_t*				quota_opt ; 
extern	bool					borg_format ; 

extern  C_OLAP_Clients				g_Clients ; 
extern	C_OLAPOperator_Queue			g_OperatorExecQueue ;
extern	C_OLAPMsg_Router_Thread*        	g_RouterThread ;
extern	C_OLAPMsg_Inquiry_Thread*		g_MsgInquiryThread ; 

C_OLAPBkend_WrkrThreads				g_bkend_thread_alloc ; 

smutex_t					g_startMutex ;
scond_t						g_startCond ;


C_OLAPMain_Thread::C_OLAPMain_Thread()
{ 
	m_isMsgSvcThread = NULL ; m_isClientSvcThread = NULL ; 
}


C_OLAPMain_Thread::~C_OLAPMain_Thread()
{ 
	if ( m_isMsgSvcThread ) delete m_isMsgSvcThread ; 
	if ( m_isClientSvcThread ) delete m_isClientSvcThread ; 

	for ( u_int32 i = 0 ; i < m_isOpThreads.getsize() ; i++ )
		delete m_isOpThreads[i] ; 

	for ( u_int32 i = 0 ; i < m_isBackendThreads.getsize() ; i++ )
		delete m_isBackendThreads[i] ; 

	for ( u_int32 i = 0 ; i < m_isBackendSocks.getsize() ; i++ )
		delete m_isBackendSocks[i] ; 
}


void C_OLAPMain_Thread::run()
{
	// Create an instance of storage manager object
	sm = new ss_m() ; 
	if ( !sm ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	cerr << "Starting backend threads" ; 

	// Create a set of back end threads 
	for ( int schema = 0 ; schema < 1 ; schema++ )
	{ 
		C_OLAPBkend_DbWrkrThreads* sch_thread_info = 
				new C_OLAPBkend_DbWrkrThreads ; 
		if ( !sch_thread_info ) W_FATAL(BORG_OUTOFMEMORY) ; 

		for ( int i = 0 ; 
			i < C_OLAPBkend_DbWrkrThreads::max_db_threads ; i++ )
        	{
			C_OLAPBkend_Thread* bkend_thread = NULL ; 
                	bkend_thread = new C_OLAPBkend_Thread(sch_thread_info->m_isDatabase,i) ;
                	if ( !bkend_thread ) W_FATAL(BORG_OUTOFMEMORY);

			m_isBackendThreads.push_back( bkend_thread ) ; 
                	bkend_thread->Start() ;

			cerr << "." ; 

			SOCKET s = bkend_thread->Synchronize() ;  

			C_OLAPBkend_WrkrInfo* info = new C_OLAPBkend_WrkrInfo ;
			if ( !info ) W_FATAL(BORG_OUTOFMEMORY) ; 

			info->m_isSocket = new sfile_t(s, sfile_t::OPEN_RDWR) ;
			if ( !info->m_isSocket) W_FATAL(BORG_OUTOFMEMORY) ; 

			m_isBackendSocks.push_back( info->m_isSocket ) ; 

			sch_thread_info->m_isInfo.push_back(info) ; 
        	}
		
		g_bkend_thread_alloc.m_isMap.push( sch_thread_info ) ; 
	}

	cerr << "Done" << endl ; 

	cerr << "Starting operators" ; 

	C_OLAPOperator_Thread*     op_thread = NULL ;

	// Create a set of operator threads
	for ( int32 i = 0 ; i < C_OLAPOperator_Queue::max_requests / 2 ; i++ )
	{
		op_thread = new C_OLAPOperator_Thread() ; 
               	if ( !op_thread ) W_FATAL(BORG_OUTOFMEMORY);

		m_isOpThreads.push_back( op_thread ) ; 
 
		W_COERCE( op_thread->fork() ) ; 		

		cerr << "." ; 
	}
	cerr << "Done" << endl ; 

	// Initialize Shore
	cerr << "Initializing storage manager..." ; 
	W_COERCE(InitShore( borg_format, dev_opt, quota_opt )) ; 
	cerr << "Done" << endl ;

	cerr << "Starting msg service..." ;
	// Create the msg service thread
	m_isMsgSvcThread = new C_OLAPMsg_Service_Thread( BORG_Msg_Sock ) ; 
	if ( !m_isMsgSvcThread ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	W_COERCE( m_isMsgSvcThread->fork() ) ;
	cerr << "Done" << endl ; 

	cerr << "Starting client service... ";
	// Create the client service thread
	m_isClientSvcThread = new C_OLAPClient_Service_Thread( BORG_Clnt_Sock ) ;
	if ( !m_isClientSvcThread ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	W_COERCE( m_isClientSvcThread->fork() ) ;
	cerr << "Done" << endl ; 

	// Wait until the message service completed starting
	W_COERCE( g_startMutex.acquire() ) ; 
	W_COERCE( g_startCond.wait(g_startMutex) ) ; 
	g_startMutex.release() ; 

	cerr << "BORG is ready to process cubes!" << endl ;

	// Wait for shutting down signal
	W_COERCE( WaitForShutdown() ) ; 
	
	W_COERCE( ShutdownServer() ) ; 

	// clean up shore
	cerr << "Shutting down storage manager..." ; 
	W_COERCE(ShutdownShore()) ; 
	cerr << "Done" << endl ; 

	W_COERCE(sm->dismount_all());
	delete sm;
}

rc_t C_OLAPMain_Thread::ShutdownServer()
{
	cerr << "Shutting down server..." << endl ; 
	W_COERCE( WaitForRequestDenialShutdown() ) ; 
	W_COERCE( WaitForClientsShutdown() ) ; 
	W_COERCE( WaitForMsgClientsShutdown() ) ; 
	W_COERCE( WaitForOperatorShutdown() ) ; 
	W_COERCE( WaitForBackendShutdown() ) ; 
	W_COERCE( WaitForMsgSystemShutdown() ) ; 

	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForRequestDenialShutdown()
{
	cerr << "Shutting down services..." << endl ;   
	W_COERCE( BORG_Shutdown_Sock->close() ) ; 
	
	cerr << "Shutting down msg service..." ;   
	W_COERCE( BORG_Msg_Sock->cancel() ) ; 
	W_COERCE( BORG_Msg_Sock->close() ) ; 
	W_COERCE( 
		m_isMsgSvcThread->wait( C_OLAPMsg_Service_Thread::WAIT_FOREVER ) 
	) ; 
	cerr << "Done" << endl ; 

	cerr << "Shutting down client service..." ;   
	W_COERCE( BORG_Clnt_Sock->cancel() ) ; 
	W_COERCE( BORG_Clnt_Sock->close() ) ; 
	W_COERCE( 
		m_isClientSvcThread->wait( C_OLAPClient_Service_Thread::WAIT_FOREVER ) 	) ; 
	cerr << "Done" << endl ; 

	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForClientsShutdown()
{
	cerr << "Waiting for active clients to complete..." ; 
	W_COERCE( g_Clients.WaitForActiveClientsFinish() ) ; 

	W_COERCE( g_Clients.RemoveInactiveClients() ) ; 
	
	cerr << "Done" << endl ; 
	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForMsgClientsShutdown()
{
	cerr << "Waiting for msg active clients to complete..." ; 
	cerr << "Done" << endl ; 
	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForOperatorShutdown()
{
	cerr << "Shutting down operators" ; 

	dynarray<C_OLAPOperator_Shutdown*>	shutdown_operators ; 
	
	for ( int32 i = 0 ; i < C_OLAPOperator_Queue::max_requests / 2 ; i++ )
	{
		C_OLAPOperator_Shutdown* op_request = new C_OLAPOperator_Shutdown ; 
		if ( !op_request ) W_FATAL(BORG_OUTOFMEMORY) ; 

		shutdown_operators.push_back( op_request ) ;  
		W_COERCE( g_OperatorExecQueue.enqueue( op_request ) ) ; 
	}

	for ( u_int32 i = 0 ; i < m_isOpThreads.getsize() ; i++ )
	{
		W_IGNORE( m_isOpThreads[i]->wait(
				C_OLAPOperator_Thread::WAIT_FOREVER) ) ;
		cerr << "." ;
	}

	for ( u_int32 i = 0 ; i < shutdown_operators.getsize() ; i++ )
	{
		delete shutdown_operators[i] ; 
	}

	cerr << "Done" << endl ; 

	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForBackendShutdown()
{
	int32 nbytes = 0;
	C_OLAPBkend_Shutdown_Request*	shutdown_request ; 

	shutdown_request = new C_OLAPBkend_Shutdown_Request ; 
	if ( !shutdown_request ) W_FATAL( BORG_OUTOFMEMORY ) ; 

	cerr << "Shutting down backend threads" ; 

	for ( u_int32 i = 0 ; i < m_isBackendThreads.getsize() ; i++ )
	{
		// send the shutdown request to a particular thread
        	W_COERCE( m_isBackendSocks[i]->write((char *)&shutdown_request, 
			sizeof(shutdown_request), nbytes) ) ;

		// wait until the back end thread terminates
		m_isBackendThreads[i]->Wait() ; 

		// Now close the connection
		W_COERCE( m_isBackendSocks[i]->close() ) ;

		cerr << "." ; 
	}

	delete shutdown_request ; 
	cerr << "Done" << endl ; 

	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForMsgSystemShutdown() 
{
	cerr << "Shutting down msg system..." << endl ; 

	cerr << "Shutting down msg inquiry service..." ;
	W_COERCE( BORG_Msg_Inquiry_Sock->cancel() ) ; 
	W_COERCE( BORG_Msg_Inquiry_Sock->close() ) ; 
	W_COERCE( 
		g_MsgInquiryThread->wait( C_OLAPMsg_Inquiry_Thread::WAIT_FOREVER ) 
	) ; 
	delete g_MsgInquiryThread ; 
	cerr << "Done" << endl ;

	cerr << "Shutting down msg routing service..." ; 
	C_OLAPMsg		msg_item( C_OLAPMsg::OLAP_MSG_ALL_CLIENTS, 
					  C_OLAPMsg::OLAP_MSG_SERVER_EXIT,
					  C_OLAPMsg::OLAP_MSG_SERVER_FATAL, 
					  0, "Server shutdown" ) ; 

	W_COERCE( g_MsgQueue.enqueue( msg_item ) ) ; 
	W_IGNORE( g_RouterThread->wait( C_OLAPMsg_Router_Thread::WAIT_FOREVER ) ) ; 
	delete g_RouterThread ; 
	cerr << "Done" << endl ; 

	return RCOK ; 
}

rc_t C_OLAPMain_Thread::WaitForShutdown() 
{
	struct sockaddr_in shutdown_client ;
	int shutdown_client_len = sizeof(shutdown_client) ;
	memset( &shutdown_client, 0, shutdown_client_len ) ;

	sfile_t		shutdown_client_sock ;

	while (1)
	{
		// Wait on the shutdown port
        	w_rc_t rc = BORG_Shutdown_Sock->accept(
                            	(struct sockaddr *)&shutdown_client,
                            	&shutdown_client_len, shutdown_client_sock );

        	if ( rc ) ::exit(0) ; 

		C_OLAPComm_System_Shore	isServer(&shutdown_client_sock) ;
		C_OLAPComm_Reconciler	isReconciler ;
		C_OLAPComm_Packet*	isPkt = NULL ;

		if ( isServer.ReceiveData( isReconciler, isPkt ) == RCOK && 
		     isPkt && isPkt->GetPktType() == OLAPCOMM_SHUTDOWN_PACKET )
		{
			delete isPkt ; 
			W_COERCE( shutdown_client_sock.close() ) ; 
			return RCOK ; 
		}
	}

	W_COERCE( shutdown_client_sock.close() ) ; 
	return RCOK ;
}

rc_t C_OLAPMain_Thread::InitShore( bool format, option_t* device_opt, 
				  option_t* max_quota_opt)
{
	u_int vol_cnt = 1;
	devid_t devid;

	const char*  dev_name = device_opt->value();
	int quota = atoi(max_quota_opt->value());

	if (format) 
		W_DO(sm->format_dev(dev_name, quota, true));

	W_DO(sm->mount_dev(dev_name, vol_cnt, devid, 1));

	if ( format ) // first time, a vol needs to be created.
	{
		lvid_t lvid;
		W_DO(sm->generate_new_lvid(lvid));
		W_DO(sm->create_vol(dev_name, lvid, quota, true, BORG_VOLID));
	}

	lvid_t* lvid_list = 0;
	u_int lvid_cnt;
	W_DO(sm->list_volumes(dev_name, lvid_list, lvid_cnt));

	if (lvid_cnt != 1) return RC(smlevel_0::eBADVOL);
	W_DO(sm->lvid_to_vid(lvid_list[0], BORG_VOLID));
	delete [] lvid_list;

	return RCOK;
}

rc_t C_OLAPMain_Thread::ShutdownShore()
{
	return RCOK ; 
}

#ifdef __GNUG__
template class dynarray<C_OLAPBkend_Thread*>;
template class dynarray<C_OLAPOperator_Thread*> ;
template class dynarray<C_OLAPOperator_Shutdown*> ;
template class dynarray<sfile_t*>;
#endif

