/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_server.cpp
//
// This file contains the implementation of the main for the olap 
// server BORG.
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>

#include <iostream.h>
#include <string>

#include "olap_server_globals.h"
#include "olap_server_macros.h"
#include "olap_main_thread.h"
#include "olap_client_thread.h"

#include "olap_comm_shore.h"
#include "olap_bkend.h"

#include "w_dynarray.h"

sfile_t*	BORG_Msg_Inquiry_Sock = NULL ; 
sfile_t*	BORG_Msg_Sock = NULL ; 
sfile_t*	BORG_Clnt_Sock = NULL ; 
sfile_t*	BORG_Shutdown_Sock = NULL ; 

int32	BORG_MSG_LEVEL = 1 ; 

// Error codes for BORG
enum {  
	BORG_MIN_ERROR_CODE = 50000,
        BORG_OS,
        BORG_FAILURE,
        BORG_COMMAND_LINE,
        BORG_MAX_ERROR_CODE
};

w_error_info_t borg_error_list[]=
{
    { BORG_OS,     		"Operating System Error" },
    { BORG_FAILURE,		"General BORG Server Failure" },
    { BORG_COMMAND_LINE,	"Error while Processing Command Line" }
};

// Instance of storage manager
ss_m* 		sm = NULL;

option_t*	dev_opt = NULL ; 
option_t*	quota_opt = NULL ;
bool 		borg_format = false;
u_int32		OlapMsgInquiryPort = 0;
u_int32		OlapMsgServicePort = 0;
u_int32		OlapClientServicePort = 0;
u_int32		OlapServerShutdownPort = 0;
u_int32		OlapDbServerPort = 0;

vid_t		BORG_VOLID = vid_t::null ; 

C_OLAP_Clients	g_Clients ;

void SetPortNumbers()
{
	OlapMsgInquiryPort = OLAP_MSG_INQUIRY_PORT;
	OlapMsgServicePort = OLAP_MSG_SERVICE_PORT;
	OlapClientServicePort = OLAP_CLIENT_SERVICE_PORT;
	OlapServerShutdownPort = OLAP_SERVER_SHUTDOWN_PORT;
	OlapDbServerPort = OLAP_DB_SERVER_PORT;

	ifstream inp(".ports");
	if (!inp) return;

	inp >> OlapMsgInquiryPort ;
	while (inp.get() != '\n');
	inp >> OlapMsgServicePort ;
	while (inp.get() != '\n');
	inp >> OlapClientServicePort ;
	while (inp.get() != '\n');
	inp >> OlapServerShutdownPort ;
	while (inp.get() != '\n');
	inp >> OlapDbServerPort ;
}

void PrintUsage(ostream& isError, const char* pszImage,
                 bool long_form, option_group_t& options)
{
	if (!long_form) 
	{
		isError << "Usage: " << pszImage
			<< " [-h] [-f] [-v] " << endl;
	} 
	else 
	{
		isError << "Usage: " << pszImage << endl;
		isError << "switches:\n"
			<< "\t-h: print help message and exit\n"
			<< "\t-v: verbose printing of all option values\n"
			<< "\t-V: verbose printing of test results\n"
			<< "options:" << endl;
   	}

	options.print_usage(true, isError);
}

rc_t ProcessConfigurationFile( option_group_t& options )
{
	rc_t		rc ; 

	ostrstream      isErrors;
	string		isOptionFile = "./.borgsrc";

	option_file_scan_t opt_scan(isOptionFile.c_str(), &options);
	rc = opt_scan.scan(true, isErrors);
	isErrors << ends;

	char* pszErrorMsg = isErrors.str();

	if ( rc ) 
	{
		cerr 	<< "Error while reading option file: " 
			<< isOptionFile 
			<< "  " 
			<< endl;

		cerr	<< "\t" 
			<< pszErrorMsg 
			<< endl;

		rc = RCOK;
	}
	
	isErrors.rdbuf()->freeze(0) ; 
	
	return rc ; 
}

rc_t ParseCommandLine( int argc, char* argv[], 
		       option_group_t& options, string& isExecutableImage )
{
	rc_t		rc ; 

	ostrstream      isErrors;
	rc = options.parse_command_line((const char**)argv, argc, 2, &isErrors);
	isErrors << ends;

	char* pszErrorMsg = isErrors.str();

	if ( rc ) 
	{
		cerr << "Error on command line " << ".borgsrc" << endl;
		cerr << "\t" << w_error_t::error_string(rc.err_num()) << endl;
		cerr << "\t" << pszErrorMsg << endl;
		PrintUsage(cerr, isExecutableImage.c_str(), false, options);
	}

	isErrors.rdbuf()->freeze(0) ; 
	
	return rc ;
}

rc_t ParseBORGSpecificFlags( int argc, char* argv[], 
			     option_group_t& options, string isExecutableImage )
{
	int option; rc_t rc ; 
	bool verbose_opt = false; 

	while ((option = getopt(argc, argv, "fhvV:")) != -1) 
	{
		switch ( option ) 
		{
			case 'f':
				borg_format = true;
				break;

			case 'v':
				verbose_opt = true;
				break;

			case 'V':
				verbose_opt = true;
				break;

			case 'h':
            			PrintUsage(cerr, isExecutableImage.c_str(), true, options);
				rc = RCOK;
				exit(0);
				break;

			default:
				cerr	<< "Unrecognized command line flag: " 
					<< char(option) 
					<< endl;
            			rc = RC(BORG_COMMAND_LINE);
         	}
	}

	if (verbose_opt) options.print_values(false, cerr);
	return rc ;  
}

rc_t CheckRequiredOptions( option_group_t& options, string& isExecutableImage ) 
{
	ostrstream      isErrors ;

	rc_t rc = options.check_required(&isErrors);
	isErrors << ends;

	char* pszErrorMsg = isErrors.str();

	if ( rc ) 
	{
		cerr << "These required options are not set:" << endl;
		cerr << pszErrorMsg << endl;
		PrintUsage(cerr, isExecutableImage.c_str(), false, options);
	}

	isErrors.rdbuf()->freeze(0) ; 
	
	return rc ; 
}

main(int argc, char *argv[])
{
	string			isDiskRW ;
	string			isDeviceFile ;
	SOCKET			sock ; 

	if (!(w_error_t::insert("BORG Main", 
				borg_error_list, 
				BORG_MAX_ERROR_CODE - BORG_MIN_ERROR_CODE - 1)))
	{
		cerr << "Failure in initializing borg error handler" << endl;
		exit(1);
	}

	string isExecutableImage = argv[0] ; 

	string::size_type pos = isExecutableImage.rfind('/') ; 
	if ( pos != string::npos )
		isExecutableImage.erase( 0, pos+1 ) ;

	// Get the diskrw and device file paths
	string isPath = argv[0] ; 
	pos = isPath.rfind('/');
	if ( pos != string::npos )
		isPath.erase(pos+1, isPath.length()) ;
	else
		isPath = "./" ;

	isDiskRW = isDeviceFile = isPath ; 
	isDiskRW += "diskrw" ; isDeviceFile += ".devices" ; 

	option_group_t options(3);
	W_COERCE( options.add_class_level("shoreserver") );
	W_COERCE( options.add_class_level("borg") );
	W_COERCE( options.add_class_level(isExecutableImage.c_str()) );

	// Request the storage manager to add the options 
	W_COERCE(ss_m::setup_options(&options));

        // Location of diskrw 
        W_COERCE(options.set_value("sm_diskrw", false,
                        isDiskRW.c_str(), true, &cerr));

	// Add the device and quota 
	W_COERCE( options.add_option("sm_device", "device name", NULL,
				"Storage Device Path", true, 
				option_t::set_value_charstr, dev_opt) );
   
	W_COERCE( options.add_option("sm_dev_quota", ">#1000", "100000",
				"Maximum Storage Device Size", false, 
				option_t::set_value_charstr, quota_opt) );
   
	// Process the default configuration file .borgsrc
	rc_t rc = ProcessConfigurationFile( options ) ;

	// Process the BORG command line
	if ( !rc ) 
		rc = ParseCommandLine( argc, argv, options, isExecutableImage ) ;

	// Process BORG specific flags 
	if ( !rc ) 
		rc = ParseBORGSpecificFlags( argc, argv, options, isExecutableImage ) ;

	// Check the necessary options are valid
	if ( !rc )
		rc = CheckRequiredOptions( options, isExecutableImage ) ; 

	// Any problems so far, then exit
	if ( rc ) 
	{
		rc = RCOK; exit(-1);
   	}

	// Read port number from file
	SetPortNumbers();

	// Initialize the communication sub system
	W_COERCE( C_OLAPComm_System::Initialize() ) ; 

	// Create the msg client service socket
	sock = C_OLAPComm_System::CreateNetDomainSvrSocket( OlapMsgServicePort ) ; 
	if ( sock == INVALID_SOCKET )
	{
		cerr 	<< argv[0] << " " 
			<< "Unable to bind on olap msg service port" << endl ; 
		exit(1) ; 
	}
	
	BORG_Msg_Sock = new sfile_t( sock, sfile_t::OPEN_RDWR ) ; 
	if ( !BORG_Msg_Sock )	W_FATAL(BORG_OUTOFMEMORY) ; 
	
	// Create the client service socket 
	sock = C_OLAPComm_System::CreateNetDomainSvrSocket( OlapClientServicePort ); 
	if ( sock == INVALID_SOCKET )
	{
		cerr 	<< argv[0] << " " 
			<< "Unable to bind on olap client service port" << endl ; 

		W_COERCE( BORG_Msg_Sock->close() ) ; 
		exit(1) ; 
	}

	BORG_Clnt_Sock = new sfile_t( sock, sfile_t::OPEN_RDWR ) ; 
	if ( !BORG_Clnt_Sock )	W_FATAL(BORG_OUTOFMEMORY) ; 

	// Create the client service socket 
	sock = C_OLAPComm_System::CreateNetDomainSvrSocket( OlapMsgInquiryPort ); 

	if ( sock == INVALID_SOCKET )
	{
		cerr 	<< argv[0] << " " 
			<< "Unable to bind on olap msg inquiry port" << endl ; 

		W_COERCE( BORG_Msg_Sock->close() ) ; 
		W_COERCE( BORG_Clnt_Sock->close() ) ; 
		exit(1) ; 
	}

	BORG_Msg_Inquiry_Sock = new sfile_t( sock, sfile_t::OPEN_RDWR ) ; 
	if ( !BORG_Msg_Inquiry_Sock )	W_FATAL(BORG_OUTOFMEMORY) ; 

	// Create the client service socket 
	sock = C_OLAPComm_System::CreateNetDomainSvrSocket( OlapServerShutdownPort ); 

	if ( sock == INVALID_SOCKET )
	{
		cerr 	<< argv[0] << " " 
			<< "Unable to bind on olap msg inquiry port" << endl ; 

		W_COERCE( BORG_Msg_Sock->close() ) ; 
		W_COERCE( BORG_Clnt_Sock->close() ) ; 
		W_COERCE( BORG_Msg_Inquiry_Sock->close() ) ; 
		exit(1) ; 
	}

	BORG_Shutdown_Sock = new sfile_t( sock, sfile_t::OPEN_RDWR ) ; 
	if ( !BORG_Shutdown_Sock )	W_FATAL(BORG_OUTOFMEMORY) ; 

	C_OLAPMain_Thread* BORG_main_thread = new C_OLAPMain_Thread();
	if ( !BORG_main_thread ) W_FATAL(fcOUTOFMEMORY);

	W_COERCE( BORG_main_thread->fork() );
	W_COERCE( BORG_main_thread->wait(C_OLAPMain_Thread::WAIT_FOREVER) );
	delete BORG_main_thread ;

	delete BORG_Msg_Sock ; 
	delete BORG_Clnt_Sock ; 
	delete BORG_Msg_Inquiry_Sock ; 
	delete BORG_Shutdown_Sock ; 

	// Shutdown the communication system
	W_COERCE( C_OLAPComm_System::Shutdown() ) ; 
	
#if defined(_WINDOWS)
	return(0) ; 
#else
	exit(0);
#endif 
}
