/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

///////////////////////////////////////////////////////////////////
//
// olap_stream_abstract.h
//
// This file defines the abstract class for the streams. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_STREAM_ABSTRACT_H)
#define	__OLAP_STREAM_ABSTRACT_H

#include "w.h"
#include "olap_types.h"
#include "borgst_error_def_gen.h"

class C_OLAP_Tuple ; 
class C_OLAP_Page ; 

class C_OLAPStream {
public:

	enum OLAP_Stream_Constants {
		OLAP_MIN_CHUNKS = 8, 
		OLAP_MIN_PAGE_BUFFERS = 8 
	} ; 
		
	enum OLAP_Stream_Type {
		OLAP_TUPLE_STREAM 
	} ; 

	enum OLAP_Stream_Msg_Type {

		OLAP_REQUEST_MORE_DATA = 1, 
		OLAP_MORE_DATA,
		OLAP_END_OF_STREAM, 
		OLAP_ABNORMAL_TERMINATION,
		OLAP_EXCEPTION_ACKNOWLEDGE
	} ; 

	C_OLAPStream() 
	{ } 

	virtual ~C_OLAPStream()
	{ } 

	// Identify the type of stream
	virtual OLAP_Stream_Type TypeOfStream() = 0 ; 

	// Read a tuple from the stream
	virtual w_rc_t read( C_OLAP_Tuple& aTuple, bool& eos, bool& exception ) = 0 ; 

	// Write a tuple into the stream
	virtual w_rc_t write( C_OLAP_Tuple& aTuple, bool& exception ) = 0;

	// Read a tuple from the stream
	virtual w_rc_t read( C_OLAP_Page*& aPage, bool& eos, bool& exception ) = 0 ; 

	// Write a tuple into the stream
	virtual w_rc_t write( C_OLAP_Page* aPage, bool& exception ) = 0;

	// Send end of input in the stream
	virtual w_rc_t SendEndOfStream() = 0 ; 

	// Send abnormal termination down stream
	virtual w_rc_t SendExceptionDownStream() = 0 ; 

	// Send abnormal termination up stream
	virtual w_rc_t SendExceptionUpStream() = 0 ; 

	// Reset the stream for yet another operation
	virtual void Reset() = 0 ;  
} ; 

#endif 
