/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_memory_stream.h
//
// This file consists of the definition of the memory stream used
// to transport tuples from one operator to another operator. Uses
// synchronization mutexes between operators threads for control
// flow.
// 
// It is derived from the abstract stream class. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_MEMORY_STREAM_H)
#define	__OLAP_MEMORY_STREAM_H

#include "olap_server_globals.h"
#include "w_list.h"

#include "olap_stream_abstract.h"

class C_OLAPMemory_Stream : public C_OLAPStream {
protected:
	struct mem_stream_t {
		C_OLAPChunk		m_isChunk ; 
		OLAP_Stream_Msg_Type	m_isMsgType ; 
		w_link_t                m_isLink ;
	} ;
public:

	C_OLAPMemory_Stream( int32 nchunks = OLAP_MIN_CHUNKS ) : 
		m_isInputList(offsetof(mem_stream_t, m_isLink)),
		m_isOutputList(offsetof(mem_stream_t, m_isLink)),
		m_isStreamCond("stream condition")
	{ 
		m_iNumChunks = nchunks ; 
		init() ; 
	}  

	virtual w_rc_t	read( C_OLAPTuple& aTuple, bool& endChunk ) ; 
	virtual w_rc_t	write( C_OLAPTuple& tuple, bool endChunk = false ) ; 

	virtual w_rc_t	read( C_OLAPChunk& aChunk ) ; 
	virtual w_rc_t	write( C_OLAPChunk& aChunk ) ; 

	virtual w_rc_t	SendEndOfStream() ; 
	virtual w_rc_t	SendAbnormalTermination() ; 

protected:

	w_rc_t	init() ; 

	int32				m_iNumChunks ; 
	w_list_t<mem_stream_t>		m_isInputList ; 
	w_list_t<mem_stream_t>		m_isOutputList ; 

	C_OLAPChunk			m_currInput ; 
	C_OLAPChunk			m_currOutput ; 
	
	scond_t                 	m_isStreamCond ;
        smutex_t                	m_isStreamMutex ;

	mem_stream_t*			m_isException ; 
} ; 

#endif 
