/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

///////////////////////////////////////////////////////////////////
//
// olap_tuple_client_shore.cpp
//
// This file defines the shore side of the stream that streams  
// tuples to the client
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_tuple_client_shore.h"
#include "olap_comm_generic.h"

w_rc_t
C_OLAPTuple_ShoreClientStream::read( C_OLAP_Tuple& /*aTuple*/, 
				     bool& /*eos*/, bool& /*exception*/ )
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
        return RCOK ;
}

w_rc_t
C_OLAPTuple_ShoreClientStream::SendExceptionDownStream()
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
        return RCOK ;
}

w_rc_t 
C_OLAPTuple_ShoreClientStream::sendAPage( bool& exception )
{
	C_OLAPComm_Reconciler		isReconciler ; 
	C_OLAPComm_Stream_Packet	isPacket(m_isOutputPage) ; 
	bool				block ; 

	exception = false ; 

	W_DO( m_isSystem->WillBlockRead( block ) ) ; 

	if ( !block ) 
	{
		C_OLAPComm_Packet*	isPkt ; 
		W_DO( m_isSystem->ReceiveData( isReconciler, isPkt ) ) ;  
		
		if ( isPkt->GetPktType() == OLAPCOMM_ERROR_PACKET )
			exception = true ; 

		delete isPkt ; return RCOK ; 
	}

	W_DO( m_isSystem->SendData( isReconciler, isPacket ) ) ; 

	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreClientStream::write( C_OLAP_Tuple& aTuple, bool& exception )
{
	exception = false ; 

	bool page_full ; 
	m_isOutputPage.write( aTuple, page_full ) ;
	
	if ( page_full )
	{
		W_DO( sendAPage( exception ) ) ;
		m_isOutputPage.reset() ; 

		if ( !exception )
			m_isOutputPage.write( aTuple, page_full ) ;
	}

	return RCOK ;
}

w_rc_t 
C_OLAPTuple_ShoreClientStream::SendEndOfStream()
{
	bool exception ;

	// Flush out the current contents
	W_DO( sendAPage( exception ) ) ; 
	if ( exception ) return RCOK ; 
	
	C_OLAPComm_Reconciler		isReconciler ; 
	C_OLAPComm_Empty_Packet		isPacket ; 

	W_DO( m_isSystem->SendData( isReconciler, isPacket ) ) ; 
	
	return RCOK ; 
}

void C_OLAPTuple_ShoreClientStream::Reset()
{
	if ( m_isInputPage ) delete m_isInputPage ; 
	m_isInputPage = NULL ; 
	m_isOutputPage.reset() ; 
}

w_rc_t 
C_OLAPTuple_ShoreClientStream::SendExceptionUpStream()
{
	C_OLAPComm_Reconciler		isReconciler ; 
	C_OLAPComm_Error_Packet		isPacket(NULL) ; 
	bool				block ; 

	// check if any exception is already on the way
	W_DO( m_isSystem->WillBlockRead( block ) ) ; 

	if ( !block ) 
	{
		C_OLAPComm_Packet*	isPkt ; 
		W_DO( m_isSystem->ReceiveData( isReconciler, isPkt ) ) ;  
		
		if ( isPkt->GetPktType() == OLAPCOMM_ERROR_PACKET )
		{
			delete isPkt ; return RCOK ; 
		}
	}

	W_DO( m_isSystem->SendData( isReconciler, isPacket ) ) ; 
	
	return RCOK ; 
}

#ifdef __GNUG__
template class w_auto_delete_t<C_OLAPTuple_ShoreClientStream>;
template class w_auto_delete_t<C_OLAPTuple_ShoreClientStream>;
#endif

