/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

///////////////////////////////////////////////////////////////////
//
// olap_tuple_memory_stream.h
//
// This file defines the defines the tuple memory stream. It is 
// derived from the abstract tuple stream class.  
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_TUPLE_MEMORY_STREAM_H)
#define	__OLAP_TUPLE_MEMORY_STREAM_H

#include <sthread.h>
#include "w_list.h"
#include "olap_tuple_stream.h"

class C_OLAPTuple_MemoryStream : public C_OLAPTuple_Stream {
protected:
	struct mem_stream_t {
		C_OLAP_Page*		m_isPage ; 
		OLAP_Stream_Msg_Type	m_isMsgType ; 
		w_link_t                m_isLink ;

		mem_stream_t() ; 
		~mem_stream_t() ; 
	} ;
public:

	C_OLAPTuple_MemoryStream( int32 nbuffers = OLAP_MIN_PAGE_BUFFERS ) : 
		m_isInputList(offsetof(mem_stream_t, m_isLink)),
		m_isOutputList(offsetof(mem_stream_t, m_isLink)),
		m_isStreamCond("stream condition")
	{ 
		m_iBuffers = nbuffers ; 
		init() ; 
	}  

        virtual ~C_OLAPTuple_MemoryStream() ;
	
	virtual w_rc_t read( C_OLAP_Tuple& aTuple, bool& eos, bool& exception ) ; 
	virtual w_rc_t write( C_OLAP_Tuple& aTuple, bool& exception ) ; 

	virtual w_rc_t read( C_OLAP_Page*& aPage, bool& eos, bool& exception ) ; 
	virtual w_rc_t write( C_OLAP_Page* aPage, bool& exception ) ; 

	virtual w_rc_t SendEndOfStream() ; 
	virtual w_rc_t SendExceptionUpStream() ; 
	virtual w_rc_t SendExceptionDownStream() ; 
	virtual void Reset() ;  

protected:

	virtual w_rc_t sendAPage( mem_stream_t* page, bool eos = false ) ; 
	virtual w_rc_t recvAPage( mem_stream_t*& page, bool& eos, bool& exception ) ; 
	virtual w_rc_t releaseAPage( mem_stream_t* page ) ; 
	virtual w_rc_t getAFreePage( mem_stream_t*& page, bool& exception ) ; 

	void				init() ; 

	int32				m_iBuffers ;
	w_list_t<mem_stream_t>		m_isInputList ; 
	w_list_t<mem_stream_t>		m_isOutputList ; 

	mem_stream_t*			m_currInput ; 
	mem_stream_t*			m_currOutput ; 

	scond_t                 	m_isStreamCond ;
        smutex_t                	m_isStreamMutex ;

	mem_stream_t*			m_isInputException ; 
	mem_stream_t*			m_isOutputException ; 
} ; 

#endif 
