/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_tuple_network_stream.cpp
//
// This file defines the network tuple stream for moving tuples from
// one operator to another operator. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if defined(SOLARIS2)
#define BSD_COMP
#include <sys/ioctl.h>
#undef BSD_COMP
#endif

#include "olap_server_globals.h"
#include "olap_tuple_network_stream.h"
#include "sfile.h"

#include "olap_thread_print.h"

C_OLAPTuple_NetworkStream::network_stream_t::network_stream_t()
{
        m_isPage = new C_OLAP_Page ;
        if ( !m_isPage ) W_FATAL( BORG_OUTOFMEMORY ) ;
}

C_OLAPTuple_NetworkStream::network_stream_t::~network_stream_t()
{
        if ( m_isPage ) delete m_isPage ;
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::recvAPage( network_stream_t*& page,
                                            bool& eos, bool& exception )
{
	eos = exception = false ; 

	int32 nbytes = 0 ; 
	W_DO( actual_read( (char *)&page, sizeof(page), nbytes ) ) ; 

	//cerr << lockp << "Shore Received page (recvAPage) " << page << endl << unlockp
; 

	if ( page->m_isMsgType == OLAP_END_OF_STREAM )
	{
		// release the page
		page->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
		page->m_isPage->reset() ; 

		nbytes = 0 ; 
		W_DO( actual_write( (char *)&page, sizeof(page), nbytes ) ) ; 

		//cerr << lockp << "Shore Sent end of stream page (recvAPage)" << page << endl << unlockp ; 

		eos = true ; exception = false ; 
		return RCOK ; 
	} 
	else if ( page->m_isMsgType == OLAP_ABNORMAL_TERMINATION )
        {
		page->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
                page->m_isPage->reset() ;

		nbytes = 0 ; 
		W_DO( actual_write( (char *)&page, sizeof(page), nbytes ) ) ; 

		//cerr << lockp << "Shore Sent exception page (recvAPage)" << page << endl << unlockp ; 

		eos = false ; exception = true ; 
                return RCOK ;
        }

	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::releaseAPage( network_stream_t* page )
{
	page->m_isMsgType = OLAP_REQUEST_MORE_DATA ;
        page->m_isPage->reset() ;

	int32 nbytes = 0 ; 
	W_DO( actual_write( (char *)&page, sizeof(page), nbytes ) ) ; 

	//cerr << lockp << "Shore Sent page (releaseAPage)" << page << endl << unlockp
; 
	
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::read( C_OLAP_Tuple& aTuple, 
                                       bool& eos, bool& exception )
{
	eos = exception = false ; 
	if ( !m_currOutput )
	{
                W_DO( recvAPage( m_currOutput, eos, exception ) ) ;
		if ( eos || exception ) return RCOK ;
	}

	bool eop ; 
	m_currOutput->m_isPage->read( aTuple, eop ) ;

        if ( eop )
        {
                W_DO( releaseAPage( m_currOutput ) ) ;
		m_currOutput = NULL ; 

                W_DO( recvAPage( m_currOutput, eos, exception ) ) ;
		if ( !eos && !exception )
                	m_currOutput->m_isPage->read( aTuple, eop ) ;
        }

	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::read( C_OLAP_Page*& page, 
                                       bool& eos, bool& exception )
{
	eos = exception = false ; 

	network_stream_t*   new_page = NULL ;

	W_DO( recvAPage( new_page, eos, exception ) ) ;
	if ( eos || exception ) return RCOK ;

	page = new_page->m_isPage ;

	new_page->m_isPage = new C_OLAP_Page ;
	if ( !new_page->m_isPage ) W_FATAL(BORG_OUTOFMEMORY) ;

	W_DO( releaseAPage( new_page ) ) ;

	return RCOK ;
}

w_rc_t
C_OLAPTuple_ShoreNetwork_Stream::flushAllPages()
{
	network_stream_t*	freep = NULL ; 
	int32			nbytes = 0 ; 

	W_DO( actual_read( (char *)&freep, sizeof(freep), nbytes) ); 
		
	while ( freep->m_isMsgType != OLAP_EXCEPTION_ACKNOWLEDGE )
	{
		//cerr << lockp << "Shore Received page (flushAllPages) " << freep << endl << unlockp ; 

                freep->m_isMsgType = OLAP_REQUEST_MORE_DATA ;
                freep->m_isPage->reset() ;

		// Re-send the page
		W_DO( actual_write( (char *)&freep, sizeof(freep), nbytes) ) ;
		//cerr << lockp << "Shore Sent page (flushAllPages) " << freep << endl << unlockp ; 

		// Read the next page
		nbytes = 0 ; 
		W_DO( actual_read( (char *)&freep, sizeof(freep), nbytes) ); 
	}

	W_DO( actual_write( (char *)&freep, sizeof(freep), nbytes) ) ;
	//cerr << lockp << "Shore Sent last page (flushAllPages) " << freep << endl << unlockp ; 

	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::SendExceptionDownStream()
{
	bool eos, exception ;

	if ( !m_currOutput )
	{
                W_DO( recvAPage( m_currOutput, eos, exception ) ) ;
		if ( eos || exception ) return RCOK ; 
	}

	m_currOutput->m_isMsgType = OLAP_ABNORMAL_TERMINATION ; 
       	m_currOutput->m_isPage->reset() ;

	// Send the exception down stream
	int32 nbytes = 0 ; 
	W_DO( actual_write((char *)&m_currOutput, sizeof(m_currOutput), nbytes)); 
	//cerr << lockp << "Shore Sent Exception page (SendDownStreamException)" << m_currOutput << endl << unlockp ; 

	m_currOutput = NULL ; 

	W_DO( flushAllPages() ) ; 

	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::write( C_OLAP_Tuple& /*aTuple*/, 
                                        bool& /*exception*/ )
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::SendEndOfStream() 
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::SendExceptionUpStream() 
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
	return RCOK ; 
}

C_OLAPTuple_NativeNetwork_Stream::C_OLAPTuple_NativeNetwork_Stream( 
					SOCKET sock, 
					u_int32 nbuffers ) :
		 m_isFreeList(offsetof(network_stream_t, m_isLink))
{
	m_iSocket = sock ; m_iBuffers = nbuffers ;
	m_currInput = NULL ; 

	for ( u_int32 i = 0 ; i < m_iBuffers ; i++ )
        {
                network_stream_t*   isNetStream = new network_stream_t ;
                if ( !isNetStream ) W_FATAL(BORG_OUTOFMEMORY) ;

                isNetStream->m_isMsgType = OLAP_REQUEST_MORE_DATA ;
                isNetStream->m_isPage->reset() ;

                m_isFreeList.push( isNetStream ) ;
        }
}

C_OLAPTuple_NativeNetwork_Stream::~C_OLAPTuple_NativeNetwork_Stream() 
{
	network_stream_t*	isNetStream = NULL ; 

	while ( (isNetStream = m_isFreeList.pop()) )
	{
		delete isNetStream ; 
	} 
}

void C_OLAPTuple_NativeNetwork_Stream::Reset()
{
	m_currInput = NULL ; 
}

w_rc_t
C_OLAPTuple_NativeNetwork_Stream::recvAllPages()
{
	int32 nbytes ; 

	// make sure we have received all the pages
	while ( m_isFreeList.num_members() < m_iBuffers )
	{
		network_stream_t*	freep = NULL ; 

		nbytes = 0 ; 
		W_DO( actual_read((char *)&freep, sizeof(freep), nbytes) );

		//cerr << lockp << "recvAllPages " << freep << endl << unlockp ; 
		freep->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
       		freep->m_isPage->reset() ;
		m_isFreeList.push( freep ) ; 
	}

	w_assert3( m_isFreeList.num_members() == m_iBuffers ) ; 

	return RCOK ; 
}

w_rc_t
C_OLAPTuple_NativeNetwork_Stream::checkPossibleExceptions( bool& exception )
{
	network_stream_t*       freep = NULL ;
	int32			nbytes = 0 ; 
	exception = false ; 

	while( !willReadBlock() )
        {
		nbytes = 0 ;
		W_DO( actual_read((char *)&freep, sizeof(freep), nbytes) );

		//cerr << lockp << "Received page (checkPossibleExceptions)" << freep << endl << unlockp ;

		if ( freep->m_isMsgType == OLAP_ABNORMAL_TERMINATION )
		{
                        exception = true ;

			// Send exception acknowledgement
                	freep->m_isMsgType = OLAP_EXCEPTION_ACKNOWLEDGE ;
                	freep->m_isPage->reset() ;

                	W_DO( actual_write( (char *)&freep, sizeof(freep), nbytes ) );

                	return RCOK ;
		}


                freep->m_isMsgType = OLAP_REQUEST_MORE_DATA ;
                freep->m_isPage->reset() ;
                m_isFreeList.push( freep ) ;
        }

	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::sendAPage( network_stream_t*& page, 
                                             bool& exception, bool eos )
{
	int32	nbytes = 0 ; 

	// Before sending the page check if there are any exceptions
	W_DO( checkPossibleExceptions( exception ) ) ;

	if ( exception ) 
	{
		// make sure the page that is being sent
		page->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
		page->m_isPage->reset() ; 

		m_isFreeList.push( page ) ;
		page = NULL ; return RCOK ; 
	}
	
	// Now send the page
        page->m_isMsgType = eos ? OLAP_END_OF_STREAM : OLAP_MORE_DATA ;

	W_DO( actual_write( (char *)&page, sizeof(page), nbytes ) ) ; 

	//cerr << lockp << "Sent page (sendAPage)" << page << endl << unlockp ; 

	page = NULL ; 
        return RCOK ;
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::getAFreePage( network_stream_t*& page, 
                                                bool& exception )
{
	exception = false ; 

	if ( !m_isFreeList.is_empty() ) 
	{
		page = m_isFreeList.pop() ; 				
		return RCOK ; 
	}

	network_stream_t*	freep = NULL ;
	int32 nbytes = 0 ; 

	// Wait until a free page is available 
	W_DO( actual_read( (char *)&freep, sizeof(freep), nbytes) ) ; 

	//cerr << lockp << "Received page (getAFreePage)" << freep << endl << unlockp ; 

	if ( freep->m_isMsgType == OLAP_ABNORMAL_TERMINATION )
	{
		exception = true ; page = NULL ; 

		freep->m_isMsgType = OLAP_EXCEPTION_ACKNOWLEDGE ; 
		freep->m_isPage->reset() ; 
			
		W_DO( actual_write( (char *)&freep, sizeof(freep), nbytes ) ) ; 
		return RCOK ;  
	}

	freep->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
       	freep->m_isPage->reset() ;

	page = freep ;

	//cerr << lockp << "free page (getAFreePage) " << page << endl << unlockp ; 
	return RCOK ;
}


w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::write( C_OLAP_Tuple& aTuple, bool& exception )
{
	exception = false ; 

	if ( !m_currInput )
	{
        	W_DO( getAFreePage( m_currInput, exception ) ) ;
		if ( exception ) 
		{
			W_DO( recvAllPages() ) ; return RCOK ; 
		}
	}	

	bool page_full ; 
        m_currInput->m_isPage->write(aTuple, page_full);

	if ( page_full ) 
        {
                W_DO( sendAPage( m_currInput, exception ) ) ;
		if ( exception ) 
		{
			W_DO( recvAllPages() ) ; return RCOK ; 
		}

                W_DO( getAFreePage( m_currInput, exception ) ) ;

		if ( !exception ) 
                	m_currInput->m_isPage->write(aTuple, page_full) ;
		else
			W_DO( recvAllPages() ) ; 
        }

        return RCOK ;
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::SendEndOfStream()
{
	bool exception = false ; 

	// Flush the last page
	if ( m_currInput )
	{
		W_DO( sendAPage( m_currInput, exception ) ) ;
		if ( exception ) 
		{
			W_DO( recvAllPages() ) ;
			return RCOK ; 
		}
	}

	// Get a free page
	W_DO( getAFreePage( m_currInput, exception ) ) ;
	if ( exception ) 
	{
		W_DO( recvAllPages() ) ;
		return RCOK ; 
	}
	
	// Send end of stream page
        W_DO( sendAPage( m_currInput, exception, true ) ) ;
	if ( exception ) 
	{
		W_DO( recvAllPages() ) ; 
		return RCOK ; 
	}

	// Wait until all the pages are returned back
	W_DO( recvAllPages() ) ; 
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::SendExceptionUpStream()
{
	bool exception = false ; 

	if ( !m_currInput )
	{
        	W_DO( getAFreePage( m_currInput, exception ) ) ;
		if ( exception ) 
		{
			W_DO( recvAllPages() ) ; 
			return RCOK ; 	
		}
	}
	
	m_currInput->m_isMsgType = OLAP_ABNORMAL_TERMINATION ; 
        m_currInput->m_isPage->reset() ;

        W_DO( sendAPage( m_currInput, exception ) ) ;
	if ( exception ) 
	{
		W_DO( recvAllPages() ) ; return RCOK ; 
	}

	W_DO( recvAllPages() ) ; 
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::read( C_OLAP_Tuple& /*aTuple*/,
	                                bool& /*eos*/, bool& /*exception*/ )
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::SendExceptionDownStream()
{
	W_FATAL( BORGST_ERROR_STREAM_USAGE );
	return RCOK ; 
}

bool 
C_OLAPTuple_ShoreNetwork_Stream::willReadBlock()
{
	fd_set		rset ; 
	FD_ZERO(&rset) ; 
	FD_SET(m_isSocket->fd(), &rset) ; 
	struct timeval timeout;
	timeout.tv_sec  = 0; timeout.tv_usec = 0;

	int32 n = select( m_isSocket->fd()+1 , &rset, NULL, NULL, &timeout ) ;  

	if ( n < 0 ) W_FATAL( fcOS ) ; 

        return n > 0 ? true : false ;
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::actual_read( 
				char* where, int32 size, 
				int32& num_bytes )
{
	int32   pkt_read = 0, temp = 0 ;
	while ( pkt_read < size )
	{
		W_DO( m_isSocket->read( where + pkt_read, size - pkt_read, temp ) ) ; 
		pkt_read += temp ;
	}

	num_bytes = pkt_read ; 
	return RCOK ; 
}

w_rc_t 
C_OLAPTuple_ShoreNetwork_Stream::actual_write( 
				char* where, int32 size, 
				int32& num_bytes )
{
	int32   pkt_write = 0, temp = 0 ;
	while ( pkt_write < size )
	{
		W_DO( m_isSocket->write( where + pkt_write, size - pkt_write, temp ) ) ; 
		pkt_write += temp ;
	}

	num_bytes = pkt_write ; 
	return RCOK ; 
}

bool 
C_OLAPTuple_NativeNetwork_Stream::willReadBlock()
{
	fd_set		rset ; 
	FD_ZERO(&rset) ; 
	FD_SET(m_iSocket, &rset) ; 
	struct timeval timeout;
	timeout.tv_sec  = 0; timeout.tv_usec = 0;

	int32 n = select( m_iSocket +1 , &rset, NULL, NULL, &timeout ) ;  

	if ( n < 0 ) W_FATAL( fcOS ) ; 

        return n > 0 ? true : false ;
}

w_rc_t 
C_OLAPTuple_NativeNetwork_Stream::actual_read( 
				char* where, int size, 
				int& num_bytes )
{
	int32   pkt_read = 0, temp = 0 ;
	while ( pkt_read < size )
	{
		temp = recv( m_iSocket, where + pkt_read, size - pkt_read, 0 ) ; 
		if ( temp  == SOCKET_ERROR) 
			return RC(BORGST_STREAM_READ_ERROR) ; 

		pkt_read += temp ;
	}

	num_bytes = pkt_read ; 
	return RCOK ; 
}

w_rc_t C_OLAPTuple_NativeNetwork_Stream::actual_write( 
				char* where, 
				int size, 
				int& num_bytes )
{
	int32   pkt_write = 0, temp = 0 ;

	while ( pkt_write < size )
	{
		temp = send(m_iSocket, where + pkt_write, size - pkt_write, 0) ;
	
		if ( temp  == SOCKET_ERROR) 
			return RC(BORGST_STREAM_WRITE_ERROR) ; 

		pkt_write += temp ;
	}

	num_bytes = pkt_write ; 
	return RCOK ; 
}

#ifdef __GNUG__
template class w_list_t<C_OLAPTuple_NetworkStream::network_stream_t>;
template class w_list_i<C_OLAPTuple_NetworkStream::network_stream_t>;
#endif

