/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

///////////////////////////////////////////////////////////////////
//
// olap_tuple_network_stream.h
//
// This file defines the stream that transports tuples between 
// backend thread and shore thread operators. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_TUPLE_NETWORK_STREAM_H)
#define	__OLAP_TUPLE_NETWORK_STREAM_H

#include "w_list.h"
#include "olap_tuple_stream.h"

class sfile_t ; 

class C_OLAPTuple_NetworkStream : public C_OLAPTuple_Stream {
protected:
	struct network_stream_t {
		C_OLAP_Page*            m_isPage ;
                OLAP_Stream_Msg_Type    m_isMsgType ;
                w_link_t                m_isLink ;

		network_stream_t() ; 
		~network_stream_t() ; 
	} ; 

	virtual void Reset()
	{ }

	virtual bool willReadBlock() = 0 ; 
	virtual w_rc_t actual_read(char* where, int32 size, int32& num_bytes) = 0 ; 
	virtual w_rc_t actual_write(char* where, int32 size, int32& num_bytes) = 0 ;
} ; 

class C_OLAPTuple_ShoreNetwork_Stream : public C_OLAPTuple_NetworkStream {
public:
	C_OLAPTuple_ShoreNetwork_Stream( sfile_t* bkendsock ) 
	{
		m_isSocket = bkendsock ; m_currOutput = NULL ;  
	}
				
	virtual w_rc_t  read(C_OLAP_Tuple& aTuple, bool& eos, bool& exception);
	virtual w_rc_t  read(C_OLAP_Page*& aPage, bool& eos, bool& exception);
        virtual w_rc_t  SendExceptionDownStream() ;

        virtual w_rc_t  write( C_OLAP_Tuple& tuple, bool& exception ) ;

        virtual w_rc_t write( C_OLAP_Page* /*page*/, bool& exception )
        {
                exception = true ;
                return RC(BORG_NOTIMPLEMENTED) ;
        }

        virtual w_rc_t  SendEndOfStream() ;
        virtual w_rc_t  SendExceptionUpStream() ;

protected:

	w_rc_t flushAllPages() ; 

	w_rc_t recvAPage( network_stream_t*& page, bool& eos, 
                                  bool& exception ) ; 
	w_rc_t releaseAPage( network_stream_t* page ) ; 

	virtual bool willReadBlock() ; 
	virtual w_rc_t actual_read( char* where, int32 size, int32& num_bytes ) ; 
	virtual w_rc_t actual_write( char* where, int32 size, int32& num_bytes ) ; 

	network_stream_t*		m_currOutput ; 
	sfile_t*			m_isSocket ; 
} ;

class C_OLAPTuple_NativeNetwork_Stream : public C_OLAPTuple_NetworkStream {
public:
	C_OLAPTuple_NativeNetwork_Stream( SOCKET sock, 
				u_int32 nbuffers = OLAP_MIN_PAGE_BUFFERS ) ; 
				
	virtual ~C_OLAPTuple_NativeNetwork_Stream() ;

	virtual w_rc_t  read( C_OLAP_Tuple& aTuple, bool& eos, bool& exception ) ;
        virtual w_rc_t  SendExceptionDownStream() ;

        virtual w_rc_t  write( C_OLAP_Tuple& tuple, bool& exception ) ;

        virtual w_rc_t read( C_OLAP_Page*& /*page*/, bool& eos, bool& exception )
        {
                eos = exception = true ;
                return RC(BORG_NOTIMPLEMENTED) ;
        }

        virtual w_rc_t write( C_OLAP_Page* /*page*/, bool& exception )
        {
                exception = true ;
                return RC(BORG_NOTIMPLEMENTED) ;
        }

        virtual w_rc_t  SendEndOfStream() ;
        virtual w_rc_t  SendExceptionUpStream() ;

	virtual void	Reset() ;
	
protected:

	w_rc_t checkPossibleExceptions( bool& exception ) ; 

	w_rc_t recvAllPages() ; 

	w_rc_t sendAPage( network_stream_t*& page, 
                                  bool& exception, 
                                  bool eos = false ) ; 
	w_rc_t getAFreePage( network_stream_t*& page, bool& exception );

	virtual bool willReadBlock() ; 
	virtual w_rc_t	actual_read( char* where, int32 size, int32& num_bytes ) ; 
	virtual w_rc_t	actual_write( char* where, int32 size, int32& num_bytes ) ; 


	w_list_t<network_stream_t>	m_isFreeList ; 
	u_int32				m_iBuffers ;

	SOCKET				m_iSocket ; 
	network_stream_t*		m_currInput ; 
} ;

#endif 
