/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_page.h
//
// Declares the page class used by BORG. BORG uses a slotted page
// structure.
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////



#if !defined(__OLAP_PAGE_H)
#define __OLAP_PAGE_H

#include "w.h"
#include "olap_types.h"
#include "olap_tuple.h"

class C_OLAP_Page {
public:

    typedef int16	slot_offset_t; 	// -1 if vacant
    typedef u_int16	slot_length_t;
    typedef int16	slotid_t; 

    struct slot_t {
	slot_offset_t offset;		// -1 if vacant
	slot_length_t length;
    };
    
    C_OLAP_Page() { reset() ; } 

    u_int32 tuple_size(slotid_t idx) const;
    void* tuple_addr(slotid_t idx) const;

    u_int32 ntuples() const ; 

    u_int32 free_space_start() const ;  
    u_int32 free_tuple_space() const ; 
    u_int32 free_space() const ; 
    
    slotid_t read( C_OLAP_Tuple& aTuple, bool& eop ) ; 
    void read( C_OLAP_Tuple& aTuple, int slot, bool& eop ) ; 
    void write( C_OLAP_Tuple& aTuple, bool& page_full ) ;

    void reset_scan() {
        next = 0;
    }

    void reset() {
	free = 0 ; nslots = 0 ; next = 0 ; 
    }

    enum {
	page_sz = 512,
	data_sz = (page_sz - 2 * sizeof(int16)
			- 2 * sizeof(slotid_t) - 2 * sizeof(slot_t)),
	max_slot = data_sz / sizeof(slot_t) + 2
    };

    u_int16	nslots;			// number of slots occupied
    u_int16	free;			// offset to free data area
    slotid_t	pad;			// pad for alignment
    slotid_t	next;			// next slot - for iterating 
    char 	data[data_sz];		// must be aligned
    slot_t	reserved_slot[1];	// 2nd slot (declared to align
					// end of _data)
    slot_t	slot[1];		// 1st slot
};

inline u_int32
C_OLAP_Page::tuple_size(slotid_t idx) const
{
    w_assert3(idx >= 0 && idx < nslots);
    return slot[-idx].length;
}

inline void*
C_OLAP_Page::tuple_addr(slotid_t idx) const
{
    w_assert3(idx >= 0 && idx < nslots);
    return (void*) (data + slot[-idx].offset);
}

inline u_int32
C_OLAP_Page::free_space() const
{
    return data_sz - free - sizeof(slot_t)*(nslots > 2 ? nslots - 2 : 0) ; 
}

inline u_int32
C_OLAP_Page::free_tuple_space() const
{
    return free_space() - sizeof(slot_t) ;  
}

inline u_int32
C_OLAP_Page::free_space_start() const
{
   return free ;
}

inline u_int32
C_OLAP_Page::ntuples() const
{
   return nslots ; 
}

#endif 
