/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_tuple.cpp
//
// Implementation of the tuple related classes such as C_OLAP_TupleInfo
// and C_OLAP_Tuple etc declared in "olap_tuple.h"
//
// Author: Prasad M. Deshpande
//
/////////////////////////////////////////////////////////////////////



#include <borgtypes.h>
#include <olap_tuple.h>
#include "olap_utils.h"

OlapType OlapType::BasicOlapTypes[] = {
    { BORG_UNDEF, 0, borgalign(0) },
    { BORG_INT,   sizeof(int), borgalign(sizeof(int)) },
    { BORG_FLOAT, sizeof(float), borgalign(sizeof(float)) },
    { BORG_STR, VARIABLE_LENGTH, VARIABLE_LENGTH }
};

void C_OLAP_Tuple::FillAttr(C_OLAPAttrDesc &attr, void *buf, int32 length)
{
    int newsize;

    if (attr.m_type->VarLength())
	newsize = m_varOffset + length;
    else
	newsize = attr.m_offset + attr.m_type->m_alignedlength;
    
    alloc(newsize);

    if (attr.m_type->VarLength())
    {
	borg_assert3(length >= 0);
	u_int32 len = length;
	memcpy(m_tupBuf + attr.m_offset, &m_varOffset, sizeof(u_int32));
	memcpy(m_tupBuf + attr.m_offset + sizeof(u_int32), &len,
	       sizeof(u_int32));
	memcpy(m_tupBuf + m_varOffset, buf, len);
	m_varOffset += len;
    }
    else
    {
	memset(m_tupBuf + attr.m_offset, 0, attr.m_type->m_alignedlength);
	memcpy(m_tupBuf + attr.m_offset, buf, attr.m_type->m_length);
    }
}
	
void *C_OLAP_Tuple::GetAttr(C_OLAPAttrDesc &attr, u_int32 &length) const
{
    u_int32 varoffset;
    if (attr.m_type->VarLength())
    {
	memcpy(&varoffset, m_tupBuf + attr.m_offset, sizeof(u_int32));
	memcpy(&length, m_tupBuf + attr.m_offset + sizeof(u_int32), 
	       sizeof(u_int32));
	return (void *)(m_tupBuf + varoffset);
    }
    else	// Fixed length attribute...
    {
        length = attr.m_type->m_length;
	return (void *)(m_tupBuf + attr.m_offset);
    }
}    

void C_OLAP_Tuple::print(ostream &str)
{
    if (!m_tupInfo)
    {
	str << "TupInfo not initialized " << endl;
	return ;
    }

    OlapType 	*attrtype;
    void	*buf;
    u_int32	i, len;

    str << "( " ;
    for (i=0; i<m_tupInfo->GetNumAttrs(); i++)
    {
	if (i != 0)
	    str << ", " ;
	attrtype = (*m_tupInfo)[i].m_type;
	buf = GetAttr(i, len);
	attrtype->print(buf, len, str);
    }
    str << " )";
}

void C_OLAP_Tuple::alloc(u_int32 len)
{
    if (len <= m_size)
	return;

    /* expand buffer 2 times */
    u_int32 nsize = m_size ? (m_size << 1) : TupInitSize;

    if (nsize < len)
	nsize = len;

    char *tbuf = new char[nsize]; 
    memcpy(tbuf, m_tupBuf, m_size);
    memset(tbuf+m_size, 0, nsize-m_size);

    if (m_allocFlag)
	DelA(m_tupBuf);

    m_allocFlag = true;
    m_tupBuf = tbuf;
    m_size = nsize;
}

void C_OLAP_TupleInfo::EndAttrDesc()
{
    u_int32 i;

    m_fixedLen = 0;

    for (i=0; i<m_numAttr; i++)
    {
	m_attrDescs[i].m_offset = m_fixedLen;
	if (m_attrDescs[i].m_type->VarLength())
	    m_fixedLen += 2*sizeof(u_int32);
	else
	    m_fixedLen += m_attrDescs[i].m_type->m_alignedlength;
    }
}

C_OLAP_TupleInfo &C_OLAP_TupleInfo::operator=(C_OLAP_TupleInfo &rhs)
{
	m_numAttr = rhs.m_numAttr;
	
	m_attrDescs = rhs.m_attrDescs;

	m_fixedLen = rhs.m_fixedLen;

	return *this;
}

void C_OLAP_TupleInfo::Linearize( C_OLAPComm_DataLinearizer &dLin )
{
    dLin.Pack((int32)m_numAttr);
    ::Linearize(m_attrDescs, dLin);
    dLin.Pack((int32)m_fixedLen);
}

void C_OLAP_TupleInfo::DeLinearize( C_OLAPComm_DataDelinearizer &deLin )
{
    int32 tmp;
    deLin.Unpack(tmp);
    m_numAttr = tmp;
    ::DeLinearize(m_attrDescs, deLin);
    deLin.Unpack(tmp);
    m_fixedLen = tmp;
}

void C_OLAPAttrDesc::Linearize( C_OLAPComm_DataLinearizer &dLin )
{
    dLin.Pack(int32(m_type->m_typeid));
    dLin.Pack((int32)m_offset);
}

void C_OLAPAttrDesc::DeLinearize( C_OLAPComm_DataDelinearizer &deLin )
{
    int32 attrtypeid;
    deLin.Unpack(attrtypeid);

    m_type =  &OlapType::BasicOlapTypes[attrtypeid];
    int32 tmp;
    deLin.Unpack(tmp);
    m_offset = tmp;
}

void OlapType::print(void *buf, u_int32 /*len */, ostream &str)
{
    switch(m_typeid)
    {
    case BORG_INT:
	str << *(int *)buf;
	break;
    case BORG_FLOAT:
	str << *(float *)buf;
	break;
    case BORG_STR:
	str << (char *)buf;
	break;
    default:
	str << "Unknown type " ;
    }
}

#ifdef __GNUG__
template class dynarray<C_OLAPAttrDesc>;
template void Linearize(dynarray <C_OLAPAttrDesc>& arr, C_OLAPComm_DataLinearizer &dLin);
template void DeLinearize(dynarray <C_OLAPAttrDesc>& arr, C_OLAPComm_DataDelinearizer &deLin);
#endif
