/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_type_converter.cpp
//
// Implementation of the type converter routines declared in 
// olap_type_converter.h
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////


#include <assert.h>
#include <stdio.h>

#include <string>

#include "olap_type_converter.h"


void C_OLAP_TypeConverter::string2type(w_string& str, BorgTypes type, DataItem& dataItem)
{
    switch (type)
    {
        case BORG_INT:
            dataItem.Name = (void *) new char[sizeof(int)];
            dataItem.Len  = sizeof(int);
            *( (int *) dataItem.Name) = atoi(str.c_str());
            break;

        case BORG_FLOAT:
            dataItem.Name = (void *) new char[sizeof(float)];
            dataItem.Len  = sizeof(float);
            *( (float *) dataItem.Name) = atof(str.c_str());
            break;

        case BORG_STR:
            dataItem.Name = (void *) new char[strlen(str.c_str())];
            dataItem.Len  = strlen(str.c_str());
            strncpy((char*) dataItem.Name, str.c_str(), dataItem.Len);
            break;

        default:
            assert(0);
    }
}



void C_OLAP_TypeConverter::type2string(DataItem& dataItem, BorgTypes type, w_string& str)
{
    char buf[1024];

    switch (type)
    {
        case BORG_INT:
            assert(dataItem.Len == sizeof(int));
            sprintf(buf, "%d", * (int *) dataItem.Name);
            str = buf;
            break;

        case BORG_FLOAT:
            assert(dataItem.Len == sizeof(float));
            sprintf(buf, "%f", * (float *) dataItem.Name);
            str = buf;
            break;

        case BORG_STR:
            str = (char *) dataItem.Name;
            break;

        default:
            assert(0);
    }
}

