/*<std-header orig-src=SHORE incl-file-exclusion=W_LIST_H>

 $Id: w_nt_string.h,v 1.1.1.1 1998/09/17 21:44:18 karthik Exp $

SHORE -- Scalable Heterogeneous Object REpository

Copyright (c) 1994-98 Computer Sciences Department, University of
                      Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE AUTHORS AND THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY
OF WISCONSIN - MADISON ALLOW FREE USE OF THIS SOFTWARE IN ITS
"AS IS" CONDITION, AND THEY DISCLAIM ANY LIABILITY OF ANY KIND
FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.

This software was developed with support by the Advanced Research
Project Agency, ARPA order number 018 (formerly 8230), monitored by
the U.S. Army Research Laboratory under contract DAAB07-91-C-Q518.

*/

#ifndef W_NT_STRING_H
#define	W_NT_STRING_H

#include "w_stream.h"
#include "w.h"
#include <stdio.h>

class w_string {
public :

	typedef w_base_t::int4_t	sz_type ; 

	typedef const char&		cref_type ; 
	typedef char&			ref_type ; 

	typedef const char*		cptr_type ; 
	typedef char*			ptr_type ; 

	static const sz_type		npos ; 

	w_string() ;

	w_string( cptr_type str ) ; 
	w_string( cptr_type str, sz_type n ) ; 

	w_string( const w_string& str ) ; 
	w_string( const w_string& str, sz_type pos, sz_type n ) ; 
	w_string( const w_string& str, sz_type n ) ; 

	w_string( sz_type n, char c ) ; 

	w_string& operator = ( const w_string& str ) ; 
	w_string& operator = ( cptr_type str ) ; 
	w_string& operator = ( char c ) ; 

	char at(sz_type pos) const ; 
        void set_at(sz_type pos, char c) ;

	char operator [](sz_type pos) const ; 
	
	cptr_type c_str() const ; 
	cptr_type data() const ; 	

	sz_type length() const ; 
	sz_type size() const ; 

	bool empty() const ; 

	w_string& operator += (const w_string& str) ; 
	w_string& operator += (cptr_type str) ; 
	w_string& operator += (char c) ; 

	sz_type find(const w_string& str) const ;  
	sz_type find(cptr_type str) const ;  
	sz_type	find(char c) const ;  

	sz_type rfind(char c) const ; 

	int operator == ( const w_string& str ) const ; 

	int compare(const w_string& str) const ;
	int compare(cptr_type str) const ;

	w_string& erase(sz_type s = 0, sz_type n = npos) ;

        friend istream& operator >> (istream& is, w_string& str) ;
        friend ostream& operator << (ostream& os, const w_string& str) ;

private :

	CString			m_strrep ; 
} ;

inline 
w_string::w_string() 
{
}

inline
w_string::w_string( cptr_type str ) : m_strrep(str)
{
}

inline
w_string::w_string( cptr_type str, sz_type n ) 
{
    if (n != npos)
	m_strrep += CString(str, n);
    else
	m_strrep += CString(str);
}

inline
w_string::w_string( const w_string& str ) : m_strrep(str.m_strrep)
{
}
 
inline
w_string::w_string( const w_string& str, sz_type pos, sz_type n ) 
{
    if (pos != npos)
	if (n != npos)
	    m_strrep += CString(str.c_str() + pos, n);
	else
	    m_strrep += CString(str.c_str());
}


inline
w_string::w_string( const w_string& str, sz_type n ) 
{
    if (n != npos)
	m_strrep += CString(str.c_str(), n);
    else
	m_strrep += CString(str.c_str());
}

inline
w_string::w_string( sz_type n, char c ) : m_strrep(c, n)
{
}

inline w_string&
w_string::operator = ( const w_string& str ) 
{
	m_strrep = str.m_strrep ; return *this ; 
}

inline w_string& 
w_string::operator = ( cptr_type str ) 
{
	m_strrep = str ; return *this ;  
}

inline w_string&
w_string::operator = ( char c ) 
{
	m_strrep = c ; return *this ; 
}

inline char
w_string::at(sz_type pos) const 
{
	return m_strrep.GetAt(pos) ; 
}

inline void
w_string::set_at(sz_type pos, char c) 
{
	m_strrep.SetAt(pos, c) ;
}


inline char 
w_string::operator [](sz_type pos) const 
{
	return m_strrep.GetAt(pos) ; 
}

	
inline w_string::cptr_type 
w_string::c_str() const 
{
	return (LPCTSTR) m_strrep ; 
}

inline w_string::cptr_type
w_string::data() const
{
	return (LPCSTR) m_strrep; 
}

inline w_string::sz_type 
w_string::length() const 
{
	return m_strrep.GetLength() ; 
}

inline w_string::sz_type 
w_string::size() const 
{
	return m_strrep.GetLength() ; 
}

inline bool 
w_string::empty() const
{
    return m_strrep.IsEmpty() == TRUE ? true : false ;
}

inline w_string& 
w_string::operator += (const w_string& str) 
{
	m_strrep += str.m_strrep ; return *this ; 
}

inline w_string& 
w_string::operator += (cptr_type str) 
{
	m_strrep += str ; return *this ; 
}

inline w_string& 
w_string::operator += (char c) 
{
	m_strrep += c ; return *this ;
}

inline w_string::sz_type 
w_string::find(const w_string& str) const 
{
	sz_type pos = m_strrep.Find(str.m_strrep) ; 
	return (pos == -1) ? npos : pos ;
}

inline w_string::sz_type 
w_string::find(cptr_type str) const 
{
	sz_type pos = m_strrep.Find(str) ; 
	return (pos == -1) ? npos : pos ;
}

inline w_string::sz_type	
w_string::find(char c) const 
{
	sz_type pos = m_strrep.Find(c) ; 
	return (pos == -1) ? npos : pos ;
}

inline w_string::sz_type 
w_string::rfind(char c) const 
{
	sz_type pos = m_strrep.ReverseFind(c) ; 
	return (pos == -1) ? npos : pos ;
}

inline int 
w_string::operator == (const w_string& str) const
{
	return m_strrep == str.m_strrep ;  
}

inline int 
w_string::compare(const w_string& str) const
{
	return m_strrep.Compare(str.c_str()) ; 
}

inline int 
w_string::compare(cptr_type str) const
{
	return m_strrep.Compare(str) ; 
}

inline w_string&
w_string::erase(sz_type s, sz_type n)
{
	w_string    tmp(*this);
	sz_type	    pos;

	m_strrep.Empty();
	if ((s == npos) || (n == npos))
	    pos = npos;
	else
	    if (s + n < tmp.length())
	        pos = s + n;
	    else
	        pos = npos;
	*this = w_string(tmp, 0, s);
	*this += w_string(tmp, pos, npos);
	return *this;
}

inline ostream&
operator << (ostream& os, const w_string& str)
{
        os << str.m_strrep ; 
	return os ;
}

inline istream&
operator >> (istream& is, w_string& str)
{
	char	buf[BUFSIZ];	
        is >> buf ; 
	buf[BUFSIZ-1] = '\0';
	str.m_strrep = buf; 
	return is ;
}


#endif

