/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_thread_thread.cpp
//
// This file contains the implementation of the abstraction for 
// threads for solaris. It implements various operations like starting
// of the thread, waiting for a thread to complete, yielding for another
// thread etc 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_types.h"
#include <thread.h>
#include "olap_thread_thread.h"

const u_int32 C_OLAPThread_Thread::STACK_SIZE = 128 ;

/* static thread locals for storing the pointer to the thread instance */
/* Microsoft specific. In other systems you can use thread locals */

C_OLAPThread_Thread* OLAPGetCurrentThread()
{
	return NULL ;
}

vo_ptr C_OLAPThread_Thread::StartFunction(vo_ptr arg)
{
	C_OLAPThread_Thread*	isOlapThread = (C_OLAPThread_Thread *)arg ;

	isOlapThread->m_isRetVal = isOlapThread->Run() ;

	pthread_exit(&(isOlapThread->m_isRetVal)) ; 

	return isOlapThread->m_isRetVal ;
}

int32 C_OLAPThread_Thread::Start()
{
	// Initialize attr needed for pthread_create 
        pthread_attr_setscope(&m_isThreadAttr, PTHREAD_SCOPE_SYSTEM);
        pthread_attr_setstacksize( &m_isThreadAttr, STACK_SIZE * 1024 );

	int32 status = pthread_create( &m_isThreadHandle, &m_isThreadAttr, 
                                 C_OLAPThread_Thread::StartFunction,
                                 (C_OLAPThread_Thread *)this );

	if ( status < 0 ) return -1 ; 

	if ( !m_isThreadHandle ) return -1 ;

	return 0 ; 
}

void C_OLAPThread_Thread::Wait()
{
	vo_ptr	retVal ; 
	
	pthread_join( m_isThreadHandle, &retVal ) ;
}

int32 C_OLAPThread_Thread::Join( C_OLAPThread_Thread& isOlapThread )
{
	int*	status ;

	if ( pthread_join( isOlapThread.m_isThreadHandle, 
			   (void **)&status ) == 0 )
		return 0 ;

	return -1 ; 
}

void C_OLAPThread_Thread::Yield()
{
	thr_yield() ; 
}
