/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_thread_mutex.h
//
// This file contains the interface abstraction for thread mutexes
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_THREAD_MUTEX_H)
#define  __OLAP_THREAD_MUTEX_H

class C_OLAPThread_Mutex {

public :

	C_OLAPThread_Mutex( const char* szName = NULL )
	{
		m_isMutexHandle = CreateMutex(NULL,FALSE,szName) ; 

		if ( GetLastError() == ERROR_ALREADY_EXISTS )
		{
			m_isMutexHandle = OpenMutex(SYNCHRONIZE, FALSE, szName) ; 
		}

		m_bValid = m_isMutexHandle ? true : false ;
	}

	~C_OLAPThread_Mutex()
	{
		CloseHandle(m_isMutexHandle) ; 
		m_bValid = false ;
	}

	int32 Lock( int32 iWaitTime = INFINITE )
	{
		if ( !m_bValid ) return -1 ;

		int32 iRetVal = WaitForSingleObject(m_isMutexHandle, iWaitTime) ;

		if ( iRetVal == WAIT_TIMEOUT ) return -2 ;
		if ( iRetVal == WAIT_OBJECT_0 ) return -1 ; 

		return 0 ; 
	}

	int32 Unlock()
	{
		if ( !m_bValid ) return -1 ; 

		ReleaseMutex(m_isMutexHandle) ; 
		return 0 ; 
	}

protected :

	HANDLE		m_isMutexHandle ;
	bool		m_bValid ;

} ;

#endif
