/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_thread_thread.cpp
//
// This file contains the implementation of the abstraction for
// threads for nt. It implements various operations like starting
// of the thread, waiting for a thread to complete, yielding for another
// thread etc
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <process.h>
#include "olap_types.h"
#include "olap_thread_thread.h"

const u_int32 C_OLAPThread_Thread::STACK_SIZE = 128 ;

/* static thread locals for storing the pointer to the thread instance */
/* Microsoft specific. In other systems you can use thread locals */

static __declspec(thread) C_OLAPThread_Thread*	gt_isMyThread ;

C_OLAPThread_Thread* OLAPGetCurrentThread()
{
	return gt_isMyThread ;
}

u_int32 _stdcall C_OLAPThread_Thread::StartFunction(void* arg)
{
	C_OLAPThread_Thread*	isOlapThread = (C_OLAPThread_Thread *)arg ;

	gt_isMyThread = isOlapThread ;

	int32	iRetVal = isOlapThread->Run() ;

	_endthreadex(iRetVal) ; 

	return iRetVal ;
}


int32 C_OLAPThread_Thread::Start()
{
	u_int32		iThreadId ;

	// We create the thread suspended. This is because handle
	// to the thread is valid member and there is a race condition
	m_isThreadHandle = (HANDLE)_beginthreadex( 
								NULL, 
								STACK_SIZE * 1024,
								C_OLAPThread_Thread::StartFunction,
								this, CREATE_SUSPENDED, &iThreadId ) ;
	
	if ( !m_isThreadHandle ) return -1 ;

	if ( ResumeThread(m_isThreadHandle)  == 0xFFFFFFFF )
		return -1 ;

	return 0 ; 
}

int32 C_OLAPThread_Thread::Join( C_OLAPThread_Thread& isOlapThread )
{
	if ( WaitForSingleObject(isOlapThread.m_isThreadHandle, 
							INFINITE) == WAIT_OBJECT_0 )
		return 0 ;

	return -1 ;
}
